#include "jboci.h"

void ReplaceCharInString(char *str, char find, char replace)
{
	char *pos = str;
	while ((pos = strchr(pos, find)))
		*pos = replace;
}

void DeleteCharInString(char *str, char find)
{
	char *pos = str;
	while ((pos = strchr(pos, find)))
		memmove(pos, pos + 1, strlen(pos));
}

void ExportCSVData(TextWidget *text, char *filename)
{
	int x, do_it = 1, bufsize = 0, recordlength = 0;
	char *buffer;
	FILE *of;
	size_t bytes_read;
	HCURSOR c;
	char *outbuf;

	c = SetCursor(LoadCursor(NULL, IDC_WAIT));
	of = fopen(filename, "w+");
	outbuf = malloc(1024*1024); // 1MB
	if (!outbuf)
	{
		WPrintf("Out of memory\n");
		SetCursor(c);
		return;
	}
	setvbuf(of, outbuf, _IOFBF, 1024*1024);
	if (of == NULL)
	{
		Error(NULL, "%s", strerror(errno));
		free(outbuf);
		SetCursor(c);
		return;
	}
	fseek(tmpfile_read, 0, SEEK_SET);
	for (x = 0; x < ncols; x++)
	{
		if (text->datasource == DS_ORACLE)
		{
			recordlength += oci_cols[x].col_size;
			if (oci_cols[x].col_size > bufsize)
				bufsize = oci_cols[x].col_size;
			if (x)
				fprintf(of, ";");
			fprintf(of, "%s", oci_cols[x].col_name);
		}
		else
		{
			recordlength += (int)odbc_cols[x].col_size;
			if ((int)odbc_cols[x].col_size > bufsize)
				bufsize = (int)odbc_cols[x].col_size;
			if (x)
				fprintf(of, ";");
			fprintf(of, "%s", odbc_cols[x].col_name);
		}
	}
	fprintf(of, "\n");
	if (bufsize)
		buffer = (char*)malloc(bufsize * 2 + 1);
	else
		return;
	if (buffer == NULL)
	{
		Error(NULL, "Out of memory");
		return;
	}
	while (do_it)
	{
		for (x = 0; x < ncols; x++)
		{
			if (text->datasource == DS_ORACLE)
				bytes_read = fread(buffer, 1, oci_cols[x].col_size, tmpfile_read);
			else
				bytes_read = fread(buffer, 1, odbc_cols[x].col_size, tmpfile_read);
			if (bytes_read == 0)
			{
				do_it = 0;
				break;
			}
			buffer[bytes_read] = '\0';
			if (x)
				fwrite(";\"", 1, 2, of);
			else
				fwrite("\"", 1, 1, of);
			if (strchr(buffer, '\"'))
				ReplaceString(buffer, "\"", "\"\"");
			ReplaceCharInString(buffer, '\n', ' ');
			DeleteCharInString(buffer, '\r');
			fwrite(buffer, 1, strlen(buffer), of);
			fwrite("\"", 1, 1, of);
		}
		fprintf(of, "\n");
	}
	fclose(of);
	free(buffer);
	free(outbuf);
	SetCursor(c);
}

void ExportSQLData(TextWidget *text, char *filename)
{
	int x, y, do_it = 1, bufsize = 0, recordlength = 0;
	char *buffer, date_format[100],*outbuf;
	FILE *of;
	size_t bytes_read;
	HCURSOR c;

	c = SetCursor(LoadCursor(NULL, IDC_WAIT));
	of = fopen(filename, "w+");
	if (of == NULL)
	{
		Error(NULL, "%s", strerror(errno));
		SetCursor(c);
		return;
	}

	outbuf = malloc(1024*1024); // 1MB
	if (!outbuf)
	{
		WPrintf("Out of memory\n");
		SetCursor(c);
		return;
	}
	setvbuf(of, outbuf, _IOFBF, 1024 * 1024);

	fseek(tmpfile_read, 0, SEEK_SET);
	if (text->datasource == DS_ORACLE)
	{
		for (x = 0; x < ncols; x++)
		{
			recordlength += oci_cols[x].col_size;
			if (oci_cols[x].col_size > bufsize)
				bufsize = oci_cols[x].col_size;
		}
	}
	else
	{
		for (x = 0; x < ncols; x++)
		{
			recordlength += (int)odbc_cols[x].col_size;
			if ((int)odbc_cols[x].col_size > bufsize)
				bufsize = (int)odbc_cols[x].col_size;
		}
	}
	if (bufsize)
		buffer = (char*)malloc(bufsize * 2 + 1);
	else
		return;
	if (buffer == NULL)
	{
		SetCursor(c);
		Error(NULL, "Out of memory");
		free(outbuf);
		return;
	}
	if (text->datasource == DS_ODBC || GetSingleValue(g_oci,"select sys_context('userenv','nls_date_format') from dual", date_format, 100) != 0)
		date_format[0] = '\0';
	fprintf(of, "set scan off\n");
	while (do_it)
	{
		for (x = 0; x < ncols; x++)
		{
			if (text->datasource == DS_ORACLE)
				bytes_read = fread(buffer, 1, oci_cols[x].col_size, tmpfile_read);
			else
				bytes_read = fread(buffer, 1, odbc_cols[x].col_size, tmpfile_read);
			if (bytes_read == 0)
			{
				do_it = 0;
				break;
			}
			if (x)
				fwrite(",", 1, 1, of);
			else
			{
				fwrite("insert into (", 1, 13, of);
				fwrite(old_commands[nold_commands - 1].command, 1, strlen(old_commands[nold_commands - 1].command),of);
				fwrite(") (", 1, 3, of);
				for (y = 0; y < ncols; y++)
				{
					if (y)
						fwrite(",\"", 1, 2, of);
					else
						fwrite("\"", 1, 1, of);
					if (text->datasource == DS_ORACLE)
						fwrite(oci_cols[y].col_name, 1, strlen(oci_cols[y].col_name), of);
					else
						fwrite(odbc_cols[y].col_name, 1, strlen(odbc_cols[y].col_name), of);
					fwrite("\"", 1, 1, of);
				}
				fwrite(")\nvalues\n(", 1, 10, of);
			}
			buffer[bytes_read] = '\0';
			if (text->datasource == DS_ORACLE)
			{
				switch (oci_cols[x].datatype)
				{
					case OCI_TYPECODE_REAL:
					case OCI_TYPECODE_DOUBLE:
					case OCI_TYPECODE_FLOAT:
					case OCI_TYPECODE_NUMBER:
					case OCI_TYPECODE_DECIMAL:
					case OCI_TYPECODE_OCTET:
					case OCI_TYPECODE_INTEGER:
					case OCI_TYPECODE_SMALLINT:
						if (buffer[0] != '\0')
							fwrite(buffer,1,strlen(buffer),of);
						else
							fwrite("NULL", 1, 4, of);
						break;
					case OCI_TYPECODE_RAW:
					case OCI_TYPECODE_VARCHAR2:
					case OCI_TYPECODE_VARCHAR:
					case OCI_TYPECODE_CHAR:
					case OCI_TYPECODE_CLOB:
					case OCI_TYPECODE_BLOB:
					case OCI_TYPECODE_BFILE:
					case SQLT_LNG:
					case SQLT_BIN:
					case SQLT_LBI:
					case SQLT_RID:
					case SQLT_RDD:
						if (buffer[0] != '\0')
						{
							ReplaceString(buffer, "'", "''");
							fwrite("'", 1, 1, of);
							fwrite(buffer, 1, strlen(buffer), of);
							fwrite("'", 1, 1, of);
						}
						else
						{
							fwrite("NULL", 1, 4, of);
						}
						break;
					case OCI_TYPECODE_DATE:
						if (buffer[0] != '\0')
						{
							if (date_format[0])
							{
								fwrite("to_date('", 1, 9, of);
								fwrite(buffer, 1, strlen(buffer), of);
								fwrite("','", 1, 3, of);
								fwrite(date_format, 1, strlen(date_format), of);
								fwrite("')", 1, 2, of);
							}
							else
							{
								fwrite("'", 1, 1, of);
								fwrite(buffer, 1, strlen(buffer), of);
								fwrite("'", 1, 1, of);
							}
						}
						else
						{
							fwrite("NULL", 1, 4, of);
						}
						break;
				}
			}
			else
			{
				switch (odbc_cols[x].datatype)
				{
					case SQL_DECIMAL:
					case SQL_NUMERIC:
					case SQL_SMALLINT:
					case SQL_INTEGER:
					case SQL_REAL:
					case SQL_FLOAT:
					case SQL_DOUBLE:
					case SQL_BIT:
					case SQL_TINYINT:
					case SQL_BIGINT:
					case SQL_GUID:
						if (buffer[0] != '\0')
							fwrite(buffer, 1, strlen(buffer), of);
						else
							fwrite("NULL", 1, 4, of);
						break;
					case SQL_CHAR:
					case SQL_VARCHAR:
					case SQL_WCHAR:
					case SQL_WVARCHAR:
					case SQL_BINARY:
					case SQL_VARBINARY:
					case SQL_LONGVARCHAR:
					case SQL_WLONGVARCHAR:
					case SQL_LONGVARBINARY:
					case SQL_TYPE_DATE:
					case SQL_TYPE_TIME:
					case SQL_TYPE_TIMESTAMP:
					case SQL_INTERVAL_MONTH:
					case SQL_INTERVAL_YEAR:
					case SQL_INTERVAL_YEAR_TO_MONTH:
					case SQL_INTERVAL_DAY:
					case SQL_INTERVAL_HOUR:
					case SQL_INTERVAL_MINUTE:
					case SQL_INTERVAL_SECOND:
					case SQL_INTERVAL_DAY_TO_HOUR:
					case SQL_INTERVAL_DAY_TO_MINUTE:
					case SQL_INTERVAL_DAY_TO_SECOND:
					case SQL_INTERVAL_HOUR_TO_MINUTE:
					case SQL_INTERVAL_HOUR_TO_SECOND:
					case SQL_INTERVAL_MINUTE_TO_SECOND:
						if (buffer[0] != '\0')
						{
							fwrite("'", 1, 1, of);
							ReplaceString(buffer, "'", "''");
							fwrite(buffer, 1, strlen(buffer), of);
							fwrite("'", 1, 1, of);
						}
						else
						{
							fwrite("NULL", 1, 4, of);
						}
						break;
				}
			}
		}
		if (do_it)
		{
			fwrite(")\n/\n", 1, 4, of);
		}
	}
	fclose(of);
	free(buffer);
	free(outbuf);
	SetCursor(c);
}

void ExportXMLData(TextWidget *text, char *filename)
{
	int x, do_it = 1, bufsize = 0, row = 0, recordlength = 0;
	char *buffer;
	wchar_t *bufferutf;
	FILE *of;
	size_t bytes_read;
	HCURSOR c;
	char col_name[512], *outbuf;

	c = SetCursor(LoadCursor(NULL, IDC_WAIT));
	of = fopen(filename, "w+, ccs=UTF-8");
	if (of == NULL)
	{
		Error(NULL, "%s", strerror(errno));
		SetCursor(c);
		return;
	}
	outbuf = malloc(1024*1024); // 1MB
	if (!outbuf)
	{
		WPrintf("Out of memory\n");
		SetCursor(c);
		return;
	}
	setvbuf(of, outbuf, _IOFBF, 1024 * 1024);

	fseek(tmpfile_read, 0, SEEK_SET);
	if (text->datasource == DS_ORACLE)
	{
		for (x = 0; x < ncols; x++)
		{
			recordlength += oci_cols[x].col_size;
			if (oci_cols[x].col_size > bufsize)
				bufsize = oci_cols[x].col_size;
		}
	}
	else
	{
		for (x = 0; x < ncols; x++)
		{
			recordlength += (int)odbc_cols[x].col_size;
			if ((int)odbc_cols[x].col_size > bufsize)
				bufsize = (int)odbc_cols[x].col_size;
		}
	}
	if (bufsize)
	{
		buffer = (char*)malloc(bufsize * 2 + 1);
		bufferutf = (wchar_t*)malloc(bufsize * 2 + 1);
	}
	else
	{
		SetCursor(c);
		free(outbuf);
		return;
	}
	if (buffer == NULL)
	{
		SetCursor(c);
		Error(NULL, "Out of memory");
		free(outbuf);
		return;
	}
	fwprintf(of, L"<?xml version=\"1.0\" encoding=\"UTF-8\"?><ROWS>");
	while (do_it)
	{
		row++;
		for (x = 0; x < ncols; x++)
		{
			if (text->datasource == DS_ORACLE)
				bytes_read = fread(buffer, 1, oci_cols[x].col_size, tmpfile_read);
			else
				bytes_read = fread(buffer, 1, odbc_cols[x].col_size, tmpfile_read);
			if (bytes_read == 0)
			{
				do_it = 0;
				break;
			}
			buffer[bytes_read] = '\0';
			ReplaceString(buffer, "<", "&lt;");
			ReplaceString(buffer, ">", "&gt;");
			ReplaceString(buffer, "&", "&amp;");
			if (x == 0)
			{
				fwprintf(of, L"<ROW num=\"%d\">", row);
			}
			if (text->datasource == DS_ORACLE)
				strcpy(col_name, oci_cols[x].col_name);
			else
				strcpy(col_name, (char*)odbc_cols[x].col_name);
			ReplaceCharInString(col_name, ' ', '_');
			fwrite(L"<", sizeof(wchar_t), 1, of);
			MultiByteToWideChar(CP_UTF8, 0, col_name, -1, bufferutf, bufsize * 2 + 1);
			fwrite(bufferutf, sizeof(wchar_t), wcslen(bufferutf), of);
			fwrite(L">", sizeof(wchar_t), 1, of);
			MultiByteToWideChar(CP_UTF8, 0, buffer, -1, bufferutf, bufsize * 2 + 1);
			fwrite(bufferutf, sizeof(wchar_t), wcslen(bufferutf), of);
			fwrite(L"</", sizeof(wchar_t), 2, of);
			MultiByteToWideChar(CP_UTF8, 0, col_name, -1, bufferutf, bufsize * 2 + 1);
			fwrite(bufferutf, sizeof(wchar_t), wcslen(bufferutf), of);
			fwrite(L">", sizeof(wchar_t), 1, of);
		}
		if (do_it)
		{
			fwrite(L"</ROW>", sizeof(wchar_t), 6, of);
		}
	}
	fwrite(L"</ROWS>", sizeof(wchar_t), 7, of);
	fclose(of);
	free(buffer);
	free(outbuf);
	SetCursor(c);
}

void ExportData(TextWidget *text)
{
	OPENFILENAME openFile;
	char filename[MAX_PATH] = "\0";
	char filter[200] = "Comma Separated Value Files (*.csv, *.txt))\0*.CSV;*.TXT\0XML Files (*.xml)\0*.xml\0SQL Scripts (*.sql)\0*.sql\0\0\0";
	char extentions[][5] = { ".csv",".xml",".sql" };
	char *pos1, *pos2;

	if (tmpfile_write == NULL)
	{
		Error(NULL, "No data to export");
		return;
	}

	openFile.lStructSize = sizeof(OPENFILENAME);
	openFile.hwndOwner = g_hWnd;
	openFile.hInstance = NULL;
	openFile.lpstrFilter = filter;
	openFile.lpstrCustomFilter = NULL;
	openFile.nMaxCustFilter = 255;
	openFile.nFilterIndex = 0;
	openFile.lpstrFile = filename;
	openFile.nMaxFile = MAX_PATH;
	openFile.lpstrFileTitle = NULL;
	openFile.nMaxFileTitle = 0;
	openFile.lpstrInitialDir = NULL;
	openFile.lpstrTitle = NULL;
	openFile.Flags = OFN_OVERWRITEPROMPT;
	openFile.nFileOffset = 0;
	openFile.nFileExtension = 0;
	openFile.lpstrDefExt = NULL;
	openFile.lCustData = 0;
	openFile.lpfnHook = NULL;
	openFile.lpTemplateName = NULL;
	GetSaveFileName(&openFile);
	if (filename[0] == '\0')
		return;
	pos1 = strrchr(filename, '.');
	pos2 = strrchr(filename, '\\');
	if (pos1 == NULL)
		strcat(filename, extentions[openFile.nFilterIndex - 1]);
	else
	{
		if (pos2 != NULL && pos2 > pos1)
			strcat(filename, extentions[openFile.nFilterIndex - 1]);
	}
	switch (openFile.nFilterIndex)
	{
		case 1:
			ExportCSVData(text, filename);
			break;
		case 2:
			ExportXMLData(text, filename);
			break;
		case 3:
			ExportSQLData(text, filename);
			break;
	}
}

void CopyFromListView(HWND hWnd, CopySpec copymethod)
{
	int x, y, nlines = 0, nselected = 0, found = 0, recordsize = 0, fieldsize = 0, offset = 0;
	size_t bytes_read;
	char *buffer, *field, *newclipstr, *lockedclipstr;
	HCURSOR c;

	nlines = ListView_GetItemCount(hWnd);
	nselected = ListView_GetSelectedCount(hWnd);
	if (!nlines || !nselected || ncols==0)
		return;
	if (g_text->datasource == DS_ORACLE)
	{
		for (x = 0; x < ncols; x++)
		{
			recordsize += oci_cols[x].col_size;
			if (oci_cols[x].col_size > fieldsize)
				fieldsize = oci_cols[x].col_size;
			if ((copymethod != COPY_ENTIRELINE && copymethod != COPY_ASCOMMA && copymethod != COPY_ASCOLON && copymethod != COPY_FIRSTCOLUMN) && x < g_subitem_clicked)
				offset += oci_cols[x].col_size;
		}
	}
	else
	{
		for (x = 0; x < ncols; x++)
		{
			recordsize += (int)odbc_cols[x].col_size;
			if ((int)odbc_cols[x].col_size > fieldsize)
				fieldsize = (int)odbc_cols[x].col_size;
			if ((copymethod != COPY_ENTIRELINE && copymethod != COPY_ASCOMMA && copymethod != COPY_ASCOLON && copymethod != COPY_FIRSTCOLUMN) && x < g_subitem_clicked)
				offset += (int)odbc_cols[x].col_size;
		}
	}
	buffer = (char*)malloc((recordsize + ncols + 3 * ncols + 20)*nselected);
	if (buffer == NULL)
	{
		Error(NULL, "Out of memory");
		return;
	}
	field = (char*)malloc(fieldsize + 1);
	if (field == NULL)
	{
		Error(NULL, "Out of memory");
		return;
	}
	c = SetCursor(LoadCursor(NULL, IDC_WAIT));
	if (copymethod == COPY_QUOTEDIN || copymethod == COPY_PASTEASQUOTEDIN || copymethod == COPY_UNQUOTEDIN || copymethod == COPY_PASTEASUNQUOTEDIN)
		strcpy(buffer, "(");
	else
		buffer[0] = '\0';
	for (x = 0; x < nlines && found < nselected; x++)
	{
		if (ListView_GetItemState(hWnd, x, LVIS_SELECTED)&LVIS_SELECTED)
		{
			if (found)
				strcat(buffer, "\n");
			if (copymethod == COPY_ENTIRELINE || copymethod == COPY_ASCOMMA || copymethod == COPY_ASCOLON)
			{
				fseek(tmpfile_read, recordlength*x, SEEK_SET);
				for (y = 0; y < ncols; y++)
				{
					if (g_text->datasource == DS_ORACLE)
						bytes_read = fread(field, 1, oci_cols[y].col_size, tmpfile_read);
					else
						bytes_read = fread(field, 1, odbc_cols[y].col_size, tmpfile_read);
					field[bytes_read] = '\0';
					if (copymethod == COPY_ENTIRELINE)
					{
						if (y)
							strcat(buffer, " ");
						strcat(buffer, field);
					}
					else if (copymethod == COPY_ASCOMMA)
					{
						if (y)
							strcat(buffer, ",");
						strcat(buffer, Trim(field));
					}
					else if (copymethod == COPY_ASCOLON)
					{
						if (y)
							strcat(buffer, ";");
						strcat(buffer, Trim(field));
					}
				}
			}
			else if (copymethod == COPY_FIRSTCOLUMN)
			{
				fseek(tmpfile_read, recordlength*x, SEEK_SET);
				if (g_text->datasource == DS_ORACLE)
					bytes_read = fread(field, 1, oci_cols[0].col_size, tmpfile_read);
				else
					bytes_read = fread(field, 1, odbc_cols[0].col_size, tmpfile_read);
				field[bytes_read] = '\0';
				strcat(buffer, field);
			}
			else // This colum
			{
				fseek(tmpfile_read, recordlength*x + offset, SEEK_SET);
				if (g_text->datasource == DS_ORACLE)
					bytes_read = fread(field, 1, oci_cols[g_subitem_clicked].col_size, tmpfile_read);
				else
					bytes_read = fread(field, 1, odbc_cols[g_subitem_clicked].col_size, tmpfile_read);
				field[bytes_read] = '\0';
				if (copymethod == COPY_QUOTEDIN || copymethod == COPY_PASTEASQUOTEDIN)
					strcat(buffer, "'");
				strcat(buffer, Trim(field));
				if (found < nselected - 1)
				{
					if (copymethod == COPY_QUOTEDIN || copymethod == COPY_PASTEASQUOTEDIN)
						strcat(buffer, "',");
					else if (copymethod == COPY_UNQUOTEDIN || copymethod == COPY_PASTEASUNQUOTEDIN)
						strcat(buffer, ",");
				}
				else
				{
					if (copymethod == COPY_QUOTEDIN || copymethod == COPY_PASTEASQUOTEDIN)
						strcat(buffer, "'");
				}
			}
			found++;
		}
	}
	if (copymethod == COPY_QUOTEDIN || copymethod == COPY_PASTEASQUOTEDIN || copymethod == COPY_UNQUOTEDIN || copymethod == COPY_PASTEASUNQUOTEDIN)
		strcat(buffer, ")");
	if (OpenClipboard(hWnd))
	{
		EmptyClipboard();
		newclipstr = GlobalAlloc(GMEM_DDESHARE, strlen(buffer) + 1);
		if (newclipstr != NULL)
		{
			lockedclipstr = GlobalLock(newclipstr);
			if (lockedclipstr != NULL)
			{
				strcpy(lockedclipstr, buffer);
				GlobalUnlock(newclipstr);
				if (SetClipboardData(CF_TEXT, newclipstr) == NULL)
					DisplayLastError(NULL);
			}
		}
		CloseClipboard();
	}
	free(buffer);
	if (copymethod == COPY_PASTETHISCOLUMN || copymethod == COPY_PASTEASQUOTEDIN || copymethod == COPY_PASTEASUNQUOTEDIN)
	{
		PasteFromClipboard(g_text);
		SetActiveWindow(hwndTextBox);
		SetFocus(hwndTextBox);
	}
	SetCursor(c);
}

void SelectTableData(TextWidget *text, TableItem *item)
{

	int x, y, z, nlines = 0, nselected = 0, found = 0, recordsize = 0, fieldsize = 0, offset = 0;
	size_t bytes_read;
	char *buffer, date_format[100];
	HCURSOR c;

	nlines = ListView_GetItemCount(hwndListView);
	nselected = ListView_GetSelectedCount(hwndListView);
	if (!nlines || !nselected)
		return;

	for (x = 0; x < item->ncolumns; x++)
	{
		if (oci_cols[item->columns[x]].col_size > recordsize)
			recordsize = oci_cols[item->columns[x]].col_size;
	}

	buffer = (char*)malloc(recordsize * 2 + 1);
	if (buffer == NULL)
	{
		Error(NULL, "Out of memory");
		return;
	}

	c = SetCursor(LoadCursor(NULL, IDC_WAIT));

	if (GetSingleValue(g_oci,"select sys_context('userenv','nls_date_format') from dual", date_format, 100) != 0)
		date_format[0] = '\0';

	WPrintf("\r");
	ClearCommand(text, 0, 0);

	StartFastDraw(text, 0);
	InsertLineBuffer(text, "%s", "SELECT * FROM \"");
	InsertLineBuffer(text, "%s", item->owner);
	InsertLineBuffer(text, "%s", "\".\"");
	InsertLineBuffer(text, "%s", item->table);
	InsertLineBuffer(text, "%s", "\" WHERE ");
	for (x = 0; x < nlines && found < nselected; x++)
	{
		if (ListView_GetItemState(hwndListView, x, LVIS_SELECTED)&LVIS_SELECTED)
		{
			if (found)
				InsertLineBuffer(text, "%s", " OR ");
			InsertLineBuffer(text, "%s", "(");
			for (y = 0; y < item->ncolumns; y++)
			{
				if (y)
					InsertLineBuffer(text, "%s", " AND ");
				offset = 0;
				if (item->nrcolumns)
					InsertLineBuffer(text, "\"%s\"", item->rcolumns[y]);
				else
					InsertLineBuffer(text, "\"%s\"", item->columns[y]);
				for (z = 0; z < item->columns[y]; z++)
					offset += oci_cols[z].col_size;
				fseek(tmpfile_read, recordlength*x + offset, SEEK_SET);
				bytes_read = fread(buffer, 1, oci_cols[item->columns[y]].col_size, tmpfile_read);
				buffer[bytes_read] = '\0';
				switch (oci_cols[item->columns[y]].datatype)
				{
					case OCI_TYPECODE_REAL:
					case OCI_TYPECODE_DOUBLE:
					case OCI_TYPECODE_FLOAT:
					case OCI_TYPECODE_NUMBER:
					case OCI_TYPECODE_DECIMAL:
					case OCI_TYPECODE_OCTET:
					case OCI_TYPECODE_INTEGER:
					case OCI_TYPECODE_SMALLINT:
						if (buffer[0] != '\0')
						{
							InsertLineBuffer(text, "%s", "=");
							InsertLineBuffer(text, "%s", buffer);
						}
						else
						{
							InsertLineBuffer(text, "%s", " IS NULL");
						}
						break;
					case OCI_TYPECODE_RAW:
					case OCI_TYPECODE_VARCHAR2:
					case OCI_TYPECODE_VARCHAR:
					case OCI_TYPECODE_CHAR:
					case OCI_TYPECODE_CLOB:
					case OCI_TYPECODE_BLOB:
					case OCI_TYPECODE_BFILE:
					case SQLT_LNG:
					case SQLT_BIN:
					case SQLT_LBI:
					case SQLT_RID:
					case SQLT_RDD:
						if (buffer[0] != '\0')
						{
							ReplaceString(buffer, "'", "''");
							InsertLineBuffer(text, "%s", "='");
							InsertLineBuffer(text, "%s", buffer);
							InsertLineBuffer(text, "%s", "'");
						}
						else
						{
							InsertLineBuffer(text, "%s", " IS NULL");
						}
						break;
					case OCI_TYPECODE_DATE:
						if (buffer[0] != '\0')
						{
							if (date_format[0])
							{
								InsertLineBuffer(text, "%s", "=TO_DATE('");
								InsertLineBuffer(text, "%s", buffer);
								InsertLineBuffer(text, "%s", "','");
								InsertLineBuffer(text, "%s", date_format);
								InsertLineBuffer(text, "%s", "')");
							}
							else
							{
								InsertLineBuffer(text, "%s", "'");
								InsertLineBuffer(text, "%s", buffer);
								InsertLineBuffer(text, "%s", "'");
							}
						}
						else
						{
							InsertLineBuffer(text, "%s", "IS NULL");
						}
						break;
				}
			}
			InsertLineBuffer(text, "%s", ")");
			found++;
		}
	}
	SetCursor(c);
	SetActiveWindow(hwndTextBox);
	SetFocus(hwndTextBox);
	InsertLineBuffer(text, "%s", ";");
	EndFastDraw(text);
	UpdateCaretPosition(text, 1);
	InsertLineBuffer(g_text, "\n");

}

