#include "jbsql.h"
#include <Dbghelp.h>
#include <Psapi.h>
#include <signal.h>
#include "ccsplitter.h"

HINSTANCE hInst, g_hInst;
HWND      g_hWnd;
TCHAR     g_szClassName[] = TEXT("VListVwClass");
HWND hwndListView;
HWND hwndTextBox;
HWND hwndSplitter;
HMENU g_menu, g_menu_tables;
TextWidget* g_text = NULL;
char executable[MAX_PATH];
HashTable vht;
DWORD g_tid;
HANDLE pthandle;
CRITICAL_SECTION g_mutex;
char* g_argv0;
char g_title[200] = "JBSQL";


#define SCREENBUFSIZE 1000

char* myvsprintf(const char* fmt, va_list vl)
{
    char* str = NULL;
    size_t size = 0;
    size_t new_size;
    char* new_str;

    size = SCREENBUFSIZE;
    str = (char*)malloc(size);
    for (;;)
    {
        new_size = vsnprintf(str, size - 1, fmt, vl);
        if (new_size > size - 1)
        {
            new_size++;
            if ((new_str = (char*)realloc(str, new_size)) == NULL)
            {
                MessageBox(NULL, "myvsprintf(): Out of memory", NULL, MB_OK | MB_ICONERROR);
                break;
            }
            str = new_str;
            size = new_size;
        }
        else
            break;
    }
    return str;
}

void Log(char* fmt, ...)
{
    va_list param;
    char* str;
    FILE* f;

    if (g_text && g_text->options[DEBUG].value.int_value)
    {
        va_start(param, fmt);
        str = myvsprintf(fmt, param);
        va_end(param);
        f = fopen("jbsql.log", "a+");
        if (f == NULL)
        {
            free(str);
            return;
        }
        if (str[strlen(str) - 1] == '\n')
            fprintf(f, "%s", str);
        else
            fprintf(f, "%s\n", str);
        fclose(f);
        free(str);
    }
}

void Error(HWND hWnd, char* fmt, ...)
{
    va_list param;
    char* str;

    va_start(param, fmt);
    str = myvsprintf(fmt, param);
    va_end(param);
    MessageBox(hWnd, str, NULL, MB_OK | MB_ICONERROR);
    free(str);
}

void DisplayLastErrorInternal(HWND hWnd, int line, char* file)
{
    LPVOID lpMsgBuf;

    if (!FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_FROM_HMODULE,
        GetModuleHandle("wininet.dll"),
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR)& lpMsgBuf, 0, NULL))
    {
        DisplayLastError(hWnd);
        return;
    }
    Error(hWnd, "%s (file %s, line %d)", lpMsgBuf, file, line);
    LocalFree(lpMsgBuf);
}

void SetWindowTitle(char* username, char* database, int instance, int sid)
{
    char sidstr[30] = "";

    if (sid)
    {
        sprintf(sidstr, " (%d, %d)", instance, sid);
    }
    if (username[0] != '\0')
    {
        if (database[0] != '\0')
            sprintf(g_title, "%s@%s%s", username, database, sidstr);
        else
            sprintf(g_title, "%s@local%s", username, sidstr);
    }
    else
    {
        if (odbc_connected && database[0] != '\0')
            sprintf(g_title, "%s", database);
        else
            sprintf(g_title, "Disconnected");
    }
    SetWindowTitleEx(g_hWnd, "JB/SQL");
    if (g_text->analyze)
        SetWindowTitleEx(g_text->analyze, "SQL Analyzer");
    if (g_text->schema)
        SetWindowTitleEx(g_text->schema, "Schema Browser");
    if (g_text->sqlmonitor)
        SetWindowTitleEx(g_text->sqlmonitor, "SQL Monitor");
}

LRESULT CALLBACK ChangePasswordProc(HWND hWnd,
    UINT uMessage,
    WPARAM wParam,
    LPARAM lParam)
{
    static TextWidget* text;

    switch (uMessage)
    {
        case WM_COMMAND:
            switch (GET_WM_COMMAND_ID(wParam, lParam))
            {
                case ID_LOGON_OK:
                    {
                        char username[129], password[100], new_password[100], new_password2[100], database[100], connect_as[10];

                        GetDlgItemText(GetParent(hWnd), IDC_USERNAME, username, 129);
                        GetDlgItemText(GetParent(hWnd), IDC_PASSWORD, password, 100);
                        GetDlgItemText(GetParent(hWnd), IDC_DATABASE, database, 100);
                        GetDlgItemText(GetParent(hWnd), IDC_CONNECT_AS, connect_as, 10);
                        GetDlgItemText(hWnd, IDC_PASSWORD, new_password, 100);
                        GetDlgItemText(hWnd, IDC_PASSWORD2, new_password2, 100);
                        if (new_password[0] == '\0' || new_password2[0] == '\0')
                        {
                            MessageBox(hWnd, "Password is empty", "Error", MB_OK | MB_ICONERROR);
                        }
                        else
                        {
                            if (strcmp(new_password, new_password2))
                            {
                                MessageBox(hWnd, "Passwords do not match", "Error", MB_OK | MB_ICONERROR);
                            }
                            else
                            {
                                if (ChangePassword(username, password, new_password, database, connect_as, 0) == 0)
                                    EndDialog(hWnd, TRUE);
                            }
                        }
                        return TRUE;
                    }
                case IDCANCEL:
                case ID_LOGON_CANCEL:
                    EndDialog(hWnd, FALSE);
                    return TRUE;
            }
            break;

        case WM_CLOSE:
            EndDialog(hWnd, FALSE);
            return TRUE;
        default:
            break;
    }
    return FALSE;
}



int ShowChangePasswordBox(HWND parenthWnd, TextWidget* text)
{
    FARPROC cpproc;
    LRESULT res;

    cpproc = MakeProcInstance((FARPROC)ChangePasswordProc, hInst);
    res = DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_LOGON_CHANGE_PASS), parenthWnd, (DLGPROC)cpproc, (LPARAM)text);
    FreeProcInstance(cpproc);

    return !(int)res;
}


LRESULT CALLBACK LoginProc(HWND hWnd,
    UINT uMessage,
    WPARAM wParam,
    LPARAM lParam)
{
    static TextWidget* text;

    switch (uMessage)
    {
        case WM_INITDIALOG:
            {
                int x;

                text = (TextWidget*)lParam;
                for (x = 0; x < 10; x++)
                {
                    if (last_usernames[x][0])
                    {
                        SendDlgItemMessage(hWnd, IDC_USERNAME, CB_ADDSTRING, 0, (LPARAM)last_usernames[x]);
                        if (x == 0)
                            SetDlgItemText(hWnd, IDC_USERNAME, last_usernames[x]);
                    }
                }
                for (x = 0; x < 10; x++)
                {
                    if (last_databases[x][0])
                    {
                        SendDlgItemMessage(hWnd, IDC_DATABASE, CB_ADDSTRING, 0, (LPARAM)last_databases[x]);
                        if (x == 0)
                            SetDlgItemText(hWnd, IDC_DATABASE, last_databases[x]);
                    }
                }
                SetFocus(hWnd);
            }
            SendDlgItemMessage(hWnd, IDC_CONNECT_AS, CB_ADDSTRING, 0, (LPARAM)"NORMAL");
            SendDlgItemMessage(hWnd, IDC_CONNECT_AS, CB_ADDSTRING, 0, (LPARAM)"SYSOPER");
            SendDlgItemMessage(hWnd, IDC_CONNECT_AS, CB_ADDSTRING, 0, (LPARAM)"SYSDBA");
            SendDlgItemMessage(hWnd, IDC_CONNECT_AS, CB_SETCURSEL, 0, 0);
            SendDlgItemMessage(hWnd, IDC_DATASOURCE, CB_ADDSTRING, 0, (LPARAM)"ORACLE");
            SendDlgItemMessage(hWnd, IDC_DATASOURCE, CB_ADDSTRING, 0, (LPARAM)"ODBC");
            SendDlgItemMessage(hWnd, IDC_DATASOURCE, CB_SETCURSEL, text->datasource, 0);
            break;
        case WM_COMMAND:
            switch (GET_WM_COMMAND_ID(wParam, lParam))
            {
                case ID_LOGON_OK:
                    {
                        char username[100], password[100], database[512], connect_as[10], datasource[10];

                        GetDlgItemText(hWnd, IDC_USERNAME, username, 100);
                        GetDlgItemText(hWnd, IDC_PASSWORD, password, 100);
                        GetDlgItemText(hWnd, IDC_DATABASE, database, 512);
                        GetDlgItemText(hWnd, IDC_CONNECT_AS, connect_as, 10);
                        GetDlgItemText(hWnd, IDC_DATASOURCE, datasource, 10);
                        if (!strcmp(datasource, "ORACLE"))
                        {
                            g_text->datasource = DS_ORACLE;
                            switch (ConnectToOracle(username, password, database, connect_as, 0))
                            {
                                case 0:
                                    EndDialog(hWnd, TRUE);
                                    break;
                                case 28001:
                                    if (!ShowChangePasswordBox(hWnd, g_text))
                                        EndDialog(hWnd, TRUE);
                                    else
                                        OCI_DisconnectServer();
                                    break;
                            }
                        }
                        else
                        {
                            g_text->datasource = DS_ODBC;
                            if (ConnectToODBC(username, password, database, 0) == 0)
                                EndDialog(hWnd, TRUE);
                        }
                        return TRUE;
                    }
                case IDCANCEL:
                case ID_LOGON_CANCEL:
                    EndDialog(hWnd, FALSE);
                    return TRUE;
            }
            break;

        case WM_CLOSE:
            EndDialog(hWnd, FALSE);
            return TRUE;
        default:
            break;
    }
    return FALSE;
}

int ShowLoginBox(HWND parenthWnd, TextWidget* text)
{
    FARPROC aboutproc;
    LRESULT res;

    aboutproc = MakeProcInstance((FARPROC)LoginProc, hInst);
    res = DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_LOGON), parenthWnd, (DLGPROC)aboutproc, (LPARAM)text);
    FreeProcInstance(aboutproc);
    if (!res)
        return 0;
    return 1;
}

void NewWindow(TextWidget* text)
{
    char path[1024] = "", original_jbsql[MAX_PATH];
    HANDLE f = INVALID_HANDLE_VALUE;
    PROCESS_INFORMATION p;
    STARTUPINFO si;

    strcpy(original_jbsql, executable);
    f = CreateFile(original_jbsql,
        0,
        FILE_SHARE_READ,
        NULL,
        OPEN_EXISTING,
        0,
        NULL);
    if (f == INVALID_HANDLE_VALUE)
    {
        Error(NULL, "Can't find %s!", original_jbsql);
        return;
    }
    CloseHandle(f);
    SaveCommandHistory(text);
    strcat(original_jbsql, " \"");
    strcat(original_jbsql, connectstring);
    strcat(original_jbsql, "\"");
    si.cb = sizeof(STARTUPINFO);
    si.lpReserved = NULL;
    si.lpDesktop = NULL;
    si.lpTitle = NULL;
    si.dwFlags = 0;
    si.cbReserved2 = 0;
    si.lpReserved2 = NULL;
    if (CreateProcess(NULL, original_jbsql, NULL, NULL, FALSE, 0, NULL, NULL, &si, &p) == 0)
    {
        DisplayLastError(NULL);
        return;
    }
}

int IsChildWindow(HWND parent, HWND child)
{
    while (child)
    {
        if (child == parent)
            return 1;
        child = GetParent(child);
    }
    return 0;
}

LRESULT CALLBACK MainWndProc(HWND hWnd,
    UINT uMessage,
    WPARAM wParam,
    LPARAM lParam)
{
    switch (uMessage)
    {
        case WM_CREATE:
            {
                CCSPLIT	spy;			// Y-split data
                RECT rc;

                GetClientRect(hWnd, &rc);
                hwndSplitter = CreateSplitter(g_hInst, hWnd);
                hwndTextBox = CreateTextWidget(g_hInst, hwndSplitter, &g_text);
                hwndListView = CreateListView(g_hInst, hwndSplitter);
                InitListView(hwndListView);
                spy.cbSize = sizeof(spy);
                spy.flags = CCSP_POS;
                spy.pos = rc.bottom / 2;
                CCsplitter_SetSplit(hwndSplitter, NULL, &spy);

                if (g_text->toolbar)
                    DestroyWindow(g_text->toolbar);
                g_text->toolbar = BuildToolbar(g_text, g_hInst, hWnd);

                g_hWnd = hWnd;
                SetForegroundWindow(hwndTextBox);
                EnableWindow(hwndTextBox, TRUE);
                WPrintf("JB/SQL version %s (%s)\n", JBSQL_VERSION, (sizeof(size_t) == 8) ? "x64" : "x86");
                PrintOCIVersion();
                GetXMLFunctions();
            }
            break;

        case WM_ACTIVATE:
            SetFocus(hwndTextBox);
            break;

        case WM_SIZE:
            {
                TextWidget* text;

                text = (TextWidget*)GetWindowLongPtr(hwndTextBox, 0);
                SendMessage(text->toolbar, TB_AUTOSIZE, 0, 0);
                //ResizeSplitter(hwndSplitter, hWnd);
                RECT r;

                GetWindowRect(text->toolbar, &r);
                MoveWindow(hwndSplitter, 0, r.bottom - r.top, LOWORD(lParam), HIWORD(lParam) - (r.bottom - r.top), TRUE);
                //InvalidateRect(hwndListView, NULL, TRUE);
                //UpdateWindow(hwndListView);
                break;
            }
        case WM_COMMAND:
            {
                TextWidget* text;

                text = (TextWidget*)GetWindowLongPtr(hwndTextBox, 0);
                switch (LOWORD(wParam))
                {
                    case ID_FILE_EXIT:
                        if (tid == 0)
                            DestroyWindow(hWnd);
                        break;
                    case ID_FILE_CONNECT:
                        if (tid == 0)
                            ShowLoginBox(g_hWnd, text);
                        break;
                    case ID_FILE_RECONNECT:
                        if (tid == 0 && connectstring[0])
                        {
                            ConnectWithConnectString(connectstring, 0);
                        }
                        break;
                    case ID_NEW_WINDOW:
                        NewWindow(text);
                        break;
                    case ID_FILE_DISCONNECT:
                        if (tid == 0)
                            DisconnectFromOracle();
                        break;
                    case ID_QUERY_CANCEL:
                        if (text->datasource == DS_ORACLE)
                            OCI_AbortQuery();
                        else
                            ODBC_AbortQuery();
                        break;
                    case ID_STATEMENT_PREVIOUS:
                        if (tid == 0)
                            PreviousStatement(text);
                        break;
                    case ID_STATEMENT_NEXT:
                        if (tid == 0)
                            NextStatement(text);
                        break;
                    case ID_COMMIT:
                        if (tid == 0)
                        {
                            WPrintf("\r");
                            ClearCommand(text, 0, 0);
                            InsertLineBuffer(text, "commit;\n");
                        }
                        break;
                    case ID_ROLLBACK:
                        if (tid == 0)
                        {
                            WPrintf("\r");
                            ClearCommand(text, 0, 0);
                            InsertLineBuffer(text, "rollback;\n");
                        }
                        break;
                    case ID_REPEAT:
                        if (tid == 0)
                            InsertLineBuffer(text, "/\n");
                        break;
                    case ID_OPTIONS:
                        {
                            FARPROC dlgproc;

                            dlgproc = MakeProcInstance((FARPROC)OptionProc, g_hInst);
                            if (DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_OPTIONS), hWnd, (DLGPROC)dlgproc, (LPARAM)text) == -1)
                                DisplayLastError(hWnd);
                            FreeProcInstance(dlgproc);
                        }
                        break;
                    case ID_SQL_ANALYZER:
                        if (ora_connected)
                        {
                            if (text->analyze == NULL)
                                text->analyze = ShowSqlAnalyzer(NULL);
                        }
                        break;
                    case ID_SCHEMA_BROWSER:
                        if (ora_connected)
                        {
                            if (text->schema == NULL)
                                text->schema = ShowSchemaBrowser(NULL);
                        }
                        break;
                    case ID_EXPORT:
                        ExportData(text);
                        break;
                    case ID_SELECTALL:
                        if (GetFocus() == hwndListView)
                        {
                            ListView_SetItemState(hwndListView,                 // handle to listview
                                -1,                           // all items
                                LVIS_FOCUSED | LVIS_SELECTED, // item state
                                0x000F);                    // mask
                        }
                        break;
                    case ID_PARENT_RECORDS:
                        if (!tid)
                        {
                            if (GetFocus() == hwndTextBox)
                            {
                                if (ListView_GetSelectedCount(hwndListView) == 0)
                                {
                                    // Select all record
                                    ListView_SetItemState(hwndListView, -1, LVIS_FOCUSED | LVIS_SELECTED, 0x000F);
                                }
                            }
                            text->table_children = 0;
                            ShowParentChildTables(text, hWnd);
                        }
                        break;
                    case ID_CHILD_RECORDS:
                        if (!tid)
                        {
                            if (GetFocus() == hwndTextBox)
                            {
                                if (ListView_GetSelectedCount(hwndListView) == 0)
                                {
                                    // Select all record
                                    ListView_SetItemState(hwndListView, -1, LVIS_FOCUSED | LVIS_SELECTED, 0x000F);
                                }
                            }
                            text->table_children = 1;
                            ShowParentChildTables(text, hWnd);
                        }
                        break;
                    case ID_COPY:
                        if (GetFocus() == hwndTextBox)
                        {
                            if (text->selection)
                                CopyToClipboard(text);
                        }
                        else if (GetFocus() == hwndListView)
                            CopyFromListView(hwndListView, COPY_ENTIRELINE);
                        break;
                    case ID_PASTE:
                        if (ptid == 0)
                        {
                            pthandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)PasteFromClipboard, text, 0, &ptid);
                            SetTimer(hWnd, PASTE_TIMER, PASTE_TIMEOUT, NULL);
                        }
                        break;
                    case ID_FAV_ADD:
                        AddFavourite(text, hWnd);
                        break;
                    case ID_FAV_EDIT:
                        ShowFavourites(hWnd, text);
                        break;
                    case ID_HISTORY:
                        ShowHistory(text);
                        break;
                    case ID_HELP_HELP:
                        ShellExecute(NULL, "open", "http://users.telenet.be/jbosman/jbsql", NULL, NULL, SW_SHOWNORMAL);
                        break;
                    case ID_HELP_CHECKFORUPDATE:
                        CheckNewVersion(hWnd, 1);
                        break;
                    case ID_HELP_ABOUT:
                        {
                            char str[400];

                            sprintf(str, "JBSQL version %s\nCreated by Jorgen Bosman", JBSQL_VERSION);
                            MessageBox(hWnd, str, "About", MB_ICONASTERISK);
                        }
                        break;
                    case IDC_HINTS:
                        SendMessage(hwndTextBox, WM_KEYDOWN, VK_TAB, 0);
                        break;
                    case ID_TOOLS_BINDVARIABLES:
                        ShowAllBindVariables(text, hWnd);
                        break;
                    case ID_TOOLS_SQLMONITOR:
                        if (text->sqlmonitor == NULL)
                            ShowSqlMonitor(NULL);
                        break;
                    case ID_COPY_COPYFIRSTCOLUMN:
                        CopyFromListView(hwndListView, COPY_FIRSTCOLUMN);
                        break;
                    case ID_COPY_COPYENTIRELINE:
                        CopyFromListView(hwndListView, COPY_ENTIRELINE);
                        break;
                    case ID_COPY_COPYTHISCOLUMN:
                        CopyFromListView(hwndListView, COPY_THISCOLUMN);
                        break;
                    default:
                        if (!IsFavouriteItem(text, LOWORD(wParam)))
                        {
                            if (GET_WM_COMMAND_ID(wParam, lParam) >= UM_TOOLTIP)
                            {
                                ToggleOption(text, (int)GET_WM_COMMAND_ID(wParam, lParam) - UM_TOOLTIP);
                            }
                        }
                        break;
                }
                break;
            }
        case WM_TIMER:
            switch (wParam)
            {
                case PASTE_TIMER:
                    {
                        DWORD exit_code;
                        GetExitCodeThread(pthandle, &exit_code);
                        if (exit_code != STILL_ACTIVE)
                        {
                            TextWidget* text;

                            KillTimer(hWnd, PASTE_TIMER);
                            text = (TextWidget*)GetWindowLongPtr(hwndTextBox, 0);
                            if (text->toolbar)
                                DestroyWindow(text->toolbar);
                            text->toolbar = BuildToolbar(text, g_hInst, g_hWnd);
                        }
                    }
                    break;
            }
            break;
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        case WM_NOTIFY:
            {
                NMHDR* nmhrd = (NMHDR*)lParam;

                if (nmhrd->hwndFrom == hwndListView)
                    return ListViewNotify(hwndListView, lParam);
                else
                {
                    switch (((LPNMHDR)lParam)->code)
                    {
                        case TTN_GETDISPINFO:
                            {
                                LPNMTTDISPINFO lpttt;
                                TextWidget* text;

                                text = (TextWidget*)GetWindowLongPtr(hwndTextBox, 0);

                                lpttt = (LPNMTTDISPINFO)lParam;
                                lpttt->hinst = NULL;
                                lpttt->uFlags = TTF_DI_SETITEM;

                                switch (lpttt->hdr.idFrom)
                                {
                                    case ID_FILE_CONNECT:
                                        //lpttt->lpszText = "Logon"; 
                                        strcpy(lpttt->szText, "Logon");
                                        break;
                                    case ID_NEW_WINDOW:
                                        //lpttt->lpszText = "New Window"; 
                                        strcpy(lpttt->szText, "New Window");
                                        break;
                                    case ID_COMMIT:
                                        strcpy(lpttt->szText, "Commit");
                                        break;
                                    case ID_ROLLBACK:
                                        strcpy(lpttt->szText, "Rollback");
                                        break;
                                    case ID_QUERY_CANCEL:
                                        strcpy(lpttt->szText, "Cancel Query");
                                        break;
                                    case ID_REPEAT:
                                        strcpy(lpttt->szText, "Repeat Statement");
                                        break;
                                    default:
                                        if (lpttt->hdr.idFrom >= UM_TOOLTIP)
                                        {
                                            strcpy(lpttt->szText, default_options[lpttt->hdr.idFrom - UM_TOOLTIP].tooltip);
                                        }
                                        break;
                                }
                                break;
                            }
                    }
                }
                break;
            }
        case WM_INITMENUPOPUP:
            if ((HMENU)wParam == GetSubMenu(GetMenu(hWnd), 4))
            {
                AddFavouritesMenu(g_text, hWnd);
            }
            break;
        default:
            return DefWindowProc(hWnd, uMessage, wParam, lParam);
    }
    return 0;
}


BOOL InitApplication(HINSTANCE hInstance)
{
    WNDCLASSEX  wcex;
    ATOM        aReturn;

    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style = 0;
    wcex.lpfnWndProc = (WNDPROC)DefWindowProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
    wcex.lpszMenuName = NULL;
    wcex.lpszClassName = GS_VIEW_WINDOW;
    wcex.hIcon = 0;
    wcex.hIconSm = 0;

    aReturn = RegisterClassEx(&wcex);

    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style = 0;
    wcex.lpfnWndProc = (WNDPROC)MainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wcex.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN_MENU);
    wcex.lpszClassName = g_szClassName;
    wcex.hIcon = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_JBSQL));
    wcex.hIconSm = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_JBSQL), IMAGE_ICON, 16, 16, 0);

    aReturn = RegisterClassEx(&wcex);

    if (0 == aReturn)
    {
        WNDCLASS wc;

        wc.style = 0;
        wc.lpfnWndProc = (WNDPROC)MainWndProc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = hInstance;
        wc.hIcon = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_JBSQL));
        wc.hCursor = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
        wc.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN_MENU);
        wc.lpszClassName = g_szClassName;

        aReturn = RegisterClass(&wc);
    }

    return aReturn;
}


BOOL InitInstance(HINSTANCE hInstance,
    int nCmdShow)
{
    HWND     hWnd;
    TCHAR    szTitle[MAX_PATH] = "JB/SQL";

    g_hInst = hInstance;

    hWnd = CreateWindowEx(0,
        g_szClassName,
        szTitle,
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        NULL,
        NULL,
        hInstance,
        NULL);


    if (!hWnd)
        return FALSE;

    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);
    g_hWnd = hWnd;
    return TRUE;

}

void CreateNewArgv(char*** argv, int* argc, char* commandline, int start_pos, int end_pos)
{
    *argv = (char**)realloc(*argv, (*argc + 1) * sizeof(char*));
    (*argv)[*argc] = malloc(end_pos - start_pos + 1);
    memcpy((*argv)[*argc], &commandline[start_pos], end_pos - start_pos);
    (*argv)[*argc][end_pos - start_pos] = 0;
    (*argc)++;
}

void ConvertCommandLine(char* commandline, int* argc, char*** argv)
{
    int lastpos = 0, x;
    char quote = 0;

    *argv = NULL;
    *argc = 0;
    for (x = 0; x <= (int)strlen(commandline); x++)
    {
        if (!quote && (commandline[x] == ' ' || commandline[x] == '\t' || commandline[x] == 0))
        {
            if (lastpos != x)
                CreateNewArgv(argv, argc, commandline, lastpos, x);
            lastpos = x + 1;
        }
        else if ((commandline[x] == '"' || commandline[x] == '\'') && (!quote || commandline[x] == quote))
        {
            if (!quote)
            {
                quote = commandline[x];
            }
            else
            {
                if (quote && (*argc > 1))	                                //Heinz: If statement added.
                                            //       Script can be executed during startup
                {							//	 of JBSQL with passing it as an argument.
                    CreateNewArgv(argv, argc, commandline, lastpos - 1, x + 1);
                }
                else
                {
                    CreateNewArgv(argv, argc, commandline, lastpos, x);
                }
                quote = 0;
            }
            if (quote == 0 || x == 0 || commandline[x - 1] == ' ' || commandline[x - 1] == '\t')
                lastpos = x + 1;
        }
    }
    *argv = (char**)realloc(*argv, (*argc + 1) * sizeof(char*));
    (*argv)[*argc] = NULL;
}

void RunInitialCommand(TextWidget* text, int argc, char* argv[])
{
    int x;

    text->termout = 0;
    ClearCommand(text, 0, 0);
    for (x = 2; x < argc; x++)
    {
        if (x > 2)
            InsertUnparsedLineBuffer(text, " ");
        InsertUnparsedLineBuffer(text, argv[x]);
    }
    InsertNewLineBuffer(text);
    text->termout = 1;
}

char* GetVersionInfo(char* program)
{
    LPVOID versionInfo;
    DWORD vLen = 0, langD;
    BOOL retVal;
    LPVOID retbuf = NULL;
    char fileEntry[256], programname[256];
    static char version[100] = "";

    if (strlen(program) > 4)
    {
        if (!_stricmp(&program[strlen(program) - 4], ".EXE"))
            strcpy(programname, program);
        else
            sprintf(programname, "%s.EXE", program);
    }
    else
        sprintf(programname, "%s.EXE", program);
    vLen = GetFileVersionInfoSize(programname, 0);
    if (vLen <= 0)
        return "0.0.0.0";
    versionInfo = malloc(vLen);
    GetFileVersionInfo(programname, 0, vLen, versionInfo);
    sprintf(fileEntry, "\\VarFileInfo\\Translation");
    retVal = VerQueryValue(versionInfo, fileEntry, &retbuf, (UINT*)& vLen);
    if (retVal && vLen == 4)
    {
        memcpy(&langD, retbuf, 4);
        sprintf(fileEntry, "\\StringFileInfo\\%02X%02X%02X%02X\\FileVersion", (langD & 0xff00) >> 8, langD & 0xff, (langD & 0xff000000) >> 24, (langD & 0xff0000) >> 16);
    }
    else
        sprintf(fileEntry, "\\StringFileInfo\\%04X04B0\\FileVersion", GetUserDefaultLangID());

    if (VerQueryValue(versionInfo, fileEntry, &retbuf, (UINT*)& vLen))
        strncpy(version, (char*)retbuf, 99);

    ReplaceString(version, ", ", ".");
    free(versionInfo);
    return version;
}

typedef BOOL(*MyMiniDumpWriteDumpFunc)(
    _In_  HANDLE hProcess,
    _In_  DWORD ProcessId,
    _In_  HANDLE hFile,
    _In_  MINIDUMP_TYPE DumpType,
    _In_  PMINIDUMP_EXCEPTION_INFORMATION ExceptionParam,
    _In_  PMINIDUMP_USER_STREAM_INFORMATION UserStreamParam,
    _In_  PMINIDUMP_CALLBACK_INFORMATION CallbackParam
    );

MyMiniDumpWriteDumpFunc MyMiniDumpWriteDump = NULL;

int GenerateDump(char* argv0, EXCEPTION_POINTERS* pExceptionPointers, int prepare)
{
    static BOOL bMiniDumpSuccessful;
    static char szFileName[MAX_PATH], errmsg[MAX_PATH];
    static DWORD dwBufferSize = MAX_PATH, current_process_id, current_thread_id;
    static HANDLE hDumpFile, current_process;
    static SYSTEMTIME stLocalTime;
    static MINIDUMP_EXCEPTION_INFORMATION ExpParam;

    if (prepare)
    {
        current_process = GetCurrentProcess();
        current_process_id = GetCurrentProcessId();
        current_thread_id = GetCurrentThreadId();

        GetLocalTime(&stLocalTime);
        sprintf(szFileName, "JBSQL-%s-Crashdump-%04d%02d%02d-%02d%02d%02d-%ld-%ld.dmp",
            GetVersionInfo(argv0), stLocalTime.wYear, stLocalTime.wMonth, stLocalTime.wDay,
            stLocalTime.wHour, stLocalTime.wMinute, stLocalTime.wSecond,
            current_process_id, current_thread_id);

    }
    else
    {
        SetUnhandledExceptionFilter(NULL);
        signal(SIGSEGV, NULL);
        signal(SIGILL, NULL);
        signal(SIGABRT, NULL);
        signal(SIGFPE, NULL);
        signal(SIGTERM, NULL);
        if (MyMiniDumpWriteDump)
        {
            hDumpFile = CreateFile(szFileName, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0);
            if (hDumpFile != INVALID_HANDLE_VALUE)
            {
                if (pExceptionPointers != NULL)
                {
                    ExpParam.ThreadId = GetCurrentThreadId();
                    ExpParam.ExceptionPointers = pExceptionPointers;
                    ExpParam.ClientPointers = TRUE;
                    bMiniDumpSuccessful = MyMiniDumpWriteDump(current_process, current_process_id,
                        hDumpFile, MiniDumpWithFullMemory, &ExpParam, NULL, NULL);
                }
                else
                    bMiniDumpSuccessful = MyMiniDumpWriteDump(current_process, current_process_id,
                        hDumpFile, MiniDumpWithFullMemory, NULL, NULL, NULL);
                CloseHandle(hDumpFile);
                sprintf(errmsg, "JBSQL.EXE has crashed! PLEASE UPLOAD %s to http://duofoto.be/mantis", szFileName);
                MessageBox(NULL, errmsg, "CRASH", MB_OK);
            }
            else
            {
                MessageBox(NULL, "JBSQL has crashed :-(", "CRASH", MB_OK);
            }
        }
    }
    return EXCEPTION_EXECUTE_HANDLER;
}

LONG WINAPI UnhandledException(LPEXCEPTION_POINTERS ptr)
{
    return GenerateDump(g_argv0, ptr, 0);
}

void SignalHandler(int sig)
{
    switch (sig)
    {
        case SIGABRT:
            WPrintf("ABORTING PROGRAM: Abnormal termination (SIGABORT)");
            break;
        case SIGFPE:
            WPrintf("ABORTING PROGRAM: Floating-point error (SIGFPE)");
            break;
        case SIGILL:
            WPrintf("ABORTING PROGRAM: Illegal instruction (SIGILL)");
            break;
        case SIGSEGV:
            WPrintf("ABORTING PROGRAM: Access Violation (SIGSEGV)");
            break;
        case SIGTERM:
            WPrintf("ABORTING PROGRAM: Terminate (SIGTERM)");
            break;
        default:
            WPrintf("ABORTING PROGRAM: Unknown signal (%d)", sig);
            break;
    }
    GenerateDump(g_argv0, NULL, 0);
    exit(sig);
}


int PASCAL WinMain(HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPSTR lpCmdLine,
    int nCmdShow)
{
    MSG  msg;
    char** argv;
    int argc, n, nolog = 0, x;
    HACCEL acctable;
    HACCEL acctable2;
    HWND focus;
#undef _DEBUG
#ifndef _DEBUG
    HMODULE module;

    __try
    {
        module = LoadLibrary("DbgHelp.dll");
        if (module)
        {
            MyMiniDumpWriteDump = (MyMiniDumpWriteDumpFunc)GetProcAddress(module, "MiniDumpWriteDump");
        }

        SetUnhandledExceptionFilter(UnhandledException);
        signal(SIGSEGV, SignalHandler);
        signal(SIGILL, SignalHandler);
        signal(SIGABRT, SignalHandler);
        signal(SIGFPE, SignalHandler);
        signal(SIGTERM, SignalHandler);
#endif
        g_text = (TextWidget*)malloc(sizeof(TextWidget));
        memset(g_text, 0, sizeof(TextWidget));

        g_hInst = hInstance;

        if (!hPrevInstance)
            if (!InitApplication(hInstance))
                return FALSE;

        InitCommonControls();
        InitializeCriticalSection(&g_mutex);
        InitCCsplitter();

        if (!InitInstance(hInstance, nCmdShow))
            return FALSE;

        if (InitODBC())
        {
            MessageBox(NULL, "Failed to initialize OCB", "ODBC error", MB_OK | MB_ICONERROR);
            return 0;
        }
        vht = CreateVariableHashTable();
        ConvertCommandLine(GetCommandLine(), &argc, &argv);
        g_argv0 = argv[0];
        for (x = 1; x < argc; x++)
        {
            if (argv[x][0] == '/' || argv[x][0] == '-')
            {
                if (!strcmp(&argv[x][1], "nolog"))
                    nolog = 1;
            }
        }
#ifndef _DEBUG
        GenerateDump(argv[0], NULL, 1); // prepare crash dump
#endif
        GetModuleFileName(NULL, executable, MAX_PATH);
        tid = 1;
        ReadUserCredentials();
        g_menu = LoadMenu(hInst, MAKEINTRESOURCE(IDR_LV_POPUP));
        g_menu_tables = CreatePopupMenu();
        if (argc > 1)
        {
            tid = 0;
            if (!ConnectWithConnectString(argv[1], 0))
            {
                tid = 1;
                //      ShowPrompt(g_text);
            }
            else
                tid = 1;
        }
        if (!ora_connected && !odbc_connected)
        {
            if (!nolog)
            {
                if (!ShowLoginBox(g_hWnd, g_text))
                    return 0;
            }
            else
            {
                ShowPrompt(g_text);
                SetWindowTitle("", "", 0, 0);
            }
        }
        else
        {
            GetDefaultOptions(g_text);
            SetPWD(g_text);
        }
        if (g_text->old_version < 1010000)
            MigrateHistory();
        ReadCommandHistory(g_text);
        EnableWindow(hwndTextBox, TRUE);
        tid = 0;
        g_tid = GetCurrentThreadId();
        acctable = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));
        acctable2 = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCEL_TW));

        if (argc > 2)
            RunInitialCommand(g_text, argc, argv);

        if (g_text->options[VERSION_CHECK].value.int_value && time(NULL) % 2 == 0)
        {
            DWORD t;
            CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)CheckNewVersionSilent, hwndTextBox, 0, &t);
        }
        while (GetMessage(&msg, NULL, 0x00, 0x00) > 0)
        {
            if (msg.message == WM_KEYDOWN)
            {
                focus = GetFocus();
                if (g_text && focus == GetDlgItem(g_text->schema, ID_SCHEMABROWSER_CODE))
                {
                    if (TranslateAccelerator(g_text->schema, acctable, &msg))
                        continue;
                }
                else if (g_text && (focus == hwndListView || focus == hwndTextBox))
                {
                    if (TranslateAccelerator(g_hWnd, acctable2, &msg))
                        continue;
                }
                else if (g_text && focus == GetDlgItem(g_text->analyze, IDC_SQL_STATEMENT))
                {
                    if (TranslateAccelerator(g_text->analyze, acctable, &msg))
                        continue;
                }
            }
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        hwndTextBox = NULL;
        if (tid)
        {
            if (g_text->datasource == DS_ORACLE)
                OCI_AbortQuery();
            else
                ODBC_AbortQuery();
        }
        n = 0;
        while (tid && n++ < 100)
            Sleep(20);
        if (g_text->spoolfile)
            CloseHandle(g_text->spoolfile);
        SaveCommandHistory(g_text);
        CloseTMPFile();
        if (g_text->datasource == DS_ORACLE)
            OCI_FreeCols(&oci_cols, &ncols);
        else
            ODBC_FreeCols(&odbc_cols, &ncols);
        if (!tid && ora_connected)
            DisconnectFromOracle();
#ifndef _DEBUG
    }
    __except (GenerateDump(NULL, GetExceptionInformation(), 0))
    {
    }
    FreeLibrary(module);
#endif
    return (int)msg.wParam;

}

