#include <windows.h>
#include <winbase.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdarg.h>
#include "resource.h"
#include <oci.h>
#include <time.h>
#include "hashtab.h"
#include <sql.h>
#include <sqlext.h>
#include <Windowsx.h>
#include "TreeList.h"

#ifndef __JBSQL_H__
#define __JBSQL_H__

#define JBSQL_VERSION "1.4.0.5"
#define CURRENT_JBSQL_VERSION 1040004

#define GS_VIEW_WINDOW      "VIEW_WINDOW"

#define ID_LISTVIEW       200
#define ID_EDIT           201
#define ID_SPLITTER       202
#define ID_HINTLIST       203
#define ID_TOOLBAR        204
#define ID_SCHEMABROWSER  205
#define ID_SQLSPLITTER    206
#define ID_SCHEMABROWSER_SCHEMA_TEXT 206
#define ID_SCHEMABROWSER_SCHEMA      207
#define ID_SCHEMABROWSER_TYPE_TEXT   208
#define ID_SCHEMABROWSER_TYPE        209
#define ID_SCHEMABROWSER_NAME_TEXT   210
#define ID_SCHEMABROWSER_NAME        211
#define ID_SCHEMABROWSER_CODE        212
#define ID_FAVOURITES_TEXT           213
#define ID_FAVOURITES_TREE           214
#define ID_FAVOURITES_PASTE          215
#define ID_FAVOURITES_CUT            216
#define ID_FAVOURITES_NEWFOLDER      217
#define ID_FAVOURITES_CLOSE          218
#define ID_FAVOURITES_TITLE          219
#define ID_FAVOURITES_LABEL          220
#define ID_FAVOURITES_LABEL2         221


#define UM_TOOLTIP        1000        /* DO NOT CHANGE THIS!!! */
#define ID_FAVOURITE_BASE 2000        /* DO NOT CHANGE THIS!!! */
#define ID_TABLE_BASE     3000        /* DO NOT CHANGE THIS!!! */

#define LEFT_TIMER     1
#define RIGHT_TIMER    2
#define TOP_TIMER      3
#define BOTTOM_TIMER   4
#define PASTE_TIMER    5

#define BUFLEN 5000
#define Y_OFFSET 2
#define X_OFFSET 2
#define CARET_OFFSET 2
#define SCROLL_TIMEOUT 20
#define PASTE_TIMEOUT 500
#define NTOKENS 10

#define DELIMITERS "()+-*/<>=!~;:.'@%,\\\"^&|{}?\t[] "
#define DELIMITERS_HINT "()+-*/<>=!~;:'@%,\\\"^&|{}?\t[]\n "
#define MOVE_DELIMITERS "\n()+-*/<>=!~;:.'@%,\\\"^&|{}?\t[] "
#define SQL_DELIMITERS "()+-*/<>=!~;:.'@%,\"^&|{}? 	[]\n\r"

#define MIN(x,y) (((int)(x)>(int)(y))?(y):(x))
#define MAX(x,y) (((int)(x)<(int)(y))?(y):(x))

#define STR_OPTION(o) text->options[o].value.str_value
#define INT_OPTION(o) text->options[o].value.int_value

#define DisplayLastError(h) DisplayLastErrorInternal(h,__LINE__,__FILE__)

typedef enum { TIMER_STMT, TIMER_SCRIPT } TimerSpec;
typedef enum { DS_ORACLE, DS_ODBC } DataSource;
typedef enum {
	COPY_FIRSTCOLUMN, COPY_ENTIRELINE, COPY_THISCOLUMN, COPY_PASTETHISCOLUMN, COPY_QUOTEDIN, COPY_UNQUOTEDIN,
	COPY_ASCOMMA, COPY_ASCOLON, COPY_PASTEASQUOTEDIN, COPY_PASTEASUNQUOTEDIN
} CopySpec;

typedef union {
	char          *str_value;
	OCILobLocator **lob_value;
	void          *object_value;
} Value;

typedef int(*QueryProc)(HWND hWnd, char *object);
typedef int(*SetOptionProc)(void *text, void *value);


typedef struct {
	char  *col_name;
	int    col_size;
	Value  value;
	short     *indicator;
	void *    null_struct;
	int       datatype;
	OCIType   *tdo;
	char      *buffer;
} OCIColumnDesc;

typedef struct {
	SQLCHAR       *col_name;
	SQLULEN       col_size;
	union {
		char          *str_value;
	} value;
	SQLLEN *indicator;
	SQLSMALLINT   datatype;
} ODBCColumnDesc;

enum {
	ALLOW_NEWLINE = 0,
	ARRAYSIZE,
	ASKBIND,
	AUTOCOMPLETE,
	AUTOTRACE,
	BACKGROUNDCOLOR,
	COMMANDHISTORY,
	DEBUG,
	DIRECTORY,
	EDITOR,
	EDIT_FILENAME,
	FAST_RESULTS,
	FAVDIR,
	FEEDBACK,
	FONT,
	FONT_LV,
	HEADING,
	HISTORY,
	LISTVIEW,
	LONG_SIZE,
	MAX_SCRIPTS,
	MAXLINES,
	MAXMEM,
	NULLSTR,
	SAVEHISTORY,
	SCAN,
	SCREEN,
	SERVEROUTPUT,
	SOUND,
	SOUNDFILE,
	SOUNDTIME,
	SQLPROMPT,
	TIMING,
	TRIMSPOOL,
	VERIFY,
	VERSION_CHECK
} OptionCode;

typedef enum { O_CHAR, O_INT, O_BOOLEAN, O_DIR, O_FONT, O_FONTVAR, O_COLOR } OptionType;
typedef enum { V_CHAR, V_NUMBER, V_DATE } VariableType;

typedef struct {
	char option[30];
	union {
		char *str_value;
		int int_value;
	} value;
	int value_size;
	int icon;
	int on_toolbar;
	OptionType type;
	char *tooltip;
	char *description;
	SetOptionProc proc;
	int min_value;
	int settable;
	int subkey; // 0 for global setting, 1 for per database setting
} Option;

typedef struct {
	int  statistic;
	int  value;
	char name[65];
} SessionStatistic;

typedef struct {
	char *command;
	char lastchar;
	int  sqlcommand;
} CommandHistory;

typedef struct {
	HMENU menu;
	UINT item;
	char *filename;
} FavouriteItem;

typedef struct {
	HMENU menu;
	UINT item;
	char *owner;
	char *table;
	char *constraint;
	int *columns;
	int ncolumns;
	char **rcolumns;
	int nrcolumns;
} TableItem;

typedef struct {
	char  **text;    /* Contains all text lines */
	int   lines;     /* Contains the number of lines in the text widget */
	int   top;       /* Contains the first displayed line number */
	int   left;      /* Contains the first displayed left character */
	int   width;     /* Contains the number of visible characters */
	int   height;    /* Contains the number of visible lines */
	int   f_width;   /* The font width */
	int   f_height;  /* The font height */
	int   w_width;   /* The window width */
	int   w_height;  /* The window height */
	int   cur_pos;   /* Current cursor position */
	int   cur_line;  /* Current line number, 1-based */
	HDC   hdc;       /* The device context */
	HFONT hfont;     /* The font to draw with */
	int   active;    /* Is the window active? */
	HWND  hintlist;  /* The hintlist window */
	int   selection; /* Is there an active selection */
	int   sel_left;  /* Left selection position */
	int   sel_top;   /* Top selection position */
	int   sel_right; /* Right selection position */
	int   sel_bottom;/* Bottom selection position */
	int   no_process;/* Do not process newlines */
	char  pwd[MAX_PATH]; /* The current directory */
	char  dir[MAX_PATH]; /* The script directory */
	int   termout;   /* Output commands to terminal */
	int   noredraw;  /* Must we redraw the screen on WPrintf */
	int   scripts;   /* The number of scripts currently nested */
	Option *options; /* The current options */
	DWORD start_time[2];/* The start time of a query/script */
	HANDLE spoolfile;/* The file to spool to */
	UINT_PTR  leftselect;  /* Timer for left offscreen mouse selection */
	UINT_PTR  rightselect; /* Timer for right offscreen mouse selection */
	UINT_PTR  topselect;   /* Timer for top offscreen mouse selection */
	UINT_PTR  bottomselect;/* Timer for bottom offscreen mouse selection */
	int   scanf;     /* Are we asking for user input? 0=no, 1=yes, 2=yes, but hidden */
	int   old_screen;/* temporary storage for the screen option when running scripts */
	HWND  analyze;   /* The handle of the sql analyzer window */
	HWND  schema;    /* The handle of the schema browser window */
	HWND  sqlmonitor; /* The handle of the sql monitor window */
	SessionStatistic *pre_stats; /* Pre statement statistics */
	int    npre_stats; /* The number of pre-statistics */
	HWND  toolbar;   /* The window handle of the toolbar */
	CommandHistory work_cmd; /* The command-line that the user was working on when he presses the up arrow */
	int   cancelled; /* Has the execution of a script been cancelled */
	char *(tokens[10]);
	int   ntokens;   /* The number of tokens in the command */
	FavouriteItem *favourite_items; /* The favourite menu items */
	int nfavourite_items; /* The number of items in the list */
	TableItem *table_items; /* The table menu items */
	int ntable_items; /* The number of items in the list */
	int old_version; /* The previous installed jbsql version */
	DataSource datasource; /* What are we connected to */
	int table_children; /* 0=parent databrowser menu, 1=childreen databrowser menu */
	char username[100];
	char database[200];
	char privilege[1];
	char *server_version;
	char *server_release;
	int scanf_linepos; /* The line position where the cursor was when we started a scanf */
	COLORREF backgroundcolor; /* The background color of the text */
} TextWidget;

typedef struct {
	TextWidget *text;
	char       *data;
	char       action[500];
	char       executestr[500];
	int        fast;
} ExecuteStruct;

typedef struct {
	char *token;
	char *prev_token;
	char *query;
	char *values;
} SQLToken;

#define OT_COLUMNS    1
#define OT_TABLES     2

typedef struct {
	char *token;
	char *prev_token;
	unsigned queries;
	char *values;
} ODBCToken;

typedef struct {
	char *keyword;
	char *value;
} Keyword;

typedef struct {
    int       id;
    HTREEITEM tree_item;
} PlanElement;

typedef struct {
    int         id;
    NODE_HANDLE tree_item;
} MonitorPlanElement;

typedef struct {
	char hint[40];
	int  table;
	int  index;
	char *description;
} SQLHint;

typedef struct {
	char *name;
	int datatype;
	ub4 size;
	sb2 indicator;
	union {
		char          *str_value;
		OCINumber     number_value;
		OCIDate       date_value;
		OCILobLocator *lob_value;
		float         float_value;
		double        double_value;
	} value;
	short temporary;
} BindVariable;

extern HINSTANCE hInst, g_hInst;
extern HWND g_hWnd;
extern HWND hwndListView;
extern HWND hwndTextBox;
extern HWND hwndSplitter;
extern HWND g_toolbar;
extern HMENU g_menu;
extern HMENU g_menu_tables;
extern int ora_connected;
extern int odbc_connected;
extern int recordlength;
extern FILE *tmpfile_write;
extern FILE *tmpfile_read;
extern OCIColumnDesc *oci_cols;
extern ODBCColumnDesc *odbc_cols;
extern int ncols;
extern int nrows;
extern int total_qrows;
extern int linepos;
extern int nlines;
extern DWORD tid;
extern DWORD ptid;
extern char *command;
extern int allow_blank_lines;
extern TextWidget *g_text;
extern CommandHistory *old_commands;
extern int nold_commands;
extern char last_usernames[10][100];
extern char last_databases[10][100];
extern int noptions;
extern Option default_options[];
extern char connectstring[MAX_PATH], g_database[250];
extern HashTable vht;
extern DWORD g_tid;
extern int max_command_size, cur_command_size;
extern int window_h_offset, window_v_offset;
extern int g_subitem_clicked;
extern char *planstr;
extern CRITICAL_SECTION g_mutex;
extern int do_refresh;
extern int max_fetch_array_size;
extern int query_finished, total_qrows;
extern SQLULEN rowsfetched;
extern BindVariable *bind_variables;
extern int nbind_variables;
extern char g_title[200];
extern char executable[MAX_PATH];

void ReallocCommand(int size);
void *MallocCommand(int size);
void FreeCommand(void);
char *_strdupCommand(char *str);

void Error(HWND hWnd, char *fmt, ...);
void DisplayLastErrorInternal(HWND hWnd, int line, char *file);
void Log(char *fmt, ...);
char *myvsprintf(const char *fmt, va_list vl);
void WPrintf(char *fmt, ...);
void WPrintfUP(char *str);
void WPrintfNL(void);
void WPrintfOnly(char *fmt, ...);
void FPrintf(char *fmt, ...);
void WScanf(TextWidget *text, char *str, int hidden);
int ReplaceString(char *str, char *srch, char *repl);
void ConvertCommandLine(char *commandline, int *argc, char ***argv);

HWND CreateListView(HINSTANCE hInstance, HWND hwndParent);
void ResizeListView(HWND hwndListView, HWND hwndParent);
BOOL InitListView(HWND hwndListView);
LRESULT ListViewNotify(HWND hWnd, LPARAM lParam);
void ClearList(HWND hwndListView, int remove_cols);

HWND CreateTextWidget(HINSTANCE hInst, HWND parent, TextWidget **rtext);
void InsertLineBuffer(TextWidget *text, char *fmt, ...);
void InsertUnparsedLineBuffer(TextWidget *text, char *str);
void InsertNewLineBuffer(TextWidget *text);
void UpdateCaretPosition(TextWidget *text, int scroll);
void RedrawLine(TextWidget *text, int line);
void RefreshTextWidget(TextWidget *text);

HWND CreateSplitter(HINSTANCE hInstance, HWND hwndParent);
//void ResizeSplitter(HWND hwndSplitter, HWND hwndParent);

void EnableThreads(void);

int ConnectToOracle(char *username, char *password, char *database, char *connect_as, int add);
int DisconnectFromOracle(void);
int ConnectWithConnectString(char *connectstring, int add);
void OCI_AbortQuery(void);
void ODBC_AbortQuery(void);
void SetWindowTitle(char *username, char *database, int instance, int sid);


void ShowPrompt(TextWidget *text);
void InsertCommandLine(char *str);
int ProcessCommand(TextWidget *text);
void PrintErrorCharacter(int c);
void PreviousStatement(TextWidget *text);
void NextStatement(TextWidget *text);
void RedrawCommand(TextWidget *text, int pos, int nlines);

int CreateTMPFile(TextWidget *text);
void CloseTMPFile(void);

int OCI_PrepareSource(TextWidget *text, char *p_sqlstr, OCIColumnDesc **p_cols, int *p_ncols, char *action);
int OCI_FetchData(TextWidget *text, OCIColumnDesc *cols, int ncols, char *action);
int OCI_FinishSource(OCIColumnDesc *cols, int ncols);
void OCI_FreeCols(OCIColumnDesc **cols, int *ncols);

int ODBC_PrepareSource(TextWidget *text, char *p_sqlstr, ODBCColumnDesc **p_cols, int *p_ncols, char *action);
int ODBC_FetchData(TextWidget *text, ODBCColumnDesc *cols, int ncols, char *action);
int ODBC_FinishSource(ODBCColumnDesc *cols, int ncols);
void ODBC_FreeCols(ODBCColumnDesc **cols, int *ncols);

int OCI_DoCommand(char *str, int feedback, char *executestr);
int ExecuteQuery(ExecuteStruct *s);
void ClearCommand(TextWidget *text, int add, int sql_command);
int ODBC_DoCommand(char *str, int feedback, char *executestr);

void OCI_Commit(void);
void OCI_Rollback(void);
void ODBC_Commit(void);
void ODBC_Rollback(void);
void GetUserID(char *object, char **user_id, char **object_name, int show_error);
char *GetOracleObjectType(char *owner, char *object, int show_error);
int OCI_ExecuteDDLStatement(char *str);
int ODBC_ExecuteDDLStatement(char *str);

int ShowHint(TextWidget *text, char *command, int linepos);
void DestroyHintList(TextWidget *text);
void SelectFromHint(TextWidget *text, int linepos);
void FillHintListFromQuery(TextWidget *text, char *query, char *name, char *owner, char *object_type);
void AddToHintList(TextWidget *text, char *str, int sort);

void Describe(TextWidget *text, char *command);
void ShowError(TextWidget *text, char *type, char *name);

char *Trim(char *str);

void RemoveLines(TextWidget *text, int nlines);
void SyncTextPosition(TextWidget *text);
void AppendLineBuffer(char *str);
int DeleteLineBufferCharacter(TextWidget *text);
int DeleteNextLineBufferCharacter(TextWidget *text);
void InsertTab(TextWidget *text);
void InsertNewLine(TextWidget *text);
void MoveCursorLeft(TextWidget *text, int next_word);
void MoveCursorRight(TextWidget *text, int next_word);
void MoveCursorHome(TextWidget *text, int full);
void MoveCursorEnd(TextWidget *text, int full);
void MoveCursorUp(TextWidget *text);
void MoveCursorDown(TextWidget *text);
void AppendLine(TextWidget *text, char *str);
void InsertLine(TextWidget *text);
void ClearLine(TextWidget *text);
void DeleteCharacter(TextWidget *text);
void DeleteNextCharacter(TextWidget *text);
void GetRealSelection(TextWidget *text, int *sel_left, int *sel_top, int *sel_right, int *sel_bottom);

int RunScript(ExecuteStruct *s);

BOOL FAR PASCAL OptionProc(HWND hWnd, unsigned message, DWORD wParam, LONG lParam);
int GetDefaultOptions(TextWidget *text);
void SetCommandLineOption(TextWidget *text, char *settings);
void SetPWD(TextWidget *text);
void PrintOptions(TextWidget *text);
void SaveUserCredentials(char *username, char *database, DataSource datasource);
void ReadUserCredentials(void);
void ToggleOption(TextWidget *text, int option);
void CopyOptions(Option *source, Option *target);
void FreeOptions(Option *source);
int EnableServerOutput(void *text, void *value);
void RecallCommand(TextWidget *text, CommandHistory *old_command);

void WriteRegistryString(char *key, char *subkey, char *value);
void WriteRegistryInteger(char *key, char *subkey, DWORD value);
void WriteRegistryBytes(char *key, char *subkey, void *value, int sz);
BOOLEAN ReadRegistryString(char *key, char *subkey, char *value, int sz);
BOOLEAN ReadRegistryInteger(char *key, char *subkey, DWORD *value);

int PrintServerOutput(TextWidget *text, int fast);
void SaveCommandHistory(TextWidget *text);
void ReadCommandHistory(TextWidget *text);

HWND BuildToolbar(TextWidget *text, HINSTANCE hInst, HWND hWnd);

HashTable CreateVariableHashTable(void);
void DefineVariable(char *str);
void AcceptVariable(TextWidget *text, char **tokens, int ntokens);
int CheckVariableFormat(VariableType vt, char *format, char *value);

HWND ShowSqlAnalyzer(HWND parenthWnd);
void ExplainPlan(HWND hWnd, char *sql_statement);
void ShowAutotrace(ExecuteStruct *s);
void AnalyzeError(char *fmt, ...);
void SetAnalyzerText(TextWidget *text, char *sqlstmt);

void ShowStatistics(TextWidget *text);
void GetStatistics(SessionStatistic **s, int *ns);

void DeleteCharacterFromString(char *str, int x);

char *GetCurrentUser(void);
HWND ShowSchemaBrowser(HWND parent);
void SchemaPrintf(HWND hWnd, char *fmt, ...);
int GenerateSource(HWND hWnd, char *schema, char *name, char *type);
int ProcessTable(HWND hWnd, char *schema, char *table);
int ProcessIndex(HWND hWnd, char *str);
int ProcessConstraint(HWND hWnd, char *str);
int ProcessFunction(HWND hWnd, char *schema, char *function);
int ProcessProcedure(HWND hWnd, char *schema, char *procedure);
int ProcessPackage(HWND hWnd, char *schema, char *package);
int ProcessPackageBody(HWND hWnd, char *schema, char *packagebody);
int ProcessTrigger(HWND hWnd, char *schema, char *trigger);
int ProcessView(HWND hWnd, char *schema, char *view);
int ProcessType(HWND hWnd, char *schema, char *type);
int ProcessTypeBody(HWND hWnd, char *schema, char *type);
int ProcessJavaSource(HWND hWnd, char *schema, char *javasource);
int ProcessSynonym(HWND hWnd, char *schema, char *synonym);
int ProcessSequence(HWND hWnd, char *schema, char *sequence);

void ExportData(TextWidget *text);
void CopyFromListView(HWND hWnd, int entire_line);
void CopyToClipboard(TextWidget *text);
int PasteFromClipboard(TextWidget *text);

void SaveToHistory(char *command);
void ShowHistory(TextWidget *text);

void AddFavourite(TextWidget *text, HWND parent);
void ShowFavourites(HWND parent, TextWidget *text);
int GetPreviousSQLCommand(void);
void AddFavouritesMenu(TextWidget *text, HWND hwnd);
int IsFavouriteItem(TextWidget *text, DWORD item);
void FlashJBSQL(void);
void StartTimer(TextWidget *text, int which_timer);
char *StopTimer(TextWidget *text, int which_timer, int flash);

void MigrateHistory(void);
void FillHintListFromArgumentQuery(TextWidget *text, char *query, char *name, char *owner, char *object_type);
void CheckNewVersion(HWND hWnd, int feedback);
void CheckNewVersionSilent(HWND hWnd);
void RefreshScreen(void *data);
void ResizeTextWidget(HWND hwnd, int update);
void RefreshTextWindow(HWND hwnd);
void SizeListViewItems(HWND hwndListView);

void StartFastDraw(TextWidget *text, int start_thread);
void EndFastDraw(TextWidget *text);
void ComposeConnectString(char *username, char *password, char *database, char *connect_as, char *datasource);

int ConnectToODBC(char *username, char *password, char *database, int add);
int InitODBC(void);
int DisconnectFromODBC(void);

int ODBC_FillTablesHintList(TextWidget *text, char *firstword, char *secondword, char *lastword);
int ODBC_FillColumnsHintList(TextWidget *text, char *catalog, char *schema, char *table, char *column);
int ODBC_DescribeTable(TextWidget *text, char *catalog, char *schema, char *table);
int ODBC_Dir(TextWidget *text, char *object);

void ToUpper(char *str);
void PrintOCIVersion(void);
int OCIFillMenuFromQuery(TextWidget *text, HMENU menu, char *query, char *owner, char *table);
void InsertPasteBuffer(TextWidget *text, char *str);
void SelectTableData(TextWidget *text, TableItem *item);
void GetXMLFunctions(void);

int ChangeFont(void *text, void *value);
int ChangeLVFont(TextWidget *text, char *value);
int ParseFontString(char *fontstr, char *font, int *pointsize, int *weight, int *italic);
int ChangePassword(char *username, char *password, char *new_password, char *database, char *connect_as, int add);
int OCI_DisconnectServer();

int AddBindVariable(TextWidget *text, char *name, char *datatype, int size);
void ShowBindVariables(TextWidget *text);
int FindBindVariable(char *name);
void PrintBindVariable(TextWidget *text, int bv);
int OCI_BindVariables(TextWidget *text);
void ShowAllBindVariables(TextWidget *text, HWND hWnd);

int GenerateDump(char *argv0, EXCEPTION_POINTERS* pExceptionPointers, int prepare);
void ShowParentChildTables(TextWidget *text, HWND hWnd);

char *OCIGetSysDate(void);
char *OCIGetServerVersion(void);
char *OCIGetServerRelease(void);
char *GetPrompt(void);

typedef sword(*OCIServerReleaseFunc)  (void  *hndlp, OCIError *errhp, OraText *bufp,
	ub4 bufsz,
	ub1 hndltype, ub4 *version);
extern OCIServerReleaseFunc MyOCIServerRelease;
void SetWindowTitleEx(HWND hWnd, char *window);
void NewWindow(TextWidget *text);
int ChangeBackgroundColor(TextWidget *text, char *value);
void ParseColorString(char *color, COLORREF *rgb);
void WriteToBuffer(char *buffer, int size);
HWND ShowSqlMonitor(HWND parenthWnd);


#endif