#include "jbsql.h"
#include "ccsplitter.h"

#define ITEM_COUNT 1000

int  last_ncols = 0;
char* buffer = NULL;
int g_subitem_clicked = 0;

HWND CreateListView(HINSTANCE hInstance, HWND hwndParent)
{
    DWORD       dwStyle;
    HWND        l_hwndListView;
    BOOL        bSuccess = TRUE;

    dwStyle = WS_TABSTOP |
        WS_CHILD |
        WS_VISIBLE |
        LVS_REPORT |
        LVS_OWNERDATA |
        LVS_NOSORTHEADER |
        LVS_SHOWSELALWAYS | 
        WS_EX_COMPOSITED;

    l_hwndListView = CreateWindowEx(WS_EX_CLIENTEDGE,          // ex style
        WC_LISTVIEW,               // class name - defined in commctrl.h
        "",                        // dummy text
        dwStyle,                   // style
        0,                         // x position
        0,                         // y position
        0,                         // width
        200,                         // height
        hwndParent,                // parent
        (HMENU)ID_LISTVIEW,        // ID
        g_hInst,                   // instance
        NULL);                     // no extra data

    if (!l_hwndListView)
        return NULL;

    ListView_SetExtendedListViewStyleEx(l_hwndListView, 0, LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES | LVS_EX_HEADERDRAGDROP | LVS_EX_DOUBLEBUFFER | LVS_EX_COLUMNOVERFLOW | LVS_EX_HEADERINALLVIEWS);

    //	ResizeListView(hwndListView, hwndParent);

    CCsplitter_SetHandle(hwndParent, CCSPV_BOTTOM, l_hwndListView);


    return l_hwndListView;
}

/******************************************************************************

   PositionHeader

   this needs to be called when the ListView is created, resized, the view is
   changed or a WM_SYSPARAMETERCHANGE message is recieved

******************************************************************************/

void PositionHeader(HWND hwndListView)
{
    HWND  hwndHeader = GetWindow(hwndListView, GW_CHILD);
    DWORD dwStyle = GetWindowLong(hwndListView, GWL_STYLE);

    /*To ensure that the first item will be visible, create the control without
    the LVS_NOSCROLL style and then add it here*/
    //dwStyle |= LVS_NOSCROLL;
    //SetWindowLong(hwndListView, GWL_STYLE, dwStyle);

    //only do this if we are in report view and were able to get the header hWnd
    if (((dwStyle & LVS_TYPEMASK) == LVS_REPORT) && hwndHeader)
    {
        RECT        rc;
        HD_LAYOUT   hdLayout;
        WINDOWPOS   wpos;

        GetClientRect(hwndListView, &rc);
        hdLayout.prc = &rc;
        hdLayout.pwpos = &wpos;

        Header_Layout(hwndHeader, &hdLayout);

        SetWindowPos(hwndHeader,
            wpos.hwndInsertAfter,
            wpos.x,
            wpos.y,
            wpos.cx,
            wpos.cy,
            wpos.flags | SWP_SHOWWINDOW);

        ListView_EnsureVisible(hwndListView, 0, FALSE);
    }
}

/******************************************************************************

   ResizeListView

******************************************************************************/

void ResizeListView(HWND hwndListView, HWND hwndParent)
{
    RECT  rc;

    GetClientRect(hwndParent, &rc);

    MoveWindow(hwndListView,
        rc.left,
        205,//rc.top,
        rc.right - rc.left,
        -200 + rc.bottom - rc.top,
        TRUE);

    //only call this if we want the LVS_NOSCROLL style
//	PositionHeader(hwndListView);
}

void SizeListViewItems(HWND hwndListView)
{
    int x, a_width, h_width;

    SendMessage(hwndListView, WM_SETREDRAW, FALSE, 0);
    for (x = 0; x < ncols; x++)
    {
        ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE);
        a_width = ListView_GetColumnWidth(hwndListView, x);
        ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE_USEHEADER);
        h_width = ListView_GetColumnWidth(hwndListView, x);
        if (a_width > h_width)
            ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE);
    }

    SendMessage(hwndListView, WM_SETREDRAW, TRUE, 0);
}

BOOL InsertListViewItems(HWND hwndListView, int count)
{
    //int x , a_width, h_width;


    //empty the list
    ListView_DeleteAllItems(hwndListView);

    //set the number of items in the list
  //  ListView_SetItemCountEx(hwndListView, count, LVSICF_NOINVALIDATEALL);
/*    for (x = 0; x < ncols; x++)
    {
        ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE);
        a_width = ListView_GetColumnWidth(hwndListView, x);
        ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE_USEHEADER);
        h_width = ListView_GetColumnWidth(hwndListView, x);
        if (a_width > h_width)
            ListView_SetColumnWidth(hwndListView, x, LVSCW_AUTOSIZE);
    }*/
    return TRUE;
}

void ClearList(HWND hwndListView, int remove_cols)
{
    int i;

    //empty the list
    if (remove_cols)
    {
        SendMessage(hwndListView, WM_SETREDRAW, FALSE, 0);
        ListView_DeleteAllItems(hwndListView);
        for (i = last_ncols - 1; i >= 0; i--)
            ListView_DeleteColumn(hwndListView, i);
        SendMessage(hwndListView, WM_SETREDRAW, TRUE, 0);
        last_ncols = 0;
    }
    else
    {
        SendMessage(hwndListView, WM_SETREDRAW, FALSE, 0);
        ListView_DeleteAllItems(hwndListView);
        SendMessage(hwndListView, WM_SETREDRAW, TRUE, 0);
    }
}

/******************************************************************************

   InitListView

******************************************************************************/

BOOL InitListView(HWND hwndListView)
{
    LVCOLUMN   lvColumn;
    int        i, bufsize;

    SendMessage(hwndListView, WM_SETREDRAW, FALSE, 0);
    HWND hWndHdr = (HWND)SendMessage(hwndListView, LVM_GETHEADER, 0, 0);
    int count = (int)SendMessage(hWndHdr, HDM_GETITEMCOUNT, 0, 0L);
    nrows = 0;
    if (count > ncols)
    {
        for (i = count - 1; i >= ncols; i--)
        {
            ListView_DeleteColumn(hwndListView, i);
        }
    }
    last_ncols = ncols;
    //initialize the columns
    
    lvColumn.fmt = LVCFMT_LEFT;
    lvColumn.cx = 120;
    recordlength = 0;
    bufsize = 0;
    if (buffer != NULL)
        free(buffer);
    if (g_text->datasource == DS_ORACLE)
    {
        for (i = 0; i < ncols; i++)
        {
            recordlength += oci_cols[i].col_size;
            if (oci_cols[i].col_size > bufsize)
                bufsize = oci_cols[i].col_size;
            lvColumn.pszText = oci_cols[i].col_name;
            lvColumn.cchTextMax = (int)strlen(lvColumn.pszText);
            lvColumn.iSubItem = i;
            if (i >= count)
            {
                lvColumn.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
                ListView_InsertColumn(hwndListView, i, &lvColumn);
            }
            else
            {
                lvColumn.mask = LVCF_FMT | LVCF_TEXT | LVCF_SUBITEM;
                ListView_SetColumn(hwndListView, i, &lvColumn);
            }
        }
    }
    else
    {
        for (i = 0; i < ncols; i++)
        {
            lvColumn.pszText = (char*)odbc_cols[i].col_name;
            lvColumn.cchTextMax = (int)strlen(lvColumn.pszText);
            lvColumn.iSubItem = i;
            if (i >= count)
            {
                lvColumn.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
                ListView_InsertColumn(hwndListView, i, &lvColumn);
            }
            else
            {
                lvColumn.mask = LVCF_FMT | LVCF_TEXT | LVCF_SUBITEM;
                ListView_SetColumn(hwndListView, i, &lvColumn);
            }
            recordlength += (int)odbc_cols[i].col_size;
            if ((int)odbc_cols[i].col_size > bufsize)
                bufsize = (int)odbc_cols[i].col_size;
        }
    }

    if (bufsize)
        buffer = (char*)malloc(bufsize);
    else
    {
        SendMessage(hwndListView, WM_SETREDRAW, TRUE, 0);
        RedrawWindow(hwndListView, NULL, NULL, RDW_NOERASE | RDW_FRAME | RDW_INVALIDATE | RDW_ALLCHILDREN);
        return FALSE;
    }
    InsertListViewItems(hwndListView, nrows);
    SendMessage(hwndListView, WM_SETREDRAW, TRUE, 0);
    //    InvalidateRect(hwndListView, NULL, TRUE);

    return TRUE;
}

/**************************************************************************

   ListViewNotify()

**************************************************************************/

LRESULT ListViewNotify(HWND hWnd, LPARAM lParam)
{
    LPNMHDR  lpnmh = (LPNMHDR)lParam;

    switch (lpnmh->code)
    {
        case LVN_GETDISPINFO:
            {
                LV_DISPINFO* lpdi = (LV_DISPINFO*)lParam;
                size_t bytes_read;
                LARGE_INTEGER lposition;

                if (lpdi->item.mask & LVIF_TEXT)
                {
                    int x, offset = 0;

                    if (g_text->datasource == DS_ORACLE)
                    {
                        if (oci_cols)
                        {
                            for (x = 0; x < lpdi->item.iSubItem; x++)
                                offset += oci_cols[x].col_size;
                            lposition.QuadPart = (__int64)recordlength * (__int64)lpdi->item.iItem + (__int64)offset;
                            if (_fseeki64(tmpfile_read, lposition.QuadPart, SEEK_SET))
                                Error(NULL, "%s", strerror(errno));
                            bytes_read = fread(buffer, 1, MIN(oci_cols[lpdi->item.iSubItem].col_size - 1, lpdi->item.cchTextMax - 1), tmpfile_read);
                            buffer[bytes_read] = '\0';
                        }
                    }
                    else
                    {
                        if (odbc_cols)
                        {
                            for (x = 0; x < lpdi->item.iSubItem; x++)
                                offset += (int)odbc_cols[x].col_size;
                            lposition.QuadPart = (__int64)recordlength * (__int64)lpdi->item.iItem + (__int64)offset;
                            if (_fseeki64(tmpfile_read, lposition.QuadPart, SEEK_SET))
                                Error(NULL, "%s", strerror(errno));
                            bytes_read = fread(buffer, 1, MIN(odbc_cols[lpdi->item.iSubItem].col_size - 1, lpdi->item.cchTextMax - 1), tmpfile_read);
                            buffer[bytes_read] = '\0';
                        }
                    }
                    lpdi->item.pszText = buffer;
                }
            }
            return 0;
        case LVN_ODCACHEHINT:
            {
                LPNMLVCACHEHINT   lpCacheHint = (LPNMLVCACHEHINT)lParam;
                /*
                This sample doesn't use this notification, but this is sent when the
                ListView is about to ask for a range of items. On this notification,
                you should load the specified items into your local cache. It is still
                possible to get an LVN_GETDISPINFO for an item that has not been cached,
                therefore, your application must take into account the chance of this
                occurring.
                */
            }
            return 0;

        case LVN_ODFINDITEM:
            {
                LPNMLVFINDITEM lpFindItem = (LPNMLVFINDITEM)lParam;
                /*
                This sample doesn't use this notification, but this is sent when the
                ListView needs a particular item. Return -1 if the item is not found.
                */
            }
            return 0;
        case LVN_DELETEALLITEMS:
            return TRUE;
        case NM_RCLICK:
            {
                g_subitem_clicked = ((LPNMLISTVIEW)lParam)->iSubItem;

                TextWidget* text;
                BOOL menuitem;

                if (tmpfile_write) /* There's something in the listview */
                {
                    POINTS cursorPos;
                    DWORD pos = GetMessagePos();
                    cursorPos=MAKEPOINTS(pos);

                    text = (TextWidget*)GetWindowLongPtr(hwndTextBox, 0);
                    menuitem = TrackPopupMenu(GetSubMenu(g_menu, 0),
                        TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD | TPM_RIGHTBUTTON,
                        cursorPos.x,
                        cursorPos.y,
                        0,
                        hWnd,
                        NULL);
                    switch (menuitem)
                    {
                        case ID_COPY_COPYFIRSTCOLUMN:
                            CopyFromListView(hwndListView, COPY_FIRSTCOLUMN);
                            break;
                        case ID_COPY_COPYENTIRELINE:
                            CopyFromListView(hwndListView, COPY_ENTIRELINE);
                            break;
                        case ID_COPY_COPYTHISCOLUMN:
                            CopyFromListView(hwndListView, COPY_THISCOLUMN);
                            break;
                        case ID_PASTE_THIS_COLUMN:
                            CopyFromListView(hwndListView, COPY_PASTETHISCOLUMN);
                            break;
                        case ID_COPYSPECIAL_COPYASQUOTED:
                            CopyFromListView(hwndListView, COPY_QUOTEDIN);
                            break;
                        case ID_COPYSPECIAL_COPYASUNQUOTED:
                            CopyFromListView(hwndListView, COPY_UNQUOTEDIN);
                            break;
                        case ID_COPYSPECIAL_COPYASCOMMA:
                            CopyFromListView(hwndListView, COPY_ASCOMMA);
                            break;
                        case ID_COPYSPECIAL_COPYASCOLON:
                            CopyFromListView(hwndListView, COPY_ASCOLON);
                            break;
                        case ID_COPYSPECIAL_PASTEASQUOTED:
                            CopyFromListView(hwndListView, COPY_PASTEASQUOTEDIN);
                            break;
                        case ID_COPYSPECIAL_PASTEASUNQUOTED:
                            CopyFromListView(hwndListView, COPY_PASTEASUNQUOTEDIN);
                            break;
                        case ID_DATABROWSER_CHILDRECORDS:
                        case ID_DATABROWSER_PARENTRECORDS:
                            {
                                text->table_children = (menuitem == ID_DATABROWSER_CHILDRECORDS);
                                ShowParentChildTables(text, hWnd);
                            }
                            break;
                    }
                }
                break;
            }
    }

    return 0;
}

