#include "jboci.h"

BindVariable *bind_variables = NULL;
int nbind_variables = 0;
int current_bindvariable = -1;

char *DataTypeStr2(int datatype)
{
	static char str[40];

	switch (datatype)
	{
		case SQLT_STR:
			sprintf(str, "VARCHAR2");
			break;
		case SQLT_VNU:
			strcpy(str, "NUMBER");
			break;
		case SQLT_ODT:
			strcpy(str, "DATE");
			break;
		case SQLT_CHR:
			sprintf(str, "CHAR");
			break;
		case SQLT_CLOB:
			strcpy(str, "CLOB");
			break;
		case SQLT_BLOB:
			strcpy(str, "BLOB");
			break;
		case SQLT_BFILE:
			strcpy(str, "BFILE");
			break;
		case SQLT_BFLOAT:
			strcpy(str, "BINARY_FLOAT");
			break;
		case SQLT_BDOUBLE:
			strcpy(str, "BINARY_DOUBLE");
			break;
	}
	return str;
}

int Str2DataType(char *datatype)
{
	if (!strcmp(datatype, "VARCHAR2"))
		return SQLT_STR;
	if (!strcmp(datatype, "NUMBER"))
		return SQLT_VNU;
	if (!strcmp(datatype, "DATE"))
		return SQLT_ODT;
	if (!strcmp(datatype, "CHAR"))
		return SQLT_CHR;
	if (!strcmp(datatype, "CLOB"))
		return SQLT_CLOB;
	if (!strcmp(datatype, "BLOB"))
		return SQLT_BLOB;
	if (!strcmp(datatype, "BFILE"))
		return SQLT_BFILE;
	if (!strcmp(datatype, "BINARY_FLOAT"))
		return SQLT_BFLOAT;
	if (!strcmp(datatype, "BINARY_DOUBLE"))
		return SQLT_BDOUBLE;
	return -1;
}

void FreeBindVariable(int bv, int remove)
{
	switch (bind_variables[bv].datatype)
	{
		case SQLT_STR:
			free(bind_variables[bv].value.str_value);
			bind_variables[bv].size = 0;
			break;
		case SQLT_CHR:
			free(bind_variables[bv].value.str_value);
			bind_variables[bv].size = 0;
			break;
		case SQLT_CLOB:
		case SQLT_BLOB:
		case SQLT_BFILE:
			if (bind_variables[bv].temporary)
				ChkOCIErr(OCILobFreeTemporary(g_oci->svchp, g_oci->errhp, bind_variables[bv].value.lob_value));
			ChkOCIErr(OCIDescriptorFree(bind_variables[bv].value.lob_value, OCI_DTYPE_LOB));
			break;
	}
	if (remove)
	{
		if (nbind_variables == 1)
		{
			free(bind_variables);
			bind_variables = NULL;
		}
		else if (nbind_variables - 1 == bv)
			bind_variables = (BindVariable*)realloc(bind_variables, (nbind_variables - 1) * sizeof(BindVariable));
		else
			memmove(&bind_variables[bv], &bind_variables[bv + 1], (nbind_variables - bv - 1) * sizeof(BindVariable));
		nbind_variables--;
	}
}

int NewBindVariable(char *datatype, int bv, int size)
{
	_strupr(datatype);
	bind_variables[bv].indicator = -1;
	bind_variables[bv].temporary = 0;
	if (!strcmp("VARCHAR2", datatype))
	{
		if (size > 0)
		{
			bind_variables[bv].datatype = SQLT_STR;
			bind_variables[bv].value.str_value = NULL;
			bind_variables[bv].size = size + 1;
			bind_variables[bv].value.str_value = (char*)malloc(size + 1);
		}
		else return 1;
	}
	else if (!strcmp("CHAR", datatype))
	{
		if (size <= 0)
			size = 1;
		bind_variables[bv].datatype = SQLT_CHR;
		bind_variables[bv].value.str_value = NULL;
		bind_variables[bv].size = size + 1;
		bind_variables[bv].value.str_value = (char*)malloc(size + 1);
	}
	else if (!strcmp(datatype, "NUMBER"))
	{
		bind_variables[bv].datatype = SQLT_VNU;
	}
	else if (!strcmp(datatype, "DATE"))
	{
		bind_variables[bv].datatype = SQLT_ODT;
	}
	else if (!strcmp(datatype, "CLOB"))
	{
		bind_variables[bv].datatype = SQLT_CLOB;
		ChkOCIErr(OCIDescriptorAlloc(g_oci->envhp, (void**)&bind_variables[bv].value.lob_value, OCI_DTYPE_LOB, 0, NULL));
	}
	else if (!strcmp(datatype, "BLOB"))
	{
		bind_variables[bv].datatype = SQLT_BLOB;
		ChkOCIErr(OCIDescriptorAlloc(g_oci->envhp, (void**)&bind_variables[bv].value.lob_value, OCI_DTYPE_LOB, 0, NULL));
	}
	else if (!strcmp(datatype, "BFILE"))
	{
		bind_variables[bv].datatype = SQLT_BFILE;
		ChkOCIErr(OCIDescriptorAlloc(g_oci->envhp, (void**)&bind_variables[bv].value.lob_value, OCI_DTYPE_LOB, 0, NULL));
	}
	else if (!strcmp(datatype, "BINARY_FLOAT"))
	{
		bind_variables[bv].datatype = SQLT_BFLOAT;
	}
	else if (!strcmp(datatype, "BINARY_DOUBLE"))
	{
		bind_variables[bv].datatype = SQLT_BDOUBLE;
	}
	else
		return 1;
	return 0;
}


void SetBindDataType(HWND hWnd, int dt)
{
	char datatype[40];

	ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LENGTH), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BIND_NUMBER_VALUE), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BIND_STRING_VALUE), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BIND_TIME_VALUE), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LENGTH_LABEL), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LOAD), FALSE);
	ShowWindow(GetDlgItem(hWnd, IDC_BINDS_SAVE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_BINDS_APPLY), TRUE);
	if (dt >= 0)
	{
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_DATATYPE_LABEL), TRUE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_DATATYPE_LIST), TRUE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_APPLY), TRUE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_VALUE_LABEL), TRUE);

		SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_GETLBTEXT, dt, (LPARAM)datatype);
		if (!strcmp(datatype, "VARCHAR2") || !strcmp(datatype, "CHAR"))
		{
			ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LENGTH), TRUE);
			ShowWindow(GetDlgItem(hWnd, IDC_BIND_STRING_VALUE), TRUE);
			SendDlgItemMessage(hWnd, IDC_BIND_STRING_VALUE, EM_SETREADONLY, (WPARAM)FALSE, (LPARAM)NULL);
			ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LENGTH_LABEL), TRUE);
		}
		if (!strcmp(datatype, "CLOB") || !strcmp(datatype, "BLOB") || !strcmp(datatype, "BFILE"))
		{
			ShowWindow(GetDlgItem(hWnd, IDC_BIND_STRING_VALUE), TRUE);
			SendDlgItemMessage(hWnd, IDC_BIND_STRING_VALUE, EM_SETREADONLY, (WPARAM)TRUE, (LPARAM)NULL);
			ShowWindow(GetDlgItem(hWnd, IDC_BINDS_SAVE), TRUE);
			if (!strcmp(datatype, "CLOB") || !strcmp(datatype, "BLOB"))
			{
				ShowWindow(GetDlgItem(hWnd, IDC_BINDS_LOAD), TRUE);
			}
		}
		else if (!strcmp(datatype, "NUMBER") || !strcmp(datatype, "BINARY_FLOAT") || !strcmp(datatype, "BINARY_DOUBLE"))
		{
			ShowWindow(GetDlgItem(hWnd, IDC_BIND_NUMBER_VALUE), TRUE);
		}
		else if (!strcmp(datatype, "DATE"))
		{
			ShowWindow(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE), TRUE);
			ShowWindow(GetDlgItem(hWnd, IDC_BIND_TIME_VALUE), TRUE);
		}
	}
	else
	{
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_DATATYPE_LABEL), FALSE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_DATATYPE_LIST), FALSE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_APPLY), FALSE);
		ShowWindow(GetDlgItem(hWnd, IDC_BINDS_VALUE_LABEL), FALSE);
	}
	SetDlgItemText(hWnd, IDC_BINDS_ERROR, "");
}

void ShowBindOCIError(HWND hWnd, sb4 errcode)
{
	char buffer[200];

	OCIErrorGet(g_oci->errhp,
		1,
		NULL,
		&errcode,
		(OraText*)buffer,
		200,
		OCI_HTYPE_ERROR);
	SetDlgItemText(hWnd, IDC_BINDS_ERROR, buffer);
}

int SaveBindVariable(HWND hWnd)
{
	int dt;

	SetDlgItemText(hWnd, IDC_BINDS_ERROR, "");
	if (current_bindvariable != -1)
	{
		char datatype[40];
		char size[10];

		SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_GETLBTEXT, (int)SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_GETCURSEL, 0, 0), (LPARAM)datatype);
		dt = Str2DataType(datatype);
		if (dt == SQLT_CHR || dt == SQLT_STR)
		{
			GetDlgItemText(hWnd, IDC_BINDS_LENGTH, size, 10);
			if (atoi(size) <= 0 || atoi(size) > 32767)
			{
				SetDlgItemText(hWnd, IDC_BINDS_ERROR, "Invalid length");
				return 1;
			}
		}
		if (dt != bind_variables[current_bindvariable].datatype || ((dt == SQLT_CHR || dt == SQLT_STR) && atoi(size) != bind_variables[current_bindvariable].size))
		{
			FreeBindVariable(current_bindvariable, 0);
			NewBindVariable(datatype, current_bindvariable, atoi(size));
		}

		switch (bind_variables[current_bindvariable].datatype)
		{
			case SQLT_STR:
			case SQLT_CHR:
				GetDlgItemText(hWnd, IDC_BIND_STRING_VALUE, bind_variables[current_bindvariable].value.str_value, bind_variables[current_bindvariable].size);
				bind_variables[current_bindvariable].indicator = (bind_variables[current_bindvariable].value.str_value[0] == '\0') ? -1 : 0;
				break;
			case SQLT_VNU:
				{
					char buffer[50], *fmt = "FM99999999999999999999999999999999999990.999999999999999999999999";
					GetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, buffer, 49);
					if (buffer[0] == '\0')
						bind_variables[current_bindvariable].indicator = -1;
					else
					{
						int ret = OCINumberFromText(g_oci->errhp, (OraText*)buffer, (ub4)strlen(buffer), (OraText*)fmt, (ub4)strlen(fmt), NULL, 0, &bind_variables[current_bindvariable].value.number_value);
						if (ret == OCI_ERROR)
						{
							ShowBindOCIError(hWnd, ret);
							bind_variables[current_bindvariable].indicator = -1;
						}
						else
							bind_variables[current_bindvariable].indicator = 0;
					}
					break;
				}
			case SQLT_ODT:
				{
					SYSTEMTIME st_date;
					SYSTEMTIME st_time;
					int ret_date, ret_time;

					ret_date = DateTime_GetSystemtime(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE), &st_date);
					ret_time = DateTime_GetSystemtime(GetDlgItem(hWnd, IDC_BIND_TIME_VALUE), &st_time);
					if (ret_date == GDT_NONE && ret_time == GDT_NONE)
					{
						bind_variables[current_bindvariable].indicator = -1;
					}
					else
					{
						if (ret_date == GDT_VALID)
							OCIDateSetDate(&bind_variables[current_bindvariable].value.date_value, st_date.wYear, (ub1)st_date.wMonth, (ub1)st_date.wDay);
						if (ret_time == GDT_VALID)
							OCIDateSetTime(&bind_variables[current_bindvariable].value.date_value, (ub1)st_time.wHour, (ub1)st_time.wMinute, (ub1)st_time.wSecond);
						bind_variables[current_bindvariable].indicator = 0;
					}
					break;
				}
			case SQLT_CLOB:
				break;
			case SQLT_BLOB:
				break;
			case SQLT_BFILE:
				break;
			case SQLT_BFLOAT:
				{
					char buffer[50];
					GetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, buffer, 49);
					if (buffer[0] == '\0')
						bind_variables[current_bindvariable].indicator = -1;
					else
					{
						bind_variables[current_bindvariable].indicator = 0;
						bind_variables[current_bindvariable].value.float_value = (float)atof(buffer);
					}
					break;
				}
			case SQLT_BDOUBLE:
				{
					char buffer[50];
					GetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, buffer, 49);
					if (buffer[0] == '\0')
						bind_variables[current_bindvariable].indicator = -1;
					else
					{
						bind_variables[current_bindvariable].indicator = 0;
						bind_variables[current_bindvariable].value.double_value = atof(buffer);
					}
					break;
				}
		}
	}
	return 0;
}

void ShowBindVariable(TextWidget *text, HWND hWnd, int index)
{
	DWORD txtlen;
	char *txt, *datatype, str_value[100];
	int x, n;
	ub4 value_size;

	txtlen = (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_GETTEXTLEN, index, 0);
	txt = (char*)malloc(txtlen + 1);
	SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_GETTEXT, index, (LPARAM)txt);
	for (x = 0; x < nbind_variables; x++)
	{
		if (!strcmp(bind_variables[x].name, txt))
		{
			current_bindvariable = x;
			datatype = DataTypeStr2(bind_variables[x].datatype);
			n = (int)SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_FINDSTRINGEXACT, -1, (LPARAM)datatype);
			if (n != LB_ERR)
			{
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_SETCURSEL, n, 0);
				SetBindDataType(hWnd, n);
				switch (bind_variables[x].datatype)
				{
					case SQLT_CHR:
					case SQLT_STR:
						if (bind_variables[x].indicator != -1)
							SetDlgItemText(hWnd, IDC_BIND_STRING_VALUE, bind_variables[x].value.str_value);
						else
							SetDlgItemText(hWnd, IDC_BIND_STRING_VALUE, "");
						SetDlgItemInt(hWnd, IDC_BINDS_LENGTH, bind_variables[x].size - 1, 0);
						SetFocus(GetDlgItem(hWnd, IDC_BIND_STRING_VALUE));
						break;
					case SQLT_VNU:
						if (bind_variables[x].indicator != -1)
						{
							value_size = 100;
							OCINumberToText(g_oci->errhp, &bind_variables[x].value.number_value, "TM9", 3, NULL, 0, &value_size, (oratext*)str_value);
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, str_value);
						}
						else
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, "");
						SetFocus(GetDlgItem(hWnd, IDC_BIND_NUMBER_VALUE));
						break;
					case SQLT_BFLOAT:
						if (bind_variables[x].indicator != -1)
						{
							value_size = 100;
							sprintf(str_value, "%f", bind_variables[x].value.float_value);
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, str_value);
						}
						else
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, "");
						SetFocus(GetDlgItem(hWnd, IDC_BIND_NUMBER_VALUE));
						break;
					case SQLT_BDOUBLE:
						if (bind_variables[x].indicator != -1)
						{
							value_size = 100;
							sprintf(str_value, "%lf", bind_variables[x].value.double_value);
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, str_value);
						}
						else
							SetDlgItemText(hWnd, IDC_BIND_NUMBER_VALUE, "");
						SetFocus(GetDlgItem(hWnd, IDC_BIND_NUMBER_VALUE));
						break;
					case SQLT_ODT:
						if (bind_variables[x].indicator != -1)
						{
							sb2 year;
							ub1 month, day, hour, minute, second;
							SYSTEMTIME st;

							value_size = 100;
							OCIDateGetDate(&bind_variables[x].value.date_value, &year, &month, &day);
							OCIDateGetTime(&bind_variables[x].value.date_value, &hour, &minute, &second);
							st.wDay = day;
							st.wMonth = month;
							st.wYear = year;
							st.wHour = hour;
							st.wMinute = minute;
							st.wSecond = second;
							st.wDayOfWeek = 0;
							st.wMilliseconds = 0;
							DateTime_SetSystemtime(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE), GDT_VALID, &st);
							DateTime_SetSystemtime(GetDlgItem(hWnd, IDC_BIND_TIME_VALUE), GDT_VALID, &st);
						}
						else
						{
							DateTime_SetSystemtime(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE), GDT_NONE, NULL);
							DateTime_SetSystemtime(GetDlgItem(hWnd, IDC_BIND_TIME_VALUE), GDT_NONE, NULL);
						}
						SetFocus(GetDlgItem(hWnd, IDC_BIND_DATE_VALUE));
						break;
					case SQLT_CLOB:
					case SQLT_BLOB:
					case SQLT_BFILE:
						if (bind_variables[x].indicator != -1)
						{
							char *buffer = (char*)malloc(INT_OPTION(LONG_SIZE) + 1);
							ub4 amount = INT_OPTION(LONG_SIZE);
							int ret;

							if (buffer != NULL)
							{
								if (bind_variables[x].datatype == SQLT_BFILE)
								{
									if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
										g_oci->errhp,
										bind_variables[x].value.lob_value,
										OCI_FILE_READONLY)))
									{
										return;
									}
								}
								ret = ChkOCIErr(OCILobRead(g_oci->svchp, g_oci->errhp, bind_variables[x].value.lob_value, &amount, 1, buffer, amount, 0, NULL, 0, SQLCS_IMPLICIT));
								if (ret != OCI_SUCCESS && ret != OCI_NEED_DATA)
								{
									if (ChkOCIErr(ret))
									{
										return;
									}
								}
								if (bind_variables[x].datatype == SQLT_BFILE)
								{
									if (ChkOCIErr(OCILobFileClose(g_oci->svchp,
										g_oci->errhp,
										bind_variables[x].value.lob_value)))
									{
										return;
									}
								}
								buffer[amount] = '\0';
								SetDlgItemText(hWnd, IDC_BIND_STRING_VALUE, buffer);
							}
						}
						else
						{
							SetDlgItemText(hWnd, IDC_BIND_STRING_VALUE, "");
						}
						EnableWindow(GetDlgItem(hWnd, IDC_BINDS_APPLY), FALSE);
						break;
				}
			}
		}
	}
	free(txt);
}

int LoadLobVariable(HWND hWnd, TextWidget *text)
{
	if (current_bindvariable != -1)
	{
		OPENFILENAME openFile;
		char filename[MAX_PATH] = "\0";
		HANDLE of;
		ub4 amount, offset = 1;
		void *buffer;
		DWORD bytes_read;

		openFile.lStructSize = sizeof(OPENFILENAME);
		openFile.hwndOwner = g_hWnd;
		openFile.hInstance = NULL;
		openFile.lpstrFilter = NULL;
		openFile.lpstrCustomFilter = NULL;
		openFile.nMaxCustFilter = 255;
		openFile.nFilterIndex = 0;
		openFile.lpstrFile = filename;
		openFile.nMaxFile = MAX_PATH;
		openFile.lpstrFileTitle = NULL;
		openFile.nMaxFileTitle = 0;
		openFile.lpstrInitialDir = NULL;
		openFile.lpstrTitle = NULL;
		openFile.Flags = OFN_OVERWRITEPROMPT;
		openFile.nFileOffset = 0;
		openFile.nFileExtension = 0;
		openFile.lpstrDefExt = NULL;
		openFile.lCustData = 0;
		openFile.lpfnHook = NULL;
		openFile.lpTemplateName = NULL;
		GetOpenFileName(&openFile);
		if (filename[0] != '\0')
		{
			of = CreateFile(filename,
				GENERIC_READ,
				FILE_SHARE_READ,
				NULL,
				OPEN_EXISTING,
				0,
				NULL);
			if (of == INVALID_HANDLE_VALUE)
			{
				DisplayLastError(NULL);
			}
			else
			{
				if (ChkOCIErr(OCILobCreateTemporary(g_oci->svchp, g_oci->errhp, bind_variables[current_bindvariable].value.lob_value, 0, SQLCS_IMPLICIT, (bind_variables[current_bindvariable].datatype == SQLT_BLOB) ? OCI_TEMP_BLOB : OCI_TEMP_CLOB, FALSE, OCI_DURATION_SESSION)))
				{
					CloseHandle(of);
					return 1;
				}
				bind_variables[current_bindvariable].temporary = 1;
				buffer = malloc(32768);
				while (ReadFile(of, buffer, 32768, &bytes_read, NULL) && bytes_read)
				{
					amount = bytes_read;
					if (ChkOCIErr(OCILobWrite(g_oci->svchp, g_oci->errhp, bind_variables[current_bindvariable].value.lob_value, &amount, offset, buffer, 32768, OCI_ONE_PIECE, NULL, NULL, 0, SQLCS_IMPLICIT)))
						break;
					if (bytes_read != 32768)
						break;
					offset += bytes_read;
				}
				free(buffer);
				CloseHandle(of);
				bind_variables[current_bindvariable].indicator = 0;
			}
		}
	}
	return 0;
}

int SaveLobVariable(HWND hWnd, TextWidget *text)
{
	if (current_bindvariable != -1)
	{
		OPENFILENAME openFile;
		char filename[MAX_PATH] = "\0";
		HANDLE of;
		int ret;
		ub4 amount, offset = 1;
		void *buffer;
		DWORD bytes_written;

		openFile.lStructSize = sizeof(OPENFILENAME);
		openFile.hwndOwner = g_hWnd;
		openFile.hInstance = NULL;
		openFile.lpstrFilter = NULL;
		openFile.lpstrCustomFilter = NULL;
		openFile.nMaxCustFilter = 255;
		openFile.nFilterIndex = 0;
		openFile.lpstrFile = filename;
		openFile.nMaxFile = MAX_PATH;
		openFile.lpstrFileTitle = NULL;
		openFile.nMaxFileTitle = 0;
		openFile.lpstrInitialDir = NULL;
		openFile.lpstrTitle = NULL;
		openFile.Flags = OFN_OVERWRITEPROMPT;
		openFile.nFileOffset = 0;
		openFile.nFileExtension = 0;
		openFile.lpstrDefExt = NULL;
		openFile.lCustData = 0;
		openFile.lpfnHook = NULL;
		openFile.lpTemplateName = NULL;
		GetSaveFileName(&openFile);
		if (filename[0] != '\0')
		{
			of = CreateFile(filename,
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				0,
				NULL);
			if (of == INVALID_HANDLE_VALUE)
			{
				DisplayLastError(NULL);
			}
			else
			{
				if (bind_variables[current_bindvariable].datatype == SQLT_BFILE)
				{
					if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
						g_oci->errhp,
						bind_variables[current_bindvariable].value.lob_value,
						OCI_FILE_READONLY)))
					{
						return 1;
					}
				}
				amount = 32768;
				buffer = malloc(amount);
				for (;;)
				{
					ret = OCILobRead(g_oci->svchp, g_oci->errhp, bind_variables[current_bindvariable].value.lob_value, &amount, offset, buffer, amount, 0, NULL, 0, SQLCS_IMPLICIT);
					if (ret != OCI_SUCCESS && ret != OCI_NEED_DATA)
					{
						ChkOCIErr(ret);
						break;
					}
					WriteFile(of, buffer, amount, &bytes_written, NULL);
					offset += amount;
					if (amount < (ub4)32768)
						break;
				}
				if (bind_variables[current_bindvariable].datatype == SQLT_BFILE)
				{
					if (ChkOCIErr(OCILobFileClose(g_oci->svchp,
						g_oci->errhp,
						bind_variables[current_bindvariable].value.lob_value)))
					{
						return 1;
					}
				}
				free(buffer);
			}
			CloseHandle(of);
		}
	}
	return 0;
}

LRESULT CALLBACK NewBindVariableProc(HWND hWnd,
	UINT uMessage,
	WPARAM wParam,
	LPARAM lParam)
{
	switch (uMessage)
	{
		case WM_INITDIALOG:
			break;
		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
			{
				case IDOK:
					{
						char name[40];

						GetDlgItemText(hWnd, IDC_BINDS_NEW_NAME, name, 40);
						_strupr(name);
						if (name[0] != '\0')
						{
							int bv;

							bv = FindBindVariable(name);
							if (bv == -1)
							{
								AddBindVariable(g_text, name, "VARCHAR2", 100);
								EndDialog(hWnd, TRUE);
							}
							else
								EndDialog(hWnd, FALSE);
						}
						break;
					}
				case IDCANCEL:
					EndDialog(hWnd, FALSE);
					break;
			}
			break;

		case WM_CLOSE:
			EndDialog(hWnd, FALSE);
			return TRUE;
		default:
			break;
	}
	return FALSE;
}



int ShowNewBindVariableDialog(HWND parenthWnd)
{
	FARPROC cpproc;
	LRESULT res;

	cpproc = MakeProcInstance((FARPROC)NewBindVariableProc, hInst);
	res = DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_BINDS_NEW), parenthWnd, (DLGPROC)cpproc, (LPARAM)0);
	FreeProcInstance(cpproc);

	return (int)res;
}


LRESULT CALLBACK BindVariablesProc(HWND hWnd,
	UINT uMessage,
	WPARAM wParam,
	LPARAM lParam)
{
	switch (uMessage)
	{
		case WM_INITDIALOG:
			{
				int x;

				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"VARCHAR2");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"CHAR");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"DATE");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"NUMBER");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"CLOB");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"BLOB");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"BFILE");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"BINARY_FLOAT");
				SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_ADDSTRING, 0, (LPARAM)"BINARY_DOUBLE");

				for (x = 0; x < g_oci->nvarnames; x++)
				{
					if (SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_FINDSTRINGEXACT, -1, (LPARAM)g_oci->varnames[x]) == LB_ERR)
					{
						SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_ADDSTRING, 0, (LPARAM)g_oci->varnames[x]);
						if (FindBindVariable((char*)g_oci->varnames[x]) == -1)
						{
							AddBindVariable(NULL, (char*)g_oci->varnames[x], "VARCHAR2", 100);
						}
					}
				}
				SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_SETCURSEL, 0, 0);
				if (g_oci->nvarnames)
					ShowBindVariable(g_text, hWnd, 0);
				else
					SetBindDataType(hWnd, -1);
				if (lParam)
				{
					EnableWindow(GetDlgItem(hWnd, IDC_BINDS_NEW), TRUE);
					EnableWindow(GetDlgItem(hWnd, IDC_BINDS_DELETE), TRUE);
				}
				else
				{
					EnableWindow(GetDlgItem(hWnd, IDC_BINDS_NEW), FALSE);
					EnableWindow(GetDlgItem(hWnd, IDC_BINDS_DELETE), FALSE);
				}
			}
			break;
		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
			{
				case ID_BINDS_OK:
					SaveBindVariable(hWnd);
					EndDialog(hWnd, TRUE);
					break;
				case IDC_BINDS_LIST:
					if (HIWORD(wParam) == LBN_SELCHANGE)
					{
						SaveBindVariable(hWnd);
						ShowBindVariable(g_text, hWnd, (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_GETCURSEL, 0, 0));
					}
					break;
				case IDC_BINDS_DATATYPE_LIST:
					if (HIWORD(wParam) == CBN_SELCHANGE)
						SetBindDataType(hWnd, (int)SendDlgItemMessage(hWnd, IDC_BINDS_DATATYPE_LIST, CB_GETCURSEL, 0, 0));
					EnableWindow(GetDlgItem(hWnd, IDC_BINDS_APPLY), TRUE);
					break;
				case IDC_BINDS_SAVE:
					SaveLobVariable(hWnd, g_text);
					break;
				case IDC_BINDS_LOAD:
					SaveBindVariable(hWnd);
					LoadLobVariable(hWnd, g_text);
					ShowBindVariable(g_text, hWnd, (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_GETCURSEL, 0, 0));
					break;
				case IDC_BINDS_APPLY:
					if (!SaveBindVariable(hWnd))
						ShowBindVariable(g_text, hWnd, (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_GETCURSEL, 0, 0));
					break;
				case IDC_BINDS_NEW:
					if (ShowNewBindVariableDialog(hWnd))
					{
						int pos;

						current_bindvariable = nbind_variables - 1;
						SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_ADDSTRING, 0, (LPARAM)bind_variables[current_bindvariable].name);
						pos = (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_FINDSTRINGEXACT, -1, (LPARAM)bind_variables[current_bindvariable].name);
						SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_SETCURSEL, pos, 0);
						ShowBindVariable(g_text, hWnd, pos);
					}
					break;
				case IDC_BINDS_DELETE:
					if (current_bindvariable >= 0)
					{
						int pos;

						pos = (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_FINDSTRINGEXACT, -1, (LPARAM)bind_variables[current_bindvariable].name);
						SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_DELETESTRING, pos, 0);
						FreeBindVariable(current_bindvariable, 1);
						if (current_bindvariable >= nbind_variables)
							current_bindvariable--;
						if (current_bindvariable >= 0)
						{
							pos = (int)SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_FINDSTRINGEXACT, -1, (LPARAM)bind_variables[current_bindvariable].name);
							SendDlgItemMessage(hWnd, IDC_BINDS_LIST, LB_SETCURSEL, pos, 0);
							ShowBindVariable(g_text, hWnd, pos);
						}
						else
							SetBindDataType(hWnd, -1);
					}
					break;
			}
			break;

		case WM_CLOSE:
			EndDialog(hWnd, FALSE);
			return TRUE;
		default:
			break;
	}
	return FALSE;
}



int ShowBindVariablesDialog(HWND parenthWnd, int allow_add_delete)
{
	FARPROC cpproc;
	LRESULT res;

	cpproc = MakeProcInstance((FARPROC)BindVariablesProc, hInst);
	res = DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_BINDS), parenthWnd, (DLGPROC)cpproc, (LPARAM)allow_add_delete);
	FreeProcInstance(cpproc);

	return !(int)res;
}

char *DataTypeStr(int datatype, int size)
{
	static char str[40];

	switch (datatype)
	{
		case SQLT_STR:
			sprintf(str, "VARCHAR2(%d)", size);
			break;
		case SQLT_VNU:
			strcpy(str, "NUMBER");
			break;
		case SQLT_ODT:
			strcpy(str, "DATE");
			break;
		case SQLT_CHR:
			sprintf(str, "CHAR(%d)", size);
			break;
		case SQLT_CLOB:
			strcpy(str, "CLOB");
			break;
		case SQLT_BLOB:
			strcpy(str, "BLOB");
			break;
		case SQLT_BFILE:
			strcpy(str, "BFILE");
			break;
		case SQLT_BFLOAT:
			strcpy(str, "BINARY_FLOAT");
			break;
		case SQLT_BDOUBLE:
			strcpy(str, "BINARY_DOUBLE");
			break;
	}
	return str;
}


int OCI_BindVariables(TextWidget *text)
{
	int x, bv;
	sb4 found;
	OraText **indnames = NULL;
	ub1 *indlengths = NULL, *varlengths = NULL;
	ub1 *dups = NULL;
	OCIBind **binds = NULL;

	g_oci->varnames = (OraText**)malloc(sizeof(OraText*));
	varlengths = (ub1*)malloc(sizeof(ub1));
	indnames = (OraText**)malloc(sizeof(OraText*));
	indlengths = (ub1*)malloc(sizeof(ub1));
	dups = (ub1*)malloc(sizeof(ub1));
	binds = (OCIBind**)malloc(sizeof(OCIBind*));

	OCIStmtGetBindInfo(g_oci->stmtp,
		g_oci->errhp,
		1,
		1,
		&found,
		g_oci->varnames,
		varlengths,
		indnames,
		indlengths,
		dups,
		binds);

	if (found < -1)
	{
		g_oci->nvarnames = -found;
		g_oci->varnames = (OraText**)realloc((void*)g_oci->varnames, g_oci->nvarnames * sizeof(OraText*));
		varlengths = (ub1*)realloc((void*)varlengths, g_oci->nvarnames * sizeof(ub1));
		indnames = (OraText**)realloc((void*)indnames, g_oci->nvarnames * sizeof(OraText*));
		indlengths = (ub1*)realloc((void*)indlengths, g_oci->nvarnames * sizeof(ub1));
		dups = (ub1*)realloc((void*)dups, g_oci->nvarnames * sizeof(ub1));
		binds = (OCIBind**)realloc((void*)binds, g_oci->nvarnames * sizeof(OCIBind*));
		OCIStmtGetBindInfo(g_oci->stmtp,
			g_oci->errhp,
			g_oci->nvarnames,
			1,
			&found,
			g_oci->varnames,
			varlengths,
			indnames,
			indlengths,
			dups,
			binds);
	}
	if (found)
	{
		g_oci->nvarnames = found;
		if (INT_OPTION(ASKBIND))
		{
			ShowBindVariablesDialog(g_hWnd, 0);
			StartTimer(g_text, TIMER_STMT);
		}
	}
	for (x = 0; x < found; x++)
	{
		if (!dups[x])
		{
			bv = FindBindVariable((char*)g_oci->varnames[x]);
			if (bv == -1)
			{
				WPrintf("\rBind variable \"%s\" not defined\n", g_oci->varnames[x]);
				return 1;
			}
			else
			{
				void *value_ptr;
				ub4 value_size;
				OCIBind *bindp = NULL;

				switch (bind_variables[bv].datatype)
				{
					case SQLT_STR:
						value_ptr = bind_variables[bv].value.str_value;
						value_size = bind_variables[bv].size;
						break;
					case SQLT_CHR:
						value_ptr = bind_variables[bv].value.str_value;
						value_size = bind_variables[bv].size;
						break;
					case SQLT_VNU:
						value_ptr = &bind_variables[bv].value.number_value;
						value_size = sizeof(OCINumber);
						break;
					case SQLT_ODT:
						value_ptr = &bind_variables[bv].value.date_value;
						value_size = sizeof(OCIDate);
						break;
					case SQLT_CLOB:
					case SQLT_BLOB:
					case SQLT_BFILE:
						value_ptr = &bind_variables[bv].value.lob_value;
						value_size = sizeof(OCILobLocator*);
						break;
					case SQLT_BFLOAT:
						value_ptr = &bind_variables[bv].value.float_value;
						value_size = sizeof(float);
						break;
					case SQLT_BDOUBLE:
						value_ptr = &bind_variables[bv].value.double_value;
						value_size = sizeof(double);
						break;
				}
				if (ChkOCIErr(OCIBindByName(g_oci->stmtp, &bindp, g_oci->errhp, g_oci->varnames[x], (ub4)strlen((char*)g_oci->varnames[x]), value_ptr, value_size, bind_variables[bv].datatype, &bind_variables[bv].indicator, NULL, NULL, 0, 0, OCI_DEFAULT)))
				{
					return 1;
				}
			}
		}
	}
	free(g_oci->varnames);
	free(varlengths);
	free(indnames);
	free(indlengths);
	free(dups);
	free(binds);
	return 0;
}


void PrintBindVariable(TextWidget *text, int bv)
{
	char *buffer = NULL;
	ub4 bufsize;

	WPrintf("%s\n", bind_variables[bv].name);
	WPrintf("------------\n");
	if (bind_variables[bv].indicator == -1)
		WPrintf("\n");
	else
	{
		switch (bind_variables[bv].datatype)
		{
			case SQLT_STR:
				WPrintf("%s\n", bind_variables[bv].value.str_value);
				break;
			case SQLT_CHR:
				bind_variables[bv].value.str_value[bind_variables[bv].size] = '\0';
				WPrintf("%s\n", bind_variables[bv].value.str_value);
				break;
			case SQLT_VNU:
				bufsize = 100;
				buffer = (char*)malloc(bufsize);
				ChkOCIErr(OCINumberToText(g_oci->errhp, &bind_variables[bv].value.number_value, "TM9", 3, NULL, 0, &bufsize, (oratext*)buffer));
				WPrintf("%s\n", buffer);
				free(buffer);
				break;
			case SQLT_ODT:
				bufsize = 100;
				buffer = (char*)malloc(bufsize);
				OCIDateToText(g_oci->errhp, &bind_variables[bv].value.date_value, NULL, 0, NULL, 0, &bufsize, (oratext*)buffer);
				WPrintf("%s\n", buffer);
				free(buffer);
				break;
			case SQLT_CLOB:
			case SQLT_BLOB:
				{
					ub4 amount;

					amount = INT_OPTION(LONG_SIZE);
					buffer = (char*)malloc(amount + 1);
					ChkOCIErr(OCILobRead(g_oci->svchp,
						g_oci->errhp,
						bind_variables[bv].value.lob_value,
						&amount,
						1,
						buffer,
						amount,
						0,
						NULL,
						0,
						SQLCS_IMPLICIT));
					buffer[amount] = '\0';
					WPrintf("%s\n", buffer);
					free(buffer);
				}
				break;
			case SQLT_BFILE:
				{
					ub4 amount;

					if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
						g_oci->errhp,
						bind_variables[bv].value.lob_value,
						OCI_FILE_READONLY)))
					{
						return;
					}

					amount = INT_OPTION(LONG_SIZE);
					buffer = (char*)malloc(amount + 1);
					ChkOCIErr(OCILobRead(g_oci->svchp,
						g_oci->errhp,
						bind_variables[bv].value.lob_value,
						&amount,
						1,
						buffer,
						amount,
						0,
						NULL,
						0,
						SQLCS_IMPLICIT));
					buffer[amount] = '\0';
					WPrintf("%s\n", buffer);
					free(buffer);
					ChkOCIErr(OCILobFileClose(g_oci->svchp, g_oci->errhp, bind_variables[bv].value.lob_value));
				}
				break;
			case SQLT_BFLOAT:
				WPrintf("%f\n", bind_variables[bv].value.float_value);
				break;
			case SQLT_BDOUBLE:
				WPrintf("%lf\n", bind_variables[bv].value.double_value);
				break;
		}
	}
}


void ShowBindVariables(TextWidget *text)
{
	int x;

	StartFastDraw(text, 0);
	WPrintf("%-30s %s\n", "NAME", "DATATYPE");
	WPrintf("------------------------------ ----------\n");
	for (x = 0; x < nbind_variables; x++)
	{
		WPrintf("%-30s %s\n", bind_variables[x].name, DataTypeStr(bind_variables[x].datatype, bind_variables[x].size - 1));
	}
	EndFastDraw(text);
}

int FindBindVariable(char *name)
{
	int x;

	for (x = 0; x < nbind_variables; x++)
	{
		if (!_stricmp(name, bind_variables[x].name))
			return x;
	}
	return -1;
}

int AddBindVariable(TextWidget *text, char *name, char *datatype, int size)
{
	int bv;

	bv = FindBindVariable(name);
	if (bv != -1)
		FreeBindVariable(bv, 0);
	else
	{
		bind_variables = (BindVariable*)realloc(bind_variables, (nbind_variables + 1) * sizeof(BindVariable));
		bv = nbind_variables++;
		bind_variables[bv].name = _strdup(name);
		_strupr(bind_variables[bv].name);
	}
	return NewBindVariable(datatype, bv, size);
}

void ShowAllBindVariables(TextWidget *text, HWND hWnd)
{
	int x;

	if (!tid && text->datasource == DS_ORACLE)
	{
		g_oci->varnames = (OraText**)malloc(nbind_variables * sizeof(OraText*));
		for (x = 0; x < nbind_variables; x++)
		{
			g_oci->varnames[x] = (OraText*)_strdup(bind_variables[x].name);
		}
		g_oci->nvarnames = nbind_variables;
		ShowBindVariablesDialog(hWnd, 1);
		for (x = 0; x < g_oci->nvarnames; x++)
		{
			free(g_oci->varnames[x]);
		}
		if (g_oci->varnames)
			free(g_oci->varnames);
		g_oci->nvarnames = 0;
	}
}