#include "jboci.h"

int ExecuteDDL(char* str, int verbose)
{
    ub4 qrows;
    int result;

    if (!ora_connected)
    {
        Error(NULL, "Not connected to a database\n");
        return -1;
    }
    if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        str,
        (ub4)strlen(str),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -2;
    }
    g_oci->offset = 1;
    if ((result = (OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT))) != OCI_SUCCESS)
    {
        if (verbose)
            ChkOCIErr2(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -3;
    }
    g_oci->offset = 0;
    ChkOCIErr2(OCIAttrGet(g_oci->stmtp,
        OCI_HTYPE_STMT,
        &qrows,
        NULL,
        OCI_ATTR_ROW_COUNT,
        g_oci->errhp));
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return qrows;
}

void RenumberPlan(HWND hWnd, HTREEITEM pi, int* n)
{
    HTREEITEM ti;
    TVITEM i;
    char str[2000], newstr[2020];

    ti = TreeView_GetChild(hWnd, pi);
    if (ti != NULL)
    {
        RenumberPlan(hWnd, ti, n);
        (*n)++;
    }
    while ((ti = TreeView_GetNextSibling(hWnd, ti)) != NULL)
    {
        RenumberPlan(hWnd, ti, n);
        (*n)++;
    }
    i.mask = TVIF_HANDLE | TVIF_TEXT;
    i.hItem = pi;
    i.pszText = str;
    i.cchTextMax = 2000;
    TreeView_GetItem(hWnd, &i);
    sprintf(newstr, "%d: %s", *n, str);
    i.mask = TVIF_HANDLE | TVIF_TEXT;
    i.hItem = pi;
    i.pszText = newstr;
    i.cchTextMax = 2020;
    TreeView_SetItem(hWnd, &i);
}

int FillNotes(HWND hWnd, char* str)
{
    SetDlgItemText(hWnd, IDC_HINT_TIP, str);
    return 0;
}

void ExplainPlan(HWND hWnd, char* sql_statement)
{
    char* str;
    char char_ptr[2000], stmt_id[50], tmpstr[2000];
    int len = 2000, result;
    int id, parent_id, level;
    char query[2000] = "";
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    HashTable ht;
    TVINSERTSTRUCT item_data;
    HTREEITEM tree_item;
    PlanElement pe;
    ub2 ind1;
    time_t t;
    int n;

    sprintf(stmt_id, "JBSQL_%ld", (long)time(&t));
    TreeView_DeleteAllItems(GetDlgItem(hWnd, IDC_PLAN_TREE));
    if (planstr)
    {
        free(planstr);
        planstr = NULL;
    }
    ExecuteDDL("savepoint JBSQL", 1);
    sprintf(char_ptr, "delete from plan_table where statement_id='%s'", stmt_id);
    if (ExecuteDDL(char_ptr, 1) < 0)
    {
        Error(NULL, "Can't access the plan table, maybe you need to run the utlxplan.sql script");
        return;
    }
    str = (char*)malloc(strlen(sql_statement) + 250);
    if (str == NULL)
        return;
    for (size_t x = strlen(sql_statement) - 1; x >= 0; x--)
    {
        if (sql_statement[x] != ' ' && sql_statement[x] != '\t' && sql_statement[x] != ';' && sql_statement[x] != '\r' && sql_statement[x] != '\n')
            break;
        sql_statement[x] = '\0';
    }
    sprintf(str, "EXPLAIN PLAN SET STATEMENT_ID = '%s' FOR %s", stmt_id, sql_statement);
    if (ExecuteDDL(str, 1) < 0)
    {
        free(str);
        return;
    }
    free(str);

    if (ExecuteDDL("update plan_table set access_predicates=null where 1=0", 0) < 0) /* Use <=8i plan table */
        sprintf(query, "SELECT ID,PARENT_ID,LEVEL,substr(OPERATION||DECODE(OTHER_TAG,NULL,'','*')||DECODE(OPTIONS,NULL,'',' ('||OPTIONS||')')||DECODE(OBJECT_NAME,NULL,'',' OF '''||OBJECT_OWNER||'.'||OBJECT_NAME||'''')||DECODE(OBJECT_TYPE,NULL,'',' ('||OBJECT_TYPE||')')||DECODE(ID,0,DECODE(OPTIMIZER,NULL,'',' Optimizer='||OPTIMIZER))||DECODE(COST,NULL,'',' (Cost='||COST||DECODE(CARDINALITY,NULL,'',' Rows='||CARDINALITY)||DECODE(BYTES,NULL,'',' Bytes='||BYTES)||')')||DECODE(PARTITION_START,NULL,'',' PStart='||PARTITION_START)||DECODE(PARTITION_STOP,NULL,'',' PStop='||PARTITION_STOP),1,2000) PLAN FROM PLAN_TABLE START WITH ID=0 AND STATEMENT_ID='%s' CONNECT BY PRIOR ID=PARENT_ID AND STATEMENT_ID='%s' ORDER BY ID,POSITION", stmt_id, stmt_id);
    else if (ExecuteDDL("update plan_table set qblock_name=null where 1=0", 0) < 0) /* Use 9i */
        sprintf(query, "SELECT ID,\
       PARENT_ID,\
       LEVEL,\
       substr(OPERATION||\
              DECODE(OTHER_TAG,NULL,'','*')||\
              DECODE(OPTIONS,NULL,'',' ('||OPTIONS||')')||\
              DECODE(OBJECT_NAME,NULL,'',' OF '''||OBJECT_OWNER||'.'||OBJECT_NAME||'''')||\
              DECODE(OBJECT_TYPE,NULL,'',' ('||OBJECT_TYPE||')')||\
              DECODE(ID,0,DECODE(OPTIMIZER,NULL,'',' Optimizer='||OPTIMIZER))||\
              DECODE(COST,NULL,'',' (Cost='||COST||\
              DECODE(CARDINALITY,NULL,'',' Rows='||CARDINALITY)||\
              DECODE(BYTES,NULL,'',' Bytes='||BYTES)||\
              DECODE(CPU_COST,NULL,'',' CPU='||CPU_COST)||\
              DECODE(IO_COST,NULL,'',' IO='||IO_COST)||\
              DECODE(TEMP_SPACE,NULL,'',' TEMP='||TEMP_SPACE)||')')||\
              DECODE(PARTITION_START,NULL,'',' PStart='||PARTITION_START)||\
              DECODE(PARTITION_STOP,NULL,'',' PStop='||PARTITION_STOP)||\
              DECODE(PARTITION_ID,NULL,'',' PId='||PARTITION_ID)||\
              DECODE(access_predicates,null,'',' ACCESS=('||substr(access_predicates,1,500)||')')||\
              DECODE(filter_predicates,null,'',' FILTER=('||substr(filter_predicates,1,500)||')')\
             ,1,2000) PLAN \
FROM PLAN_TABLE \
START WITH ID=0 \
AND STATEMENT_ID='%s' \
CONNECT BY PRIOR ID=PARENT_ID \
AND STATEMENT_ID='%s' \
ORDER BY ID,POSITION", stmt_id, stmt_id);
    else /* use new plan table */
        sprintf(query, "SELECT ID,\
       PARENT_ID,\
       LEVEL,\
       substr(OPERATION||\
              DECODE(OTHER_TAG,NULL,'','*')||\
              DECODE(OPTIONS,NULL,'',' ('||OPTIONS||')')||\
              DECODE(OBJECT_NAME,NULL,'',' OF '''||OBJECT_OWNER||'.'||OBJECT_NAME||'''')||\
              DECODE(OBJECT_TYPE,NULL,'',' ('||OBJECT_TYPE||')')||\
              DECODE(ID,0,DECODE(OPTIMIZER,NULL,'',' Optimizer='||OPTIMIZER))||\
              DECODE(COST,NULL,'',' (Cost='||COST||\
              DECODE(CARDINALITY,NULL,'',' Rows='||CARDINALITY)||\
              DECODE(BYTES,NULL,'',' Bytes='||BYTES)||\
              DECODE(CPU_COST,NULL,'',' CPU='||CPU_COST)||\
              DECODE(IO_COST,NULL,'',' IO='||IO_COST)||\
              DECODE(TEMP_SPACE,NULL,'',' TEMP='||TEMP_SPACE)||')')||\
              DECODE(PARTITION_START,NULL,'',' PStart='||PARTITION_START)||\
              DECODE(PARTITION_STOP,NULL,'',' PStop='||PARTITION_STOP)||\
              DECODE(PARTITION_ID,NULL,'',' PId='||PARTITION_ID)||\
              DECODE(QBLOCK_NAME,NULL,'',' QBLOCK='||QBLOCK_NAME)||\
              DECODE(access_predicates,null,'',' ACCESS=('||substr(access_predicates,1,500)||')')||\
              DECODE(filter_predicates,null,'',' FILTER=('||substr(filter_predicates,1,500)||')')\
             ,1,2000) PLAN \
FROM PLAN_TABLE \
START WITH ID=0 \
AND STATEMENT_ID='%s' \
CONNECT BY PRIOR ID=PARENT_ID \
AND STATEMENT_ID='%s' \
ORDER BY ID,POSITION", stmt_id, stmt_id);
    char_ptr[0] = '\0';
    if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return;
    }
    if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        &id,
        sizeof(int),
        SQLT_INT,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }

    if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        2,
        &parent_id,
        sizeof(int),
        SQLT_INT,
        &ind1,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }

    if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        3,
        &level,
        sizeof(int),
        SQLT_INT,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }

    if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        4,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }

    g_oci->offset = 0;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        ChkOCIErr2(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    ht = HTCreate(101, sizeof(int), sizeof(PlanElement), NULL, NULL, NULL, NULL);
    for (;;)
    {
        result = OCIStmtFetch(g_oci->stmtp,
            g_oci->errhp,
            1,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            break;
        }
        if (result == OCI_NO_DATA)
            break;

        if (ind1 == (ub2)-1)
            item_data.hParent = TVI_ROOT;
        else
        {
            HTGetRecord(ht, &parent_id, &pe);
            item_data.hParent = pe.tree_item;
        }
        item_data.hInsertAfter = TVI_LAST;
        item_data.item.mask = TVIF_STATE | TVIF_TEXT;
        item_data.item.state = TVIS_EXPANDED;
        item_data.item.stateMask = TVIS_EXPANDED;
        item_data.item.pszText = char_ptr;
        item_data.item.cchTextMax = (int)strlen(char_ptr);

        memset(tmpstr, ' ', sizeof(tmpstr));
        strcpy(&tmpstr[level], char_ptr);
        if (planstr)
        {
            planstr = realloc(planstr, strlen(planstr) + strlen(tmpstr) + 3);
            if (planstr == NULL)
                return;
            strcat(planstr, "\n");
            strcat(planstr, tmpstr);
        }
        else
        {
            planstr = _strdup(tmpstr);
        }
        tree_item = TreeView_InsertItem(GetDlgItem(hWnd, IDC_PLAN_TREE), &item_data);
        pe.id = id;
        pe.tree_item = tree_item;
        HTAddRecord(ht, &pe);
    }
    HTDestroy(ht);
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    ProcessQuery(g_oci, hWnd, "with pln as (select rownum as rn, plan_table_output from table(dbms_xplan.display(statement_id=>:stmt_id,format=>'basic note'))) select listagg(replace(pln.plan_table_output,'   -','-'), chr(13)||chr(10)) within group (order by rn)  from pln where rn > (select rn + 1 from pln where pln.plan_table_output = 'Note')", (QueryProc)FillNotes, 2000, "stmt_id", stmt_id, NULL);

    ExecuteDDL("rollback to savepoint JBSQL", 1);
    n = 1;
    RenumberPlan(GetDlgItem(hWnd, IDC_PLAN_TREE), TreeView_GetRoot(GetDlgItem(hWnd, IDC_PLAN_TREE)), &n);
}

void FillListBoxFromQuery(OCIEnvironment*g_oci, HWND hWnd, char* query, char** bindvar, int nbindvars, int listbox, int len)
{
    char* char_ptr = NULL;
    int result, x;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    ub4 qrows;
    int l_nrows;
    char l_query_finished = 0;
    ub4 l_total_qrows = 0;
    int fetch_array_size = SCHEMA_ARRAY_SIZE;
    OCIEnvironment* gg_oci;
    {
        extern OCIEnvironment* g_oci;
        gg_oci = g_oci;
    }
    if (g_oci==gg_oci && tid)
        return;
    SendMessage(hWnd, WM_SETREDRAW, FALSE, 0);
    if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return;
    }
    if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    for (x = 0; x < nbindvars; x++)
    {
        if (ChkOCIErr2(OCIBindByPos(g_oci->stmtp,
            &bndhp,
            g_oci->errhp,
            x + 1,
            bindvar[x],
            (ub4)strlen(bindvar[x]),
            SQLT_CHR,
            NULL,
            NULL,
            NULL,
            0,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
    }
    char_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * len * sizeof(char));
    if (char_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return;
    }
    if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }

    g_oci->offset = 0;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }
    while (!l_query_finished)
    {
        result = OCIStmtFetch(g_oci->stmtp,
            g_oci->errhp,
            fetch_array_size,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            if (ChkOCIErrI(result, 1406))
                break;
        }
        if (ChkOCIErr2(OCIAttrGet(g_oci->stmtp,
            OCI_HTYPE_STMT,
            &qrows,
            NULL,
            OCI_ATTR_ROW_COUNT,
            g_oci->errhp)))
        {
            break;
        }
        l_nrows = qrows - l_total_qrows;
        if (l_nrows != fetch_array_size)
            l_query_finished = 1;

        for (x = 0; x < (int)(qrows - l_total_qrows); x++)
        {
            SendMessage(hWnd, listbox ? LB_ADDSTRING : CB_ADDSTRING, 0, (LPARAM)& char_ptr[x * len]);
        }
        l_total_qrows = qrows;
    }
    SendMessage(hWnd, WM_SETREDRAW, TRUE, 0);
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    free(char_ptr);
}