#include "jboci.h"

void FillHintListFromQuery(TextWidget* text, char* query, char* name, char* owner, char* object_type)
{
    char* char_ptr = NULL;
    int len = 200, result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    ub4 qrows;
    int nrows, x;
    char query_finished = 0;
    ub4 total_qrows = 0;
    int fetch_array_size = SCHEMA_ARRAY_SIZE;

    if (ChkOCIErrN(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return;
    }
    if (ChkOCIErrN(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        name,
        (ub4)strlen(name),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (owner != NULL)
    {
        if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
            &bndhp,
            g_oci->errhp,
            2,
            owner,
            (ub4)strlen(owner),
            SQLT_CHR,
            NULL,
            NULL,
            NULL,
            0,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        if (object_type != NULL)
        {
            if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
                &bndhp,
                g_oci->errhp,
                3,
                object_type,
                (ub4)strlen(object_type),
                SQLT_CHR,
                NULL,
                NULL,
                NULL,
                0,
                NULL,
                OCI_DEFAULT)))
            {
                OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
                g_oci->stmtp = NULL;
                return;
            }
        }
    }
    char_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * len * sizeof(char));
    if (char_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }

    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }
    while (!query_finished)
    {
        result = OCIStmtFetch(g_oci->stmtp,
            g_oci->errhp,
            fetch_array_size,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            if (ChkOCIErrI(result, 1406))
                break;
        }
        if (ChkOCIErrN(OCIAttrGet(g_oci->stmtp,
            OCI_HTYPE_STMT,
            &qrows,
            NULL,
            OCI_ATTR_ROW_COUNT,
            g_oci->errhp)))
        {
            break;
        }
        nrows = qrows - total_qrows;
        if (nrows != fetch_array_size)
            query_finished = 1;

        for (x = 0; x < (int)(qrows - total_qrows); x++)
        {
            AddToHintList(text, &char_ptr[x * len], 0);
        }
        total_qrows = qrows;
    }
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    free(char_ptr);
}

void FillHintListFromArgumentQuery(TextWidget* text, char* query, char* name, char* owner, char* object_type)
{
    char* char_ptr = NULL;
    int len = 200, result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    ub4 qrows;
    int nrows, x;
    char query_finished = 0;
    ub4 total_qrows = 0;
    int fetch_array_size = SCHEMA_ARRAY_SIZE;
    int overloads[SCHEMA_ARRAY_SIZE];
    char line[4000] = "";
    int prev_overload = 1, n_in_overload = 0;

    if (ChkOCIErrN(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return;
    }
    if (ChkOCIErrN(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        name,
        (ub4)strlen(name),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (owner != NULL)
    {
        if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
            &bndhp,
            g_oci->errhp,
            2,
            owner,
            (ub4)strlen(owner),
            SQLT_CHR,
            NULL,
            NULL,
            NULL,
            0,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        if (object_type != NULL)
        {
            if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
                &bndhp,
                g_oci->errhp,
                3,
                object_type,
                (ub4)strlen(object_type),
                SQLT_CHR,
                NULL,
                NULL,
                NULL,
                0,
                NULL,
                OCI_DEFAULT)))
            {
                OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
                g_oci->stmtp = NULL;
                return;
            }
        }
    }
    char_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * len * sizeof(char));
    if (char_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }

    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        2,
        overloads,
        sizeof(int),
        SQLT_INT,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }

    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(char_ptr);
        return;
    }
    while (!query_finished)
    {
        result = OCIStmtFetch(g_oci->stmtp,
            g_oci->errhp,
            fetch_array_size,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            if (ChkOCIErrI(result, 1406))
                break;
        }
        if (ChkOCIErrN(OCIAttrGet(g_oci->stmtp,
            OCI_HTYPE_STMT,
            &qrows,
            NULL,
            OCI_ATTR_ROW_COUNT,
            g_oci->errhp)))
        {
            break;
        }
        nrows = qrows - total_qrows;
        if (nrows != fetch_array_size)
            query_finished = 1;

        for (x = 0; x < (int)(qrows - total_qrows); x++)
        {
            if (overloads[x] != prev_overload)
            {
                strcat(line, ")");
                AddToHintList(text, line, 0);
                line[0] = '\0';
                n_in_overload = 0;
                prev_overload = overloads[x];
            }
            if (strlen(line) + strlen(&char_ptr[x * len]) + 1 < 4000)
            {
                if (n_in_overload++)
                    strcat(line, ",");
                strcat(line, &char_ptr[x * len]);
            }
        }
        total_qrows = qrows;
    }
    if (n_in_overload)
    {
        strcat(line, ")");
        AddToHintList(text, line, 0);
    }
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    free(char_ptr);
}

int ColumnsPresent(TextWidget* text, char* owner, char* constraint, TableItem* item, int check)
{
    char* column_ptr = NULL;
    char query[] = "select column_name from all_cons_columns where owner=:1 and constraint_name=:2 order by position";
    int result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    ub4 qrows;
    int nrows, x;
    char query_finished = 0;
    ub4 total_qrows = 0;
    int fetch_array_size = SCHEMA_ARRAY_SIZE;
    OCIStmt* stmtp;
    int cols_found = 1, col_found = 0;
    int offset, y;

    if (ChkOCIErrN(OCIHandleAlloc(g_oci->envhp,
        &stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return 1;
    }
    if (ChkOCIErrN(OCIStmtPrepare(stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        owner,
        (ub4)strlen(owner),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(stmtp,
        &bndhp,
        g_oci->errhp,
        2,
        constraint,
        (ub4)strlen(constraint),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return 1;
    }
    column_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * 129 * sizeof(char));
    if (column_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return 1;
    }
    if (ChkOCIErrN(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        1,
        column_ptr,
        129,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        free(column_ptr);
        return 1;
    }

    offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        free(column_ptr);
        return 1;
    }
    while (!query_finished)
    {
        result = OCIStmtFetch(stmtp,
            g_oci->errhp,
            fetch_array_size,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            if (ChkOCIErrI(result, 1406))
                break;
        }
        if (ChkOCIErrN(OCIAttrGet(stmtp,
            OCI_HTYPE_STMT,
            &qrows,
            NULL,
            OCI_ATTR_ROW_COUNT,
            g_oci->errhp)))
        {
            break;
        }
        nrows = qrows - total_qrows;
        if (nrows != fetch_array_size)
            query_finished = 1;

        for (x = 0; x < (int)(qrows - total_qrows); x++)
        {
            if (check)
            {
                col_found = 0;
                for (y = 0; y < ncols; y++)
                {
                    if (!_stricmp(oci_cols[y].col_name, &column_ptr[x * 129]))
                    {
                        col_found = 1;
                        item->columns = (int*)realloc(item->columns, (item->ncolumns + 1) * sizeof(int));
                        item->columns[item->ncolumns] = y;
                        item->ncolumns++;
                        break;
                    }
                }
                if (!col_found)
                    cols_found = 0;
            }
            else
            {
                item->rcolumns = (char**)realloc(item->rcolumns, (item->nrcolumns + 1) * sizeof(char*));
                item->rcolumns[item->nrcolumns] = _strdup(&column_ptr[x * 129]);
                item->nrcolumns++;
            }
        }
        total_qrows = qrows;
    }
    OCIHandleFree(stmtp, OCI_HTYPE_STMT);
    free(column_ptr);

    return cols_found;
}


int OCIFillMenuFromQuery(TextWidget* text, HMENU menu, char* query, char* owner, char* table)
{
    char* owner_ptr = NULL, * table_ptr = NULL, * constraint_ptr = NULL, * r_constraint_ptr = NULL;
    int result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    ub4 qrows;
    int nrows, x;
    char query_finished = 0;
    ub4 total_qrows = 0;
    int fetch_array_size = SCHEMA_ARRAY_SIZE;
    char line[100] = "";
    UINT_PTR menuitem;

    if (ChkOCIErrN(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return 1;
    }
    if (ChkOCIErrN(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        owner,
        (ub4)strlen(owner),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        2,
        owner,
        (ub4)strlen(owner),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        3,
        table,
        (ub4)strlen(table),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return 1;
    }
    if (ChkOCIErrN(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        4,
        table,
        (ub4)strlen(table),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return 1;
    }
    owner_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * 129 * sizeof(char));
    if (owner_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return 1;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        owner_ptr,
        129,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(owner_ptr);
        return 1;
    }

    table_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * 129 * sizeof(char));
    if (table_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return 1;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        2,
        table_ptr,
        129,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(owner_ptr);
        free(table_ptr);
        return 1;
    }

    constraint_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * 129 * sizeof(char));
    if (constraint_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return 1;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        3,
        constraint_ptr,
        129,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(owner_ptr);
        free(table_ptr);
        free(constraint_ptr);
        return 1;
    }

    r_constraint_ptr = (char*)malloc(SCHEMA_ARRAY_SIZE * 129 * sizeof(char));
    if (r_constraint_ptr == NULL)
    {
        Error(NULL, "Out of memory");
        return 1;
    }
    if (ChkOCIErrN(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        4,
        r_constraint_ptr,
        129,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(owner_ptr);
        free(table_ptr);
        free(constraint_ptr);
        free(r_constraint_ptr);
        return 1;
    }

    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        free(owner_ptr);
        free(table_ptr);
        free(constraint_ptr);
        free(r_constraint_ptr);
        return 1;
    }
    while (!query_finished)
    {
        result = OCIStmtFetch(g_oci->stmtp,
            g_oci->errhp,
            fetch_array_size,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            if (ChkOCIErrI(result, 1406))
                break;
        }
        if (ChkOCIErrN(OCIAttrGet(g_oci->stmtp,
            OCI_HTYPE_STMT,
            &qrows,
            NULL,
            OCI_ATTR_ROW_COUNT,
            g_oci->errhp)))
        {
            break;
        }
        nrows = qrows - total_qrows;
        if (nrows != fetch_array_size)
            query_finished = 1;

        for (x = 0; x < (int)(qrows - total_qrows); x++)
        {
            int retval;

            text->table_items = (TableItem*)realloc(text->table_items, (text->ntable_items + 1) * sizeof(TableItem));
            text->table_items[text->ntable_items].owner = _strdup(&owner_ptr[x * 129]);
            text->table_items[text->ntable_items].table = _strdup(&table_ptr[x * 129]);
            text->table_items[text->ntable_items].constraint = _strdup(&constraint_ptr[x * 129]);
            text->table_items[text->ntable_items].columns = NULL;
            text->table_items[text->ntable_items].ncolumns = 0;
            text->table_items[text->ntable_items].rcolumns = NULL;
            text->table_items[text->ntable_items].nrcolumns = 0;
            if (text->table_children)
                retval = ColumnsPresent(text, text->table_items[text->ntable_items].owner, &r_constraint_ptr[x * 129], &text->table_items[text->ntable_items], 1);
            else
                retval = ColumnsPresent(text, text->table_items[text->ntable_items].owner, &constraint_ptr[x * 129], &text->table_items[text->ntable_items], 1);
            if (!retval)
            {
                if (text->table_items[text->ntable_items].ncolumns)
                {
                    free(text->table_items[text->ntable_items].columns);
                    text->table_items[text->ntable_items].ncolumns = 0;
                }
            }
            else // add the remote columns without checking
            {
                if (text->table_children)
                    ColumnsPresent(text, text->table_items[text->ntable_items].owner, &constraint_ptr[x * 129], &text->table_items[text->ntable_items], 0);
                else
                    ColumnsPresent(text, text->table_items[text->ntable_items].owner, &r_constraint_ptr[x * 129], &text->table_items[text->ntable_items], 0);
                sprintf(line, "%s.%s (%s)", &owner_ptr[x * 129], &table_ptr[x * 129], &constraint_ptr[x * 129]);
                menuitem = (UINT_PTR)ID_TABLE_BASE + text->ntable_items;
                if (!AppendMenu(menu, MF_ENABLED | MF_STRING, menuitem, line))
                    DisplayLastError(NULL);
                text->table_items[text->ntable_items].item = (UINT)menuitem;
                text->table_items[text->ntable_items].menu = menu;
                text->ntable_items++;
            }
        }
        total_qrows = qrows;
    }
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    free(owner_ptr);
    free(table_ptr);
    free(constraint_ptr);
    free(r_constraint_ptr);
    return 0;
}
