#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jboci.h"


int OCI_DoCommand(char* str, int feedback, char* executestr)
{
    ub4 qrows;

    if (!ora_connected)
    {
        WPrintf("Not connected to a database\n");
        return -1;
    }
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (feedback)
        WPrintfOnly("\rParsing statement");
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        (oratext*)str,
        (ub4)strlen(str),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -2;
    }
    if (OCI_BindVariables(g_text))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -2;
    }
    g_oci->offset = 1;
    if (feedback)
    {
        if (executestr == NULL || executestr[0] == '\0')
            WPrintfOnly("\rExecuting statement");
        else
            WPrintfOnly("\r%s", executestr);
    }
    if (ChkOCIErr(OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -3;
    }
    g_oci->offset = 0;
    ChkOCIErr(OCIAttrGet(g_oci->stmtp,
        OCI_HTYPE_STMT,
        &qrows,
        NULL,
        OCI_ATTR_ROW_COUNT,
        g_oci->errhp));
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return qrows;
}

void OCI_Commit(void)
{
    if (!ora_connected)
    {
        WPrintf("Not connected to a database\n");
        return;
    }
    if (!ChkOCIErr(OCITransCommit(g_oci->svchp,
        g_oci->errhp,
        OCI_DEFAULT)))
    {
        WPrintf("Commit complete\n");
    }

    return;
}

void OCI_Rollback(void)
{
    if (!ora_connected)
    {
        WPrintf("Not connected to a database\n");
        return;
    }
    if (!ChkOCIErr(OCITransRollback(g_oci->svchp,
        g_oci->errhp,
        OCI_DEFAULT)))
    {
        WPrintf("Rollback complete\n");
    }

    return;
}

void GetUserID(char* object, char** user_id, char** object_name, int show_error)
{
    char* str1 = "SELECT owner FROM all_objects WHERE (object_name=:v_object or object_name=upper(:v_object)) and object_type!='SYNONYM' and owner=sys_context( 'userenv', 'current_schema' )";
    char* str2 = "SELECT table_owner,table_name FROM all_synonyms WHERE (synonym_name=:v_object or synonym_name=upper(:v_object)) and owner in (user,'PUBLIC') order by decode(owner,user,0,1)";
    char char_ptr[40], char_ptr2[40];
    int len = 40, result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;

    *user_id = NULL;
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return;
    }
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        str1,
        (ub4)strlen(str1),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErr(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        object,
        (ub4)strlen(object),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErr(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        2,
        object,
        (ub4)strlen(object),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }

    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        ChkOCIErr(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return;
    }
    g_oci->offset = 0;
    if (result == OCI_NO_DATA)
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
            &g_oci->stmtp,
            OCI_HTYPE_STMT,
            0,
            NULL)))
        {
            return;
        }
        if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
            g_oci->errhp,
            str2,
            (ub4)strlen(str2),
            OCI_NTV_SYNTAX,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        if (ChkOCIErr(OCIBindByPos(g_oci->stmtp,
            &bndhp,
            g_oci->errhp,
            1,
            object,
            (ub4)strlen(object),
            SQLT_CHR,
            NULL,
            NULL,
            NULL,
            0,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
            &defhp,
            g_oci->errhp,
            1,
            char_ptr,
            len,
            SQLT_STR,
            NULL,
            NULL,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }

        if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
            &defhp,
            g_oci->errhp,
            2,
            char_ptr2,
            len,
            SQLT_STR,
            NULL,
            NULL,
            NULL,
            OCI_DEFAULT)))
        {
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }

        g_oci->offset = 1;
        result = OCIStmtExecute(g_oci->svchp,
            g_oci->stmtp,
            g_oci->errhp,
            1,
            0,
            NULL,
            NULL,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA)
        {
            ChkOCIErr(result);
            g_oci->offset = 0;
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        else if (result == OCI_NO_DATA)
        {
            if (show_error)
                WPrintf("Object %s does not exist\n", object);
            OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
            g_oci->stmtp = NULL;
            return;
        }
        *object_name = _strdup(char_ptr2);
    }
    else
        *object_name = _strdup(object);
    *user_id = _strdup(char_ptr);
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
}

char* GetOracleObjectType(char* owner, char* object, int show_error)
{
    char* str1 = "SELECT object_type FROM all_objects WHERE owner=upper(:v_owner) AND object_name=upper(:v_object) AND subobject_name IS NULL AND object_type!='PACKAGE BODY'";
    static char char_ptr[40];
    int len = 40, result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;

    char_ptr[0] = '\0';
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return NULL;
    }
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        str1,
        (ub4)strlen(str1),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    if (ChkOCIErr(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        1,
        owner,
        (ub4)strlen(owner),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    if (ChkOCIErr(OCIBindByPos(g_oci->stmtp,
        &bndhp,
        g_oci->errhp,
        2,
        object,
        (ub4)strlen(object),
        SQLT_CHR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }

    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        ChkOCIErr(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    g_oci->offset = 0;
    if (result == OCI_NO_DATA)
    {
        if (show_error)
            WPrintf("Object %s.%s does not exist\n", owner, object);
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return char_ptr;
}

char* GetCurrentUser(void)
{
    char* str1 = "SELECT user FROM dual";
    static char char_ptr[40];
    int len = 40, result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;

    char_ptr[0] = '\0';
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return NULL;
    }
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        str1,
        (ub4)strlen(str1),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        char_ptr,
        len,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        ChkOCIErr(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return NULL;
    }
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return char_ptr;
}

void OCI_AbortQuery(void)
{
    if (tid)
    {
        if (!query_finished && total_qrows)
            query_finished = 1;
        else
        {
            /* this caused a crash, removed the ChkOCIErr */
           /* if (ChkOCIErr(OCIBreak(g_oci->svchp,g_oci->errhp)))
              return;
            if (ChkOCIErr(OCIReset(g_oci->svchp,g_oci->errhp)))
              return;*/
            OCIBreak(g_oci->svchp, g_oci->errhp);
            OCIReset(g_oci->svchp, g_oci->errhp);
        }
        if (g_text->scripts)
            g_text->cancelled = 1;
    }
    else
    {
        RecallCommand(g_text, NULL);
        /*    WPrintf("\n");
            ClearCommand(g_text,0,0);*/
    }
}

int OCI_ExecuteDDLStatement(char* str)
{
    ub4 qrows;

    if (!ora_connected)
    {
        Error(NULL, "Not connected to a database\n");
        return -1;
    }
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        str,
        (ub4)strlen(str),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -2;
    }
    g_oci->offset = 1;
    if (ChkOCIErr(OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -3;
    }
    g_oci->offset = 0;
    ChkOCIErr(OCIAttrGet(g_oci->stmtp,
        OCI_HTYPE_STMT,
        &qrows,
        NULL,
        OCI_ATTR_ROW_COUNT,
        g_oci->errhp));
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return qrows;
}


int GetSingleValue(OCIEnvironment * a_oci, char* query, char* target, int target_size)
{
    int result;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    OCIEnvironment* g_oci = a_oci;

    target[0] = '\0';
    if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
        &g_oci->stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
        g_oci->errhp,
        query,
        (ub4)strlen(query),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -2;
    }
    if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
        &defhp,
        g_oci->errhp,
        1,
        target,
        target_size,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -3;
    }
    g_oci->offset = 1;
    result = OCIStmtExecute(g_oci->svchp,
        g_oci->stmtp,
        g_oci->errhp,
        1,
        0,
        NULL,
        NULL,
        OCI_DEFAULT);
    if (result != OCI_SUCCESS && result != OCI_NO_DATA)
    {
        ChkOCIErr(result);
        g_oci->offset = 0;
        OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
        g_oci->stmtp = NULL;
        return -4;
    }
    g_oci->offset = 0;
    OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
    g_oci->stmtp = NULL;
    return 0;
}

char* OCIGetSysDate(void)
{
    static char sd[100];
    OCIDate osd;
    ub4 len = 100;

    OCIDateSysDate(g_oci->errhp, &osd);
    OCIDateToText(g_oci->errhp, &osd, NULL, 0, NULL, 0, &len, (oratext*)sd);
    return sd;
}

char* OCIGetServerVersion(void)
{
    static char sd[100];
    char* pos;

    OCIServerVersion(g_oci->svchp, g_oci->errhp, (OraText*)sd, 100, OCI_HTYPE_SVCCTX);
    pos = strchr(sd, '\n');
    if (pos)
        * pos = '\0';
    return sd;
}

#define MAJOR_NUMVSN(v) ((sword)(((v) >> 24) & 0x000000FF))      /* version number */ 
#define MINOR_NUMRLS(v) ((sword)(((v) >> 20) & 0x0000000F))      /* release number */
#define UPDATE_NUMUPD(v) ((sword)(((v) >> 12) & 0x000000FF))     /* update number */ 
#define PORT_REL_NUMPRL(v) ((sword)(((v) >> 8) & 0x0000000F))    /* port release number */ 
#define PORT_UPDATE_NUMPUP(v) ((sword)(((v) >> 0) & 0x000000FF)) /* port update number */

char* OCIGetServerRelease(void)
{
    static char sd[100] = "0000000000";
    ub4 v;

    if (MyOCIServerRelease)
    {
        MyOCIServerRelease(g_oci->svchp, g_oci->errhp, (OraText*)sd, 100, OCI_HTYPE_SVCCTX, &v);
        sprintf(sd, "%02d%02d%02d%02d%02d", MAJOR_NUMVSN(v), MINOR_NUMRLS(v), UPDATE_NUMUPD(v), PORT_REL_NUMPRL(v), PORT_UPDATE_NUMPUP(v));
    }
    return sd;
}

