#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jboci.h"
#include <orid.h>
#include "match.h"
#include <fcntl.h>
#include <io.h>

int max_fetch_array_size;
FILE *tmpfile_write = NULL;
FILE *tmpfile_read = NULL;
char *lob_buffer = NULL;
int query_finished = 1, total_qrows = 0, long_column = 0;
XmlSaveDomFunc MyXmlSaveDom = NULL;
char *tmpfile_writebuf = NULL;
char *write_buffer = NULL;
int write_buffer_offset = 0;

#define OBJECT_STR_BUFSIZE 100


int OutputObjectColumn(TextWidget *text, OCIType *tdo, void *instance, void *null_struct, char *buffer, int col_size);
int OutputRefColumn(TextWidget *text, void *ref, char *buffer, int col_size);
int OutputAnyData(TextWidget *text, OCIAnyData *anydata, char *buffer, int col_size);

char *strcat_safe(char *target, char *source, int bufsize)
{
	if (source != NULL)
		return strncat(target, source, __min(strlen(source), bufsize - 1 - strlen(target)));
	else
		return NULL;
}

FILE* fopen_unixlike(const char *file, const char *mode) {
	DWORD dwDesiredAccess = 0;
	DWORD dwCreationDisposition = 0;
	DWORD dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL;
	const DWORD dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE;
	size_t i;

	HANDLE handle;
	int fd;

	/* Infer flags. */
	for (i = 0; mode[i] != '\0'; i++) {
		switch (mode[i]) {
			case 'r':
				dwDesiredAccess |= GENERIC_READ;
				dwCreationDisposition = OPEN_EXISTING;
				break;
			case 'a':
				dwDesiredAccess |= GENERIC_WRITE;
				dwCreationDisposition = CREATE_NEW;
				break;
			case 'w':
				dwDesiredAccess |= GENERIC_WRITE;
				dwCreationDisposition = CREATE_ALWAYS;
				break;
			case 'D':
				dwFlagsAndAttributes |= FILE_FLAG_DELETE_ON_CLOSE;
				break;
			case '+':
				dwDesiredAccess |= GENERIC_READ;
				dwDesiredAccess |= GENERIC_WRITE;
				break;
		}
	}

	/* Create the file. */
	handle = CreateFileA(file,
		dwDesiredAccess,
		dwShareMode,
		NULL,
		dwCreationDisposition,
		dwFlagsAndAttributes,
		NULL);
	if (handle == INVALID_HANDLE_VALUE) {
		return NULL;
	}
    /*
    short cmpression = COMPRESSION_FORMAT_DEFAULT;
    DeviceIoControl(handle, FSCTL_SET_COMPRESSION, &cmpression, sizeof(cmpression), NULL, 0, NULL, NULL);
    */
	/* Associates a C run-time file descriptor with a file HANDLE. */
	fd = _open_osfhandle((intptr_t)handle, _O_BINARY);
	if (fd == -1) {
		CloseHandle(handle);
		return NULL;
	}

	/* Associates a stream with a C run-time file descriptor. */
	return _fdopen(fd, mode);
}

int CreateTMPFile(TextWidget *text)
{
	char tmppath[MAX_PATH], tmpfilename[MAX_PATH];

	CloseTMPFile();
	if (!INT_OPTION(LISTVIEW))
		return 0;
	if (GetTempPath(MAX_PATH, tmppath) == 0)
	{
		DisplayLastError(NULL);
		return -1;
	}
	if (GetTempFileName(tmppath, "jbsql", 0, tmpfilename) == 0)
	{
		DisplayLastError(NULL);
		return -1;
	}

	tmpfile_write = fopen_unixlike(tmpfilename, "wD+");

	if (tmpfile_write == NULL)
	{
		WPrintf("Failed to create tmp file: %s\n", strerror(errno));
		tmpfile_write = NULL;
		return -1;
	}
	if (tmpfile_writebuf == NULL)
	{
		tmpfile_writebuf = (char*)malloc(1024L * 1024L);
		if (tmpfile_writebuf)
			setvbuf(tmpfile_write, tmpfile_writebuf, _IOFBF, 1024L * 1024L);
	}
	tmpfile_read = fopen_unixlike(tmpfilename, "r");

	if (tmpfile_read == NULL)
	{
		WPrintf("Failed to open tmp file: %s\n", strerror(errno));
		tmpfile_write = NULL;
		return -1;
	}
	return 0;
}

void CloseTMPFile(void)
{
	if (tmpfile_read != NULL)
		fclose(tmpfile_read);
	tmpfile_read = NULL;
	if (tmpfile_write != NULL)
		fclose(tmpfile_write);
	tmpfile_write = NULL;
}

void OCI_PrintHeaders(TextWidget *text, OCIColumnDesc *cols, int ncols)
{
	int x;
	char *lines;
	size_t len;

	if (INT_OPTION(HEADING))
	{
		if (INT_OPTION(SCREEN))
		{
			WPrintf("\r");
			for (x = 0; x < (int)ncols; x++)
			{
				len = MAX(cols[x].col_size - 1, strlen(cols[x].col_name));
				if (INT_OPTION(TRIMSPOOL) && x == ncols - 1)
					len = strlen(cols[x].col_name);
				if (x)
					WPrintf(" %-*s", len, cols[x].col_name);
				else
					WPrintf("%-*s", len, cols[x].col_name);
			}
			WPrintfNL();
			for (x = 0; x < (int)ncols; x++)
			{
				if (x)
					WPrintf(" ");
				lines = (char*)malloc(MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name)) + 1);
				memset(lines, '-', MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name)));
				lines[MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name))] = '\0';
				WPrintf("%s", lines);
				free(lines);
			}
			//text->noredraw=0;
			WPrintfNL();
		}
		else if (text->spoolfile != NULL)
		{
			for (x = 0; x < (int)ncols; x++)
			{
				len = MAX(cols[x].col_size - 1, strlen(cols[x].col_name));
				if (INT_OPTION(TRIMSPOOL) && x == ncols - 1)
					len = strlen(cols[x].col_name);
				if (x)
					FPrintf(" %-*s", len, cols[x].col_name);
				else
					FPrintf("%-*s", len, cols[x].col_name);
			}
			FPrintf("\n");
			for (x = 0; x < (int)ncols; x++)
			{
				if (x)
					FPrintf(" ");
				lines = (char*)malloc(MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name)) + 1);
				memset(lines, '-', MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name)));
				lines[MAX(cols[x].col_size - 1, (int)strlen(cols[x].col_name))] = '\0';
				FPrintf("%s", lines);
				free(lines);
			}
			FPrintf("\n");
		}
	}
}

int OCI_PrepareSource(TextWidget *text, char *p_sqlstr, OCIColumnDesc **p_cols, int *p_ncols, char *action)
{
	int           x, y;
	OCIParam      *param;
	ub2           dtype;
	char          *col_name;
	ub4           col_name_len;
	ub4           ncols;
	OCIColumnDesc *pcols;
	ub2           len;
	OCIDefine     *defhp = NULL;
	char          *char_ptr;
	int           record_size = 0;
	int           total_memory;
	int           objects = 0;
	oratext       *type_name;
	oratext       *type_schema_name;

	*p_ncols = 0;
	if (!ora_connected)
	{
		WPrintf("Not connected to a database\n");
		return -1;
	}
	if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp,
		(void**)&g_oci->stmtp,
		OCI_HTYPE_STMT,
		0,
		NULL)))
	{
		return -1;
	}
	if (action[0] != '\0' && INT_OPTION(FEEDBACK))
	{
		HANDLE tmp = text->spoolfile;

		text->spoolfile = NULL;
		WPrintfOnly("\rParsing statement...");
		text->spoolfile = tmp;
	}
	if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
		g_oci->errhp,
		(OraText*)p_sqlstr,
		(ub4)strlen(p_sqlstr),
		OCI_NTV_SYNTAX,
		OCI_DEFAULT)))
	{
		OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
		g_oci->stmtp = NULL;
		return -2;
	}
	if (OCI_BindVariables(g_text))
	{
		OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
		g_oci->stmtp = NULL;
		return -2;
	}

	g_oci->offset = 1;
	if (action[0] != '\0' && INT_OPTION(FEEDBACK))
	{
		HANDLE tmp = text->spoolfile;

		text->spoolfile = NULL;
		WPrintfOnly("\rExecuting query...");
		text->spoolfile = tmp;
	}
	if (ChkOCIErr(OCIStmtExecute(g_oci->svchp,
		g_oci->stmtp,
		g_oci->errhp,
		0,
		0,
		NULL,
		NULL,
		OCI_DEFAULT)))
	{
		g_oci->offset = 0;
		OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
		g_oci->stmtp = NULL;
		return -3;
	}

	g_oci->offset = 0;
	if (ChkOCIErr(OCIAttrGet(g_oci->stmtp,
		OCI_HTYPE_STMT,
		&ncols,
		NULL,
		OCI_ATTR_PARAM_COUNT,
		g_oci->errhp)))
	{
		OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
		g_oci->stmtp = NULL;
		return -4;
	}

	if ((pcols = (OCIColumnDesc*)malloc(sizeof(OCIColumnDesc)*ncols)) == NULL)
	{
		WPrintf("Can't perform malloc!\n");
		return 1;
	}
	for (x = 0; x < (int)ncols; x++)
	{
		if (ChkOCIErr(OCIParamGet(g_oci->stmtp,
			OCI_HTYPE_STMT,
			g_oci->errhp,
			(void**)&param,
			x + 1)))
		{
			break;
		}
		if (ChkOCIErr(OCIAttrGet(param,
			OCI_DTYPE_PARAM,
			&dtype,
			0,
			OCI_ATTR_DATA_TYPE,
			g_oci->errhp)))
		{
			break;
		}

		if (ChkOCIErr(OCIAttrGet(param,
			OCI_DTYPE_PARAM,
			&col_name,
			&col_name_len,
			OCI_ATTR_NAME,
			g_oci->errhp)))
		{
			break;
		}
		pcols[x].col_name = (char*)malloc(col_name_len + 1);
		memcpy(pcols[x].col_name, col_name, col_name_len);
		pcols[x].col_name[col_name_len] = '\0';
		if (ChkOCIErr(OCIAttrGet(param,
			OCI_DTYPE_PARAM,
			&len,
			NULL,
			OCI_ATTR_DATA_SIZE,
			g_oci->errhp)))
		{
			break;
		}
		pcols[x].datatype = dtype;
		switch (dtype)
		{
			case OCI_TYPECODE_DATE:
			case OCI_TYPECODE_TIMESTAMP:
			case OCI_TYPECODE_TIME:
			case OCI_TYPECODE_TIME_TZ:
			case OCI_TYPECODE_TIMESTAMP_TZ:
			case OCI_TYPECODE_TIMESTAMP_LTZ:
			case OCI_TYPECODE_INTERVAL_YM:
			case OCI_TYPECODE_INTERVAL_DS:
				pcols[x].col_size = 40;
				break;
			case OCI_TYPECODE_REAL:
			case OCI_TYPECODE_DOUBLE:
			case OCI_TYPECODE_FLOAT:
			case OCI_TYPECODE_NUMBER:
			case OCI_TYPECODE_DECIMAL:
			case OCI_TYPECODE_OCTET:
			case OCI_TYPECODE_INTEGER:
			case OCI_TYPECODE_SMALLINT:
			case OCI_TYPECODE_BFLOAT:
			case OCI_TYPECODE_BDOUBLE:
				pcols[x].col_size = 42;
				break;
			case OCI_TYPECODE_RAW:
			case OCI_TYPECODE_VARCHAR2:
			case OCI_TYPECODE_VARCHAR:
			case OCI_TYPECODE_CHAR:
				pcols[x].col_size = len + 1;
				break;
			case OCI_TYPECODE_VARRAY:
				WPrintf("Unsupported datatype: VARRAY for column %s\n", pcols[x].col_name);
				return -5;
			case OCI_TYPECODE_TABLE:
				WPrintf("Unsupported datatype: TABLE for column %s\n", pcols[x].col_name);
				return -5;
			case OCI_TYPECODE_CLOB:
			case OCI_TYPECODE_BLOB:
			case OCI_TYPECODE_BFILE:
			case SQLT_LNG:
			case SQLT_BIN:
			case SQLT_LBI:
				pcols[x].col_size = INT_OPTION(LONG_SIZE) + 1;
				break;
			case OCI_TYPECODE_REF:
			case OCI_TYPECODE_OBJECT:
				/*WPrintf("\rUnsupported datatype: OBJECT for column %s\n",pcols[x].col_name);
				return -5;*/
				pcols[x].col_size = INT_OPTION(LONG_SIZE) + 1;
				objects = 1;
				break;
			case OCI_TYPECODE_NAMEDCOLLECTION:
				WPrintf("\rUnsupported datatype: COLLECTION for column %s\n", pcols[x].col_name);
				return -5;
			case SQLT_RID:
			case SQLT_RDD:
				pcols[x].col_size = 40;
				break;
			default:
				WPrintf("\rUnknown datatype %d for column %s\n", dtype, pcols[x].col_name);
				return -5;
		}
		if (pcols[x].col_size < (int)strlen(STR_OPTION(NULLSTR)) + 1)
			pcols[x].col_size = (int)strlen(STR_OPTION(NULLSTR)) + 1;
		record_size += pcols[x].col_size;
	}
	total_memory = INT_OPTION(ARRAYSIZE)*record_size;
	write_buffer = malloc(record_size);
	if (write_buffer == NULL)
	{
		WPrintf("Can't perform malloc!\n");
		return 1;
	}
	if (objects)
		max_fetch_array_size = 1;
	else
	{
		if (total_memory<0 || total_memory>(int)INT_OPTION(MAXMEM))
		{
			max_fetch_array_size = INT_OPTION(MAXMEM) / record_size;
			if (max_fetch_array_size < 1)
				max_fetch_array_size = 1;
		}
		else
			max_fetch_array_size = INT_OPTION(ARRAYSIZE);
	}
	for (x = 0; x < (int)ncols; x++)
	{
		if ((pcols[x].indicator = (short*)malloc(max_fetch_array_size * sizeof(short))) == NULL)
		{
			WPrintf("Can't perform malloc!\n");
			return 1;
		}
		switch (pcols[x].datatype)
		{
			case OCI_TYPECODE_CLOB:
			case OCI_TYPECODE_BLOB:
			case OCI_TYPECODE_BFILE:
				{
					if ((pcols[x].value.lob_value = (OCILobLocator**)malloc(max_fetch_array_size * sizeof(OCILobLocator*))) == NULL)
					{
						WPrintf("Can't perform malloc!\n");
						return 1;
					}
					for (y = 0; y < max_fetch_array_size; y++)
					{
						defhp = NULL;
						if (ChkOCIErr(OCIDescriptorAlloc(g_oci->envhp,
							(void**)&pcols[x].value.lob_value[y],
							OCI_DTYPE_LOB,
							0,
							NULL)))
						{
							return -6;
						}
					}
					if (lob_buffer == NULL)
					{
						lob_buffer = (char*)malloc(pcols[x].col_size);
						if (lob_buffer == NULL)
						{
							WPrintf("Can't perform malloc!\n");
							return -7;
						}
					}
					defhp = NULL;
					if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
						&defhp,
						g_oci->errhp,
						x + 1,
						pcols[x].value.lob_value,
						sizeof(OCILobLocator*),
						pcols[x].datatype,
						pcols[x].indicator,
						NULL,
						NULL,
						OCI_DEFAULT)))
					{
						return -8;
					}
				}
				break;
			case OCI_TYPECODE_OBJECT:
				{
					ub4 type_len = 0, type_schema_len = 0;
					pcols[x].value.object_value = NULL;
					pcols[x].null_struct = NULL;
					pcols[x].buffer = (char*)malloc(pcols[x].col_size + 1);
					if (pcols[x].buffer == NULL)
						return 1;

					defhp = NULL;
					if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
						&defhp,
						g_oci->errhp,
						x + 1,
						NULL,
						0,
						SQLT_NTY,
						NULL,
						NULL,
						NULL,
						OCI_DEFAULT)))
					{
						return 1;
					}
					if (ChkOCIErr(OCIParamGet(g_oci->stmtp,
						OCI_HTYPE_STMT,
						g_oci->errhp,
						(void**)&param,
						x + 1)))
					{
						break;
					}
					if (ChkOCIErr(OCIAttrGet(param,
						OCI_DTYPE_PARAM,
						&type_name,
						(ub4*)&type_len,
						OCI_ATTR_TYPE_NAME,
						g_oci->errhp)))
					{
						return 1;
					}
					type_name[type_len] = '\0';
					if (ChkOCIErr(OCIAttrGet(param,
						OCI_DTYPE_PARAM,
						&type_schema_name,
						(ub4*)&type_schema_len,
						OCI_ATTR_SCHEMA_NAME,
						g_oci->errhp)))
					{
						return 1;
					}
					type_schema_name[type_schema_len] = '\0';
					if (ChkOCIErr(OCITypeByName(g_oci->envhp, g_oci->errhp, g_oci->svchp, (oratext*)type_schema_name, type_schema_len, (oratext*)type_name, type_len, NULL, 0, OCI_DURATION_SESSION, OCI_TYPEGET_ALL, &pcols[x].tdo)))
					{
						return 1;
					}
					if (ChkOCIErr(OCIDefineObject(defhp, g_oci->errhp, pcols[x].tdo, &pcols[x].value.object_value, NULL, &pcols[x].null_struct, NULL)))
						return 6;
				}
				break;
			case OCI_TYPECODE_REF:
				{
					ub4 ref_size;

					pcols[x].value.object_value = NULL;
					pcols[x].null_struct = NULL;
					pcols[x].buffer = (char*)malloc(pcols[x].col_size + 1);
					if (pcols[x].buffer == NULL)
						return 1;

					defhp = NULL;
					if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
						&defhp,
						g_oci->errhp,
						x + 1,
						NULL,
						0,
						SQLT_REF,
						NULL,
						NULL,
						NULL,
						OCI_DEFAULT)))
					{
						return 1;
					}
					if (ChkOCIErr(OCIDefineObject(defhp, g_oci->errhp, NULL, &pcols[x].value.object_value, &ref_size, NULL, NULL)))
						return 6;
				}
				break;
			default:
				{
					if ((pcols[x].value.str_value = (char*)malloc(max_fetch_array_size*(pcols[x].col_size))) == NULL)
					{
						WPrintf("Can't perform malloc!\n");
						return 1;
					}
					char_ptr = pcols[x].value.str_value;
					len = pcols[x].col_size;
					defhp = NULL;
					if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
						&defhp,
						g_oci->errhp,
						x + 1,
						char_ptr,
						len,
						SQLT_STR,
						pcols[x].indicator,
						NULL,
						NULL,
						OCI_DEFAULT)))
					{
						return -9;
					}
				}
				break;
		}
	}

	*p_ncols = ncols;
	*p_cols = pcols;
	//PrintHeaders(text,cols,ncols);
	OCIDescriptorFree(param, OCI_DTYPE_PARAM);

	return 0;
}

void WriteToBuffer(char *buffer, int size)
{
	memcpy(write_buffer + write_buffer_offset, buffer, size);
	write_buffer_offset += size;
}

int OCI_OutputColumn(TextWidget *text, int y, OCIColumnDesc *cols, char *buffer)
{
	if (INT_OPTION(SCREEN))
	{
		size_t len = MAX(cols[y].col_size - 1, strlen(cols[y].col_name));
		if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
		{
			len = strlen(buffer);
		}
		if (y)
			WPrintf(" %-*s", len, buffer);
		else
			WPrintf("\r%-*s", len, buffer);
	}
	else if (text->spoolfile != NULL)
	{
		size_t len = MAX(cols[y].col_size - 1, strlen(cols[y].col_name));
		if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
		{
			len = strlen(buffer);
		}
		if (y)
			FPrintf(" %-*s", len, buffer);
		else
			FPrintf("\r%-*s", len, buffer);
	}
	if (INT_OPTION(LISTVIEW))
	{
		WriteToBuffer(buffer, cols[y].col_size);
	}
	return 0;
}

int OCI_OutputLOBColumn(TextWidget *text, int y, OCIColumnDesc *cols, char *buffer, size_t totsize, int bufsize, int last)
{
	if (totsize == 0 && last)
	{
		if (INT_OPTION(SCREEN))
		{
			size_t len = MAX((cols[y].col_size - 1), strlen(cols[y].col_name));
			if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
				len = strlen(buffer);
			if (y)
				WPrintf(" %-*s", len, buffer);
			else
				WPrintf("\r%-*s", len, buffer);
		}
		else if (text->spoolfile != NULL)
		{
			size_t len = MAX((cols[y].col_size - 1), strlen(cols[y].col_name));
			if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
				len = strlen(buffer);
			if (y)
				FPrintf(" %-*s", len, buffer);
			else
				FPrintf("\r%-*s", len, buffer);
		}
		if (INT_OPTION(LISTVIEW))
		{
			if (bufsize)
			{
				WriteToBuffer(buffer, bufsize);
			}
		}
	}
	else if (totsize == 0 && !last)
	{
		if (INT_OPTION(SCREEN))
		{
			if (y)
				WPrintf(" %s", buffer);
			else
				WPrintf("\r%s", buffer);
		}
		else if (text->spoolfile != NULL)
		{
			if (y)
				FPrintf(" %s", buffer);
			else
				FPrintf("\r%s", buffer);
		}
		if (INT_OPTION(LISTVIEW))
		{
			WriteToBuffer(buffer, bufsize);
		}
	}
	else if (!last)
	{
		if (INT_OPTION(SCREEN))
		{
			if (y)
				WPrintf("%s", buffer);
			else
				WPrintf("%s", buffer);
		}
		else if (text->spoolfile != NULL)
		{
			if (y)
				FPrintf("%s", buffer);
			else
				FPrintf("%s", buffer);
		}
		if (INT_OPTION(LISTVIEW))
		{
			WriteToBuffer(buffer, bufsize);
		}
	}
	else
	{
		if (INT_OPTION(SCREEN))
		{
			size_t len = (cols[y].col_size - 1) - totsize;
			if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
				len = strlen(buffer);
			if (y)
				WPrintf("%-*s", len, buffer);
			else
				WPrintf("%-*s", len, buffer);
		}
		else if (text->spoolfile != NULL)
		{
			size_t len = (cols[y].col_size - 1) - totsize;
			if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
				len = strlen(buffer);
			if (y)
				FPrintf("%-*s", len, buffer);
			else
				FPrintf("%-*s", len, buffer);
		}
		if (INT_OPTION(LISTVIEW))
		{
			if (bufsize)
				WriteToBuffer(buffer, bufsize);
		}
	}
	return 0;
}

int OutputXMLAttribute(xmldocnode* instance, char *buffer, int col_size)
{
	xmlctx *xctx;
	xmlerr xerr;
	size_t len;
	ubig_ora bytes_written, required_size;
	char *bigbuffer;

	strcat_safe(buffer, "'", col_size);
	xctx = OCIXmlDbInitXmlCtx(g_oci->envhp, g_oci->svchp, g_oci->errhp, NULL, 0);
	len = strlen(buffer);
	required_size = MyXmlSaveDom(xctx, &xerr, instance, NULL);
	if (required_size)
	{
		bigbuffer = malloc(required_size + 1);
		if (bigbuffer)
		{
			bytes_written = MyXmlSaveDom(xctx, &xerr, instance, "buffer", bigbuffer, "buffer_length", required_size + 1, NULL);
			if (bytes_written > 0)
			{
				if (bytes_written > col_size - len)
					bigbuffer[col_size - len] = '\0';
				else
					bigbuffer[bytes_written] = '\0';
				strcat_safe(buffer, bigbuffer, col_size);
			}
			free(bigbuffer);
		}
	}
	OCIXmlDbFreeXmlCtx(xctx);
	strcat_safe(buffer, "'", col_size);
	return 0;
}

int OutputAttribute(TextWidget *text, void *value, void *null_struct, OCITypeCode typecode, OCIType *tdo_attr, char *buffer, int col_size)
{
	ub4 amount, offset;
	ub4 value_size = OBJECT_STR_BUFSIZE;
	ub2 value_size2 = OBJECT_STR_BUFSIZE;
	oratext str_value[OBJECT_STR_BUFSIZE];
	size_t interval_size;
	sword ret;
	char *ptr;
	ub4 ptrsize;

	switch (typecode)
	{
		case OCI_TYPECODE_RAW:
			strcat_safe(buffer, "'", col_size);
			ptrsize = OCIRawSize(g_oci->envhp, *(OCIRaw**)value);
			ptr = (char*)malloc(ptrsize + 1);
			memcpy(ptr, OCIRawPtr(g_oci->envhp, *(OCIRaw **)value), ptrsize);
			ptr[ptrsize] = '\0';
			strcat_safe(buffer, ptr, col_size);
			strcat_safe(buffer, "'", col_size);
			free(ptr);
			break;
		case OCI_TYPECODE_VARCHAR2:
		case OCI_TYPECODE_NVARCHAR2:
		case OCI_TYPECODE_VARCHAR:
		case OCI_TYPECODE_CHAR:
		case OCI_TYPECODE_NCHAR:
			strcat_safe(buffer, "'", col_size);
			strcat_safe(buffer, (char*)OCIStringPtr(g_oci->envhp, *(OCIString **)value), col_size);
			strcat_safe(buffer, "'", col_size);
			break;
		case OCI_TYPECODE_REAL:
		case OCI_TYPECODE_DOUBLE:
		case OCI_TYPECODE_FLOAT:
		case OCI_TYPECODE_NUMBER:
		case OCI_TYPECODE_DECIMAL:
		case OCI_TYPECODE_OCTET:
		case OCI_TYPECODE_INTEGER:
			str_value[0] = '\0';
			value_size = OBJECT_STR_BUFSIZE;
			ChkOCIErr(OCINumberToText(g_oci->errhp, (OCINumber *)value, "TM9", 3, NULL, 0, &value_size, str_value));
			strcat_safe(buffer, (char*)str_value, col_size);
			break;
		case OCI_TYPECODE_BDOUBLE:
			sprintf((char*)str_value, "%lf", *(double*)value);
			strcat_safe(buffer, (char*)str_value, col_size);
			break;
		case OCI_TYPECODE_BFLOAT:
			sprintf((char*)str_value, "%f", *(float*)value);
			strcat_safe(buffer, (char*)str_value, col_size);
			break;
		case OCI_TYPECODE_DATE:
			str_value[0] = '\0';
			value_size = OBJECT_STR_BUFSIZE;
			ChkOCIErr(OCIDateToText(g_oci->errhp, (OCIDate *)value, NULL, 0, NULL, 0, &value_size, str_value));
			strcat_safe(buffer, "'", col_size);
			strcat_safe(buffer, (char*)str_value, col_size);
			strcat_safe(buffer, "'", col_size);
			break;
		case OCI_TYPECODE_TIME:
		case OCI_TYPECODE_TIME_TZ:
		case OCI_TYPECODE_TIMESTAMP:
		case OCI_TYPECODE_TIMESTAMP_TZ:
		case OCI_TYPECODE_TIMESTAMP_LTZ:
			str_value[0] = '\0';
			value_size = OBJECT_STR_BUFSIZE;
			ChkOCIErr(OCIDateTimeToText(g_oci->usrhp, g_oci->errhp, *(OCIDateTime **)value, NULL, 0, 6, NULL, 0, &value_size, str_value));
			strcat_safe(buffer, "'", col_size);
			strcat_safe(buffer, (char*)str_value, col_size);
			strcat_safe(buffer, "'", col_size);
			break;
		case OCI_TYPECODE_INTERVAL_YM:
		case OCI_TYPECODE_INTERVAL_DS:
			str_value[0] = '\0';
			value_size = OBJECT_STR_BUFSIZE;
			ChkOCIErr(OCIIntervalToText(g_oci->envhp, g_oci->errhp, *(OCIInterval **)value, 0, 0, str_value, value_size, &interval_size));
			strcat_safe(buffer, "'", col_size);
			strcat_safe(buffer, (char*)str_value, col_size);
			strcat_safe(buffer, "'", col_size);
			break;
		case OCI_TYPECODE_NCLOB:
		case OCI_TYPECODE_CLOB:
		case OCI_TYPECODE_BLOB:
		case OCI_TYPECODE_BFILE:
			if (lob_buffer == NULL)
			{
				lob_buffer = (char*)malloc(INT_OPTION(LONG_SIZE) < 4000 ? INT_OPTION(LONG_SIZE) + 1 : 4001);
				if (lob_buffer == NULL)
				{
					WPrintf("Can't perform malloc!\n");
					return -7;
				}
			}
			if (typecode == OCI_TYPECODE_BFILE)
			{
				if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
					g_oci->errhp,
					*(OCILobLocator **)value,
					OCI_FILE_READONLY)))
				{
					return 1;
				}
			}
			strcat_safe(buffer, "'", col_size);
			offset = 1;
			do
			{
				amount = INT_OPTION(LONG_SIZE);
				ret = OCILobRead(g_oci->svchp,
					g_oci->errhp,
					*(OCILobLocator **)value,
					&amount,
					offset,
					lob_buffer,
					INT_OPTION(LONG_SIZE) < 4000 ? INT_OPTION(LONG_SIZE) : 4000,
					0,
					NULL,
					0,
					(typecode == OCI_TYPECODE_NCLOB) ? SQLCS_NCHAR : SQLCS_IMPLICIT);
				if (ret != OCI_SUCCESS && ret != OCI_NEED_DATA)
				{
					if (ChkOCIErr(ret))
					{
						return 1;
					}
				}
				lob_buffer[amount] = '\0';
				strcat_safe(buffer, lob_buffer, col_size);
				offset += amount;
			} while (ret == OCI_NEED_DATA && offset < (ub4)INT_OPTION(LONG_SIZE));
			strcat_safe(buffer, "'", col_size);
			if (typecode == OCI_TYPECODE_BFILE)
			{
				if (ChkOCIErr(OCILobFileClose(g_oci->svchp,
					g_oci->errhp,
					*(OCILobLocator **)value)))
				{
					return 1;
				}
			}
			else if (ret == OCI_NEED_DATA)
			{
				if (ChkOCIErr(OCILobClose(g_oci->svchp,
					g_oci->errhp,
					*(OCILobLocator **)value)))
				{
					return 1;
				}
			}
			break;
		case OCI_TYPECODE_OBJECT:
			if (OutputObjectColumn(text, tdo_attr, value, null_struct, buffer, col_size))
				return 1;
			break;
		case OCI_TYPECODE_OPAQUE:
			{
				char *type_name = NULL, *type_schema_name = NULL;
				ub4 type_len = 0, type_schema_len = 0;
				OCIDescribe *handle = NULL;
				void *parmh;

				if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp, (void**)&handle, OCI_HTYPE_DESCRIBE, 0, NULL)))
					return 1;
				if (ChkOCIErr(OCIDescribeAny(g_oci->svchp, g_oci->errhp, tdo_attr, 0, OCI_OTYPE_PTR, OCI_DEFAULT, OCI_PTYPE_TYPE, handle)))
					return 1;
				if (OCIAttrGet(handle, OCI_HTYPE_DESCRIBE, &parmh, 0, OCI_ATTR_PARAM, g_oci->errhp))
					return 1;
				if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &type_schema_name, &type_schema_len, OCI_ATTR_SCHEMA_NAME, g_oci->errhp))
					return 1;
				if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &type_name, &type_len, OCI_ATTR_NAME, g_oci->errhp))
					return 1;
				type_schema_name[type_schema_len] = '\0';
				type_name[type_len] = '\0';
				strcat_safe(buffer, type_schema_name, col_size);
				strcat_safe(buffer, ".", col_size);
				strcat_safe(buffer, type_name, col_size);
				strcat_safe(buffer, "(", col_size);
				if (!strcmp(type_schema_name, "SYS") && !strcmp(type_name, "ANYDATA"))
				{
					if (OutputAnyData(text, *(OCIAnyData**)value, buffer, col_size))
						return 1;
				}
				else if (!strcmp(type_schema_name, "SYS") && !strcmp(type_name, "XMLTYPE"))
				{
					OutputXMLAttribute(*(xmldocnode**)value, buffer, col_size);
				}
				strcat_safe(buffer, ")", col_size);
				if (ChkOCIErr(OCIHandleFree((void*)handle, OCI_HTYPE_DESCRIBE)))
					return 1;
			}
			break;
		case OCI_TYPECODE_NAMEDCOLLECTION:
			if (OutputObjectColumn(text, tdo_attr, *(OCITable**)value, null_struct, buffer, col_size))
				return 1;
			break;
		case OCI_TYPECODE_REF:
			OutputRefColumn(text, *(OCIRef**)value, buffer, col_size);
			break;
		case OCI_TYPECODE_UROWID:
			if (ChkOCIErr(OCIRowidToChar(*(OCIRowid**)value, str_value, &value_size2, g_oci->errhp)))
				return 1;
			str_value[value_size2] = '\0';
			strcat_safe(buffer, "'", col_size);
			strcat_safe(buffer, (char*)str_value, col_size);
			strcat_safe(buffer, "'", col_size);
	}
	return 0;
}

int OutputAnyDataValue(TextWidget *text, OCIAnyData *anydata, void *value, OCIType *tdo, OCITypeCode typecode, char *buffer, int col_size, int by_ref)
{
	ub4 attrlen;
	OCIInd indicator;

	if (ChkOCIErr(OCIAnyDataAccess(g_oci->svchp, g_oci->errhp, anydata, typecode, tdo, &indicator, &value, &attrlen)))
		return 1;
	if (indicator == OCI_IND_NULL)
		strcat_safe(buffer, "NULL", col_size);
	else
	{
		if (OutputAttribute(text, by_ref ? &value : value, NULL, typecode, tdo, buffer, col_size))
			return 1;
	}
	return 0;
}


int OutputAnyData(TextWidget *text, OCIAnyData *anydata, char *buffer, int col_size)
{
	OCITypeCode typecode, table_typecode;
	OCIType *tdo = NULL;
	void *object_value = NULL, *object_null = NULL;
	ub4 attrlen = 0;
	OCIString *string_value = NULL;
	OCIRaw *raw_value = NULL;
	OCINumber number_value;
	OCIDate date_value;
	OCIDateTime *datetime_value = NULL;
	OCILobLocator *lob_value = NULL;
	OCIDescribe *handle = NULL;
	OCIInterval *interval_value = NULL;
	OCIRowid *rowid_value = NULL;
	void *parmh;
	boolean isnull;

	if (ChkOCIErr(OCIAnyDataGetType(g_oci->svchp, g_oci->errhp, anydata, &typecode, &tdo)))
		return 1;
	if (ChkOCIErr(OCIAnyDataIsNull(g_oci->svchp, g_oci->errhp, anydata, &isnull)))
		return 1;
	if (isnull)
	{
		strcat_safe(buffer, "NULL", col_size);
	}
	else
	{
		switch (typecode)
		{
			case OCI_TYPECODE_OBJECT:
				if (ChkOCIErr(OCIObjectNew(g_oci->envhp, g_oci->errhp, g_oci->svchp, OCI_TYPECODE_OBJECT, tdo, NULL, OCI_DURATION_DEFAULT, TRUE, (void**)&object_value)))
					return 1;
				if (ChkOCIErr(OCIAnyDataAccess(g_oci->svchp, g_oci->errhp, anydata, typecode, tdo, &object_null, &object_value, &attrlen)))
					return 1;
				if (OutputAttribute(text, object_value, object_null, typecode, tdo, buffer, col_size))
					return 1;
				if (ChkOCIErr(OCIObjectFree(g_oci->envhp, g_oci->errhp, object_value, 0)))
					return 1;
				break;
			case OCI_TYPECODE_REF:
				if (ChkOCIErr(OCIObjectNew(g_oci->envhp, g_oci->errhp, g_oci->svchp, OCI_TYPECODE_REF, tdo, NULL, OCI_DURATION_DEFAULT, TRUE, (void**)&object_value)))
					return 1;
				if (OutputAnyDataValue(text, anydata, object_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				if (ChkOCIErr(OCIObjectFree(g_oci->envhp, g_oci->errhp, object_value, 0)))
					return 1;
				break;
			case OCI_TYPECODE_NAMEDCOLLECTION:
				if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp, (void**)&handle, OCI_HTYPE_DESCRIBE, 0, NULL)))
					return 1;
				if (ChkOCIErr(OCIDescribeAny(g_oci->svchp, g_oci->errhp, tdo, 0, OCI_OTYPE_PTR, OCI_DEFAULT, OCI_PTYPE_TYPE, handle)))
					return 1;
				if (OCIAttrGet(handle, OCI_HTYPE_DESCRIBE, &parmh, 0, OCI_ATTR_PARAM, g_oci->errhp))
					return 1;
				if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &table_typecode, 0, OCI_ATTR_COLLECTION_TYPECODE, g_oci->errhp))
					return 1;
				if (ChkOCIErr(OCIObjectNew(g_oci->envhp, g_oci->errhp, g_oci->svchp, table_typecode, tdo, NULL, OCI_DURATION_DEFAULT, TRUE, (void**)&object_value)))
					return 1;
				if (OutputAnyDataValue(text, anydata, object_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				if (ChkOCIErr(OCIObjectFree(g_oci->envhp, g_oci->errhp, object_value, 0)))
					return 1;
				if (ChkOCIErr(OCIHandleFree(handle, OCI_HTYPE_DESCRIBE)))
					return 1;
				break;
			case OCI_TYPECODE_VARCHAR2:
			case OCI_TYPECODE_VARCHAR:
			case OCI_TYPECODE_CHAR:
			case OCI_TYPECODE_NVARCHAR2:
			case OCI_TYPECODE_NCHAR:
				if (OutputAnyDataValue(text, anydata, string_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_RAW:
				if (OutputAnyDataValue(text, anydata, raw_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_REAL:
			case OCI_TYPECODE_DOUBLE:
			case OCI_TYPECODE_BDOUBLE:
			case OCI_TYPECODE_BFLOAT:
			case OCI_TYPECODE_FLOAT:
			case OCI_TYPECODE_NUMBER:
			case OCI_TYPECODE_DECIMAL:
			case OCI_TYPECODE_OCTET:
			case OCI_TYPECODE_INTEGER:
			case OCI_TYPECODE_SMALLINT:
				OCINumberSetZero(g_oci->errhp, &number_value);
				if (OutputAnyDataValue(text, anydata, &number_value, tdo, typecode, buffer, col_size, 0))
					return 1;
				break;
			case OCI_TYPECODE_DATE:
				if (OutputAnyDataValue(text, anydata, &date_value, tdo, typecode, buffer, col_size, 0))
					return 1;
				break;
			case OCI_TYPECODE_TIME:
			case OCI_TYPECODE_TIME_TZ:
			case OCI_TYPECODE_TIMESTAMP:
			case OCI_TYPECODE_TIMESTAMP_TZ:
			case OCI_TYPECODE_TIMESTAMP_LTZ:
				if (OutputAnyDataValue(text, anydata, datetime_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_CLOB:
			case OCI_TYPECODE_NCLOB:
			case OCI_TYPECODE_BLOB:
			case OCI_TYPECODE_BFILE:
				if (OutputAnyDataValue(text, anydata, lob_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_INTERVAL_DS:
				if (OutputAnyDataValue(text, anydata, interval_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_INTERVAL_YM:
				if (OutputAnyDataValue(text, anydata, interval_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
			case OCI_TYPECODE_UROWID:
				if (OutputAnyDataValue(text, anydata, rowid_value, tdo, typecode, buffer, col_size, 1))
					return 1;
				break;
		}
	}
	return 0;
}

int OutputObjectColumn(TextWidget *text, OCIType *tdo, void *instance, void *null_struct, char *buffer, int col_size)
{
	OCIDescribe *handle;
	OCIParam *attrlsthd, *attrhd, *parmh, *col_element;
	ub2 n_attributes, x;
	OCITypeCode typecode;
	oratext *attrname = NULL;
	ub4 attrlen = 0;
	OCIInd null_status = -1;
	OCIType *tdo_attr = NULL;
	void *value = NULL;
	char *type_name, *type_schema_name;
	ub4 type_len = 0, type_schema_len = 0;
	void *attr_null_struct = NULL;
	OCIRef *tdo_ref;
	sb4 index;
	void *null_element = NULL;
	int exists;
	int first = 1;
	sb4 table_size;
	OCIIter *itr;
	int eoc;

	if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp, (void**)&handle, OCI_HTYPE_DESCRIBE, 0, NULL)))
		return 1;
	if (ChkOCIErr(OCIDescribeAny(g_oci->svchp, g_oci->errhp, tdo, 0, OCI_OTYPE_PTR, OCI_DEFAULT, OCI_PTYPE_TYPE, handle)))
		return 1;
	if (OCIAttrGet(handle, OCI_HTYPE_DESCRIBE, &parmh, 0, OCI_ATTR_PARAM, g_oci->errhp))
		return 1;
	if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &type_schema_name, &type_schema_len, OCI_ATTR_SCHEMA_NAME, g_oci->errhp))
		return 1;
	if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &type_name, &type_len, OCI_ATTR_NAME, g_oci->errhp))
		return 1;
	type_schema_name[type_schema_len] = '\0';
	type_name[type_len] = '\0';
	strcat_safe(buffer, type_schema_name, col_size);
	strcat_safe(buffer, ".", col_size);
	strcat_safe(buffer, type_name, col_size);
	strcat_safe(buffer, "(", col_size);
	if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &typecode, 0, OCI_ATTR_TYPECODE, g_oci->errhp))
		return 1;
	switch (typecode)
	{
		case OCI_TYPECODE_OPAQUE:
			if (!strcmp(type_schema_name, "SYS") && !strcmp(type_name, "XMLTYPE") && MyXmlSaveDom != NULL)
				OutputXMLAttribute((xmldocnode*)instance, buffer, col_size);
			else if (!strcmp(type_schema_name, "SYS") && !strcmp(type_name, "ANYDATA"))
			{
				if (OutputAnyData(text, (OCIAnyData*)instance, buffer, col_size))
					return 1;
			}
			break;
		case OCI_TYPECODE_OBJECT:
			if (ChkOCIErr(OCIAttrGet(parmh, OCI_HTYPE_DESCRIBE, (void*)&n_attributes, NULL, OCI_ATTR_NUM_TYPE_ATTRS, g_oci->errhp)))
				return 1;
			if (ChkOCIErr(OCIAttrGet(parmh, OCI_HTYPE_DESCRIBE, (dvoid *)&attrlsthd, 0, OCI_ATTR_LIST_TYPE_ATTRS, g_oci->errhp)))
				return 1;
			for (x = 1; x <= n_attributes; x++)
			{
				if (ChkOCIErr(OCIParamGet(attrlsthd, OCI_DTYPE_PARAM, g_oci->errhp, (void**)&attrhd, x)))
					return 1;
				if (ChkOCIErr(OCIAttrGet(attrhd, OCI_DTYPE_PARAM, &typecode, 0, OCI_ATTR_TYPECODE, g_oci->errhp)))
					return 1;
				if (ChkOCIErr(OCIAttrGet(attrhd, OCI_DTYPE_PARAM, &attrname, &attrlen, OCI_ATTR_NAME, g_oci->errhp)))
					return 1;
				attrname[attrlen] = '\0';
				if (ChkOCIErr(OCIObjectGetAttr(g_oci->envhp, g_oci->errhp, instance, null_struct, tdo, &attrname, &attrlen, 1, NULL, 0, &null_status, &attr_null_struct, &value, &tdo_attr)))
					return 1;
				if (x > 1)
					strcat_safe(buffer, ",", col_size);
				if (null_status == OCI_IND_NULL)
					strcat_safe(buffer, "NULL", col_size);
				else
				{
					if (OutputAttribute(text, value, attr_null_struct, typecode, tdo_attr, buffer, col_size))
						return 1;
				}
			}
			break;
		case OCI_TYPECODE_NAMEDCOLLECTION:
			if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &typecode, 0, OCI_ATTR_COLLECTION_TYPECODE, g_oci->errhp))
				return 1;
			switch (typecode)
			{
				case OCI_TYPECODE_TABLE:
					if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &col_element, 0, OCI_ATTR_COLLECTION_ELEMENT, g_oci->errhp))
						return 1;

					if (OCIAttrGet(col_element, OCI_DTYPE_PARAM, &tdo_ref, 0, OCI_ATTR_REF_TDO, g_oci->errhp))
						return 1;

					if (ChkOCIErr(OCITypeByRef(g_oci->envhp, g_oci->errhp, tdo_ref, OCI_DURATION_SESSION, OCI_TYPEGET_HEADER, &tdo_attr)))
						return 1;

					if (ChkOCIErr(OCIAttrGet((dvoid*)col_element, OCI_DTYPE_PARAM, &typecode, NULL, OCI_ATTR_TYPECODE, g_oci->errhp)))
						return 1;

					if (ChkOCIErr(OCITableSize(g_oci->envhp, g_oci->errhp, (OCITable*)instance, &table_size)))
						return 1;

					if (table_size)
					{
						if (ChkOCIErr(OCITableFirst(g_oci->envhp, g_oci->errhp, (OCITable*)instance, &index)))
							return 1;

						for (;;)
						{
							if (first)
								first = 0;
							else
								strcat_safe(buffer, ",", col_size);

							if (ChkOCIErr(OCICollGetElem(g_oci->envhp, g_oci->errhp, (OCITable*)instance, index, &exists, &value, &null_element)))
								return 1;
							if (!exists)
								break;
							if (*(OCIInd*)null_element == OCI_IND_NULL)
								strcat_safe(buffer, "NULL", col_size);
							else
							{
								if (OutputAttribute(text, value, null_element, typecode, tdo_attr, buffer, col_size))
									return 1;
							}
							if (OCITableNext(g_oci->envhp, g_oci->errhp, index, (OCITable*)instance, &index, &exists) || !exists)
								break;
						}
					}
					break;
				case OCI_TYPECODE_VARRAY:
					if (OCIAttrGet(parmh, OCI_DTYPE_PARAM, &col_element, 0, OCI_ATTR_COLLECTION_ELEMENT, g_oci->errhp))
						return 1;

					if (ChkOCIErr(OCIAttrGet((dvoid*)col_element, OCI_DTYPE_PARAM, &typecode, NULL, OCI_ATTR_TYPECODE, g_oci->errhp)))
						return 1;

					if (OCIAttrGet(col_element, OCI_DTYPE_PARAM, &tdo_ref, 0, OCI_ATTR_REF_TDO, g_oci->errhp))
						return 1;

					if (ChkOCIErr(OCITypeByRef(g_oci->envhp, g_oci->errhp, tdo_ref, OCI_DURATION_SESSION, OCI_TYPEGET_HEADER, &tdo_attr)))
						return 1;

					if (ChkOCIErr(OCIIterCreate(g_oci->envhp, g_oci->errhp, (CONST OCIColl*) instance, &itr)))
						return 1;

					for (eoc = 0; !OCIIterNext(g_oci->envhp, g_oci->errhp, itr, (dvoid **)&value, (dvoid **)&null_element, &eoc) && !eoc;)
					{
						if (first)
							first = 0;
						else
							strcat_safe(buffer, ",", col_size);

						if (*(OCIInd*)null_element == OCI_IND_NULL)
							strcat_safe(buffer, "NULL", col_size);
						else
						{
							if (OutputAttribute(text, value, null_element, typecode, tdo_attr, buffer, col_size))
								return 1;
						}
					}
					break;
			}
			break;
	}
	strcat_safe(buffer, ")", col_size);
	if (ChkOCIErr(OCIHandleFree(handle, OCI_HTYPE_DESCRIBE)))
		return 1;
	return 0;
}

int OutputRefColumn(TextWidget *text, void *ref, char *buffer, int col_size)
{
	void *object, *null_object;
	OCIRef *type_ref;
	OCIType *tdo;

	if (ChkOCIErr(OCIObjectPin(g_oci->envhp, g_oci->errhp, (OCIRef *)ref, (OCIComplexObject *)0, OCI_PIN_ANY, OCI_DURATION_SESSION, OCI_LOCK_NONE, (dvoid **)&object)))
		return 1;
	if (ChkOCIErr(OCIObjectNew(g_oci->envhp, g_oci->errhp, g_oci->svchp, OCI_TYPECODE_REF, NULL, NULL, OCI_DURATION_DEFAULT, TRUE, (void**)&type_ref)))
		return 1;
	if (ChkOCIErr(OCIObjectGetTypeRef(g_oci->envhp, g_oci->errhp, object, type_ref)))
		return 1;
	if (ChkOCIErr(OCIObjectPin(g_oci->envhp, g_oci->errhp, type_ref, NULL, OCI_PIN_ANY, OCI_DURATION_SESSION, OCI_LOCK_NONE, (void**)&tdo)))
		return 1;
	if (ChkOCIErr(OCIObjectGetInd(g_oci->envhp, g_oci->errhp, object, &null_object)))
		return 1;
	if (OutputObjectColumn(text, tdo, object, null_object, buffer, col_size))
		return 1;
	if (OCIObjectFree(g_oci->envhp, g_oci->errhp, object, 0))
		return 1;
	return 0;
}

int OCI_FetchData(TextWidget *text, OCIColumnDesc *cols, int ncols, char *action)
{
	ub4 qrows, amount, offset;
	int result, nrows, x, y, fetch_array_size;
	sword ret;

	query_finished = 0;
	total_qrows = 0;
	while (!query_finished)
	{
		if (INT_OPTION(FAST_RESULTS) && total_qrows < (int)INT_OPTION(FAST_RESULTS))
			fetch_array_size = 1;
		else
			fetch_array_size = max_fetch_array_size;
		result = OCIStmtFetch(g_oci->stmtp,
			g_oci->errhp,
			fetch_array_size,
			OCI_FETCH_NEXT,
			OCI_DEFAULT);
		if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
		{
			if (ChkOCIErrI(result, 1406))
				break;
		}
		if (ChkOCIErr(OCIAttrGet(g_oci->stmtp,
			OCI_HTYPE_STMT,
			&qrows,
			NULL,
			OCI_ATTR_ROW_COUNT,
			g_oci->errhp)))
		{
			break;
		}


		nrows = qrows - total_qrows;
		if (total_qrows == 0 && qrows)
			OCI_PrintHeaders(text, cols, ncols);
		if (nrows != fetch_array_size)
			query_finished = 1;


		for (x = 0; x < nrows; x++)
		{
			write_buffer_offset = 0;
			for (y = 0; y < ncols; y++)
			{
				switch (cols[y].datatype)
				{
					case OCI_TYPECODE_CLOB:
					case OCI_TYPECODE_BLOB:
					case OCI_TYPECODE_BFILE:
						if (cols[y].indicator[x] == -1)
						{
							strcpy(lob_buffer, STR_OPTION(NULLSTR));
//							memset(lob_buffer, 0, INT_OPTION(LONG_SIZE) < 4000 ? INT_OPTION(LONG_SIZE) : 4000);
							if (OCI_OutputLOBColumn(text, y, cols, lob_buffer, strlen(lob_buffer), INT_OPTION(LONG_SIZE)+1, 1))
								return total_qrows + x;
						}
						else
						{
							if (cols[y].datatype == OCI_TYPECODE_BFILE)
							{
								if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
									g_oci->errhp,
									cols[y].value.lob_value[x],
									OCI_FILE_READONLY)))
								{
									return total_qrows + x;
								}
							}
							offset = 1;
							amount = (INT_OPTION(LONG_SIZE) < cols[y].col_size - 1) ? (INT_OPTION(LONG_SIZE)) : (cols[y].col_size - 1);
							Log("Reading LOB, amount=%d, offset=%d\n", amount, offset);
							ret = OCILobRead(g_oci->svchp,
								g_oci->errhp,
								cols[y].value.lob_value[x],
								&amount,
								offset,
								lob_buffer,
								cols[y].col_size - 1,
								0,
								NULL,
								0,
								SQLCS_IMPLICIT);
							Log("Read %d bytes from lob\n", amount);
							lob_buffer[amount] = '\0';
							if (OCI_OutputLOBColumn(text, y, cols, lob_buffer, amount, cols[y].col_size, 1))
								return total_qrows + x;
							if (cols[y].datatype == OCI_TYPECODE_BFILE)
							{
								if (ChkOCIErr(OCILobFileClose(g_oci->svchp,
									g_oci->errhp,
									cols[y].value.lob_value[x])))
								{
									return total_qrows + y;
								}
							}
							else if (ret == OCI_NEED_DATA)
							{
								if (ChkOCIErr(OCILobClose(g_oci->svchp,
									g_oci->errhp,
									cols[y].value.lob_value[x])))
								{
									return total_qrows + y;
								}
							}
							
						}
						break;
					case OCI_TYPECODE_OBJECT:
						cols[y].buffer[0] = '\0';
						if (cols[y].value.object_value != NULL)
						{
							if (OutputObjectColumn(text, cols[y].tdo, cols[y].value.object_value, cols[y].null_struct, cols[y].buffer, cols[y].col_size))
								return total_qrows;
							if (ChkOCIErr(OCIObjectFree(g_oci->envhp, g_oci->errhp, cols[y].value.object_value, 0)))
								return total_qrows;
							cols[y].value.object_value = NULL;
							cols[y].null_struct = NULL;
						}
						else
							strcpy(cols[y].buffer, STR_OPTION(NULLSTR));
						if (OCI_OutputColumn(text, y, cols, cols[y].buffer))
							return total_qrows;
						break;
					case OCI_TYPECODE_REF:
						cols[y].buffer[0] = '\0';
						if (cols[y].value.object_value != NULL)
						{
							if (OutputRefColumn(text, cols[y].value.object_value, cols[y].buffer, cols[y].col_size))
								return total_qrows;
							if (ChkOCIErr(OCIObjectFree(g_oci->envhp, g_oci->errhp, cols[y].value.object_value, 0)))
								return total_qrows;
							cols[y].value.object_value = NULL;
							cols[y].null_struct = NULL;
						}
						else
							strcpy(cols[y].buffer, STR_OPTION(NULLSTR));
						if (OCI_OutputColumn(text, y, cols, cols[y].buffer))
							return total_qrows;
						break;
					default:
						if (cols[y].indicator[x] == -1)
							strcpy(&cols[y].value.str_value[x*cols[y].col_size], STR_OPTION(NULLSTR));
//							memset(&cols[y].value.str_value[x*cols[y].col_size], 0, cols[y].col_size);
						if (OCI_OutputColumn(text, y, cols, &cols[y].value.str_value[x*cols[y].col_size]))
							return total_qrows;
						break;
				}
			}
			//text->noredraw=0;
			if (INT_OPTION(SCREEN))
				WPrintfNL();
			else if (text->spoolfile != NULL)
				FPrintf("\n");

			if (INT_OPTION(LISTVIEW))
			{
				if (fwrite(write_buffer, write_buffer_offset, 1, tmpfile_write) == 0)
				{
					WPrintf("\nCannot write to temp file: %s\n", strerror(errno));
					return total_qrows;
				}
			}
		}
		if (action[0] != '\0' && INT_OPTION(FEEDBACK))
		{
			HANDLE tmp = text->spoolfile;

			text->spoolfile = NULL;
			WPrintfOnly(action, qrows, qrows == 1 ? "" : "s");
			text->spoolfile = tmp;
		}
		total_qrows = qrows;
		if (INT_OPTION(LISTVIEW))
		{
			ListView_SetItemCountEx(hwndListView, total_qrows, LVSICF_NOINVALIDATEALL);
			fflush(tmpfile_write);
		}
	}
	return total_qrows;
}

int OCI_FinishSource(OCIColumnDesc *cols, int ncols)
{
	int x, y;

	if (ora_connected && g_oci->stmtp)
		OCIHandleFree(g_oci->stmtp, OCI_HTYPE_STMT);
	g_oci->stmtp = NULL;
	for (x = 0; x < ncols; x++)
	{
		switch (cols[x].datatype)
		{
			case OCI_TYPECODE_CLOB:
			case OCI_TYPECODE_BLOB:
			case OCI_TYPECODE_BFILE:
				if (cols[x].value.lob_value)
				{
					for (y = 0; y < max_fetch_array_size; y++)
					{
						OCIDescriptorFree(cols[x].value.lob_value[y], OCI_DTYPE_LOB);
					}
					free(cols[x].value.lob_value);
					cols[x].value.lob_value = NULL;
				}
				break;
			case OCI_TYPECODE_OBJECT:
			case OCI_TYPECODE_REF:
				free(cols[x].buffer);
				break;
			default:
				if (cols[x].value.str_value)
				{
					free(cols[x].value.str_value);
					cols[x].value.str_value = NULL;
				}
				break;
		}
		free(cols[x].indicator);
	}
	if (lob_buffer != NULL)
	{
		free(lob_buffer);
		lob_buffer = NULL;
	}
	if (write_buffer)
	{
		free(write_buffer);
		write_buffer = NULL;
	}
	return -1;
}


void OCI_FreeCols(OCIColumnDesc **cols, int *ncols)
{
	if (*cols)
	{
		free(*cols);
		*cols = NULL;
		*ncols = 0;
	}
}