#include "jboci.h"
#include "match.h"

static char* GS_SQL_CONSTRAINTS = "select CONSTRAINT_TYPE,TABLE_NAME,SEARCH_CONDITION,R_OWNER,R_CONSTRAINT_NAME,DELETE_RULE,STATUS,DEFERRABLE,DEFERRED,VALIDATED,GENERATED from all_constraints where owner=:p_schema and constraint_name=:p_constraint";

int ProcessConstraintColumns(HWND hWnd, char* schema, char* constraint)
{
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    OCIStmt* stmtp = NULL;
    char* columnstr = "select COLUMN_NAME from all_cons_columns where owner=:p_schema and constraint_name=:p_constraint order by position";
    char column_name[4001] = "";
    ub2 column_name_i;
    int n, result;

    if (ChkOCIErrS(OCIHandleAlloc(g_oci->envhp,
        (dvoid * *)& stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErrS(OCIStmtPrepare(stmtp,
        g_oci->errhp,
        columnstr,
        (ub4)strlen(columnstr),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_schema",
        (ub4)strlen("p_schema"),
        schema,
        (ub4)strlen(schema) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_constraint",
        (ub4)strlen("p_constraint"),
        constraint,
        (ub4)strlen(constraint) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        1,
        column_name,
        sizeof(column_name),
        SQLT_STR,
        &column_name_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIStmtExecute(g_oci->svchp,
        stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    n = 0;
    for (;;)
    {
        result = OCIStmtFetch(stmtp,
            g_oci->errhp,
            1,
            OCI_FETCH_NEXT,
            OCI_DEFAULT);
        if (result != OCI_SUCCESS && result != OCI_NO_DATA && result != OCI_SUCCESS_WITH_INFO)
        {
            break;
        }
        if (result == OCI_NO_DATA)
            break;
        if (n)
            SchemaPrintf(hWnd, " ,\"%s\"", column_name);
        else
            SchemaPrintf(hWnd, "  \"%s\"", column_name);
        SchemaPrintf(hWnd, "\n");
        n++;
    }
    OCIHandleFree(stmtp, OCI_HTYPE_STMT);
    stmtp = NULL;
    return 0;
}

int ProcessForeignConstraint(HWND hWnd, char* schema, char* r_owner, char* r_constraint)
{
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    OCIStmt* stmtp = NULL;
    char* constraintstr = "select TABLE_NAME from all_constraints where owner=:p_schema and constraint_name=:p_constraint";
    char table_name[129] = "";
    ub2 table_name_i;

    if (ChkOCIErrS(OCIHandleAlloc(g_oci->envhp,
        (dvoid * *)& stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErrS(OCIStmtPrepare(stmtp,
        g_oci->errhp,
        constraintstr,
        (ub4)strlen(constraintstr),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_schema",
        (ub4)strlen("p_schema"),
        r_owner,
        (ub4)strlen(r_owner) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_constraint",
        (ub4)strlen("p_constraint"),
        r_constraint,
        (ub4)strlen(r_constraint) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        1,
        table_name,
        sizeof(table_name),
        SQLT_STR,
        &table_name_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIStmtExecute(g_oci->svchp,
        stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIStmtFetch(stmtp,
        g_oci->errhp,
        1,
        OCI_FETCH_NEXT,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (strcmp(schema, r_owner))
        SchemaPrintf(hWnd, " \"%s\".\"%s\"", r_owner, table_name);
    else
        SchemaPrintf(hWnd, " \"%s\"", table_name);
    OCIHandleFree(stmtp, OCI_HTYPE_STMT);
    stmtp = NULL;
    return 0;
}


int ProcessConstraint(HWND hWnd, char* str)
{
    char* schema, * constraint, * pos;
    char* constraintstr = GS_SQL_CONSTRAINTS;
    char constraint_type[2] = "", table_name[129] = "", search_condition[32000] = "", r_owner[129] = "", r_constraint_name[129] = "", delete_rule[10] = "", status[9] = "", deferrable[15] = "", deferred[10] = "", validated[14] = "", generated[15] = "";
    ub2 constraint_type_i = -1, table_name_i = -1, search_condition_i = -1, r_owner_i = -1, r_constraint_name_i = -1, delete_rule_i = -1, status_i = -1, deferrable_i = -1, deferred_i = -1, validated_i = -1, generated_i = -1;
    OCIBind* bndhp = NULL;
    OCIDefine* defhp = NULL;
    OCIStmt* stmtp = NULL;

    schema = str;
    pos = strchr(str, '.');
    *pos = '\0';
    constraint = pos + 1;

    if (ChkOCIErrS(OCIHandleAlloc(g_oci->envhp,
        (dvoid * *)& stmtp,
        OCI_HTYPE_STMT,
        0,
        NULL)))
    {
        return -1;
    }
    if (ChkOCIErrS(OCIStmtPrepare(stmtp,
        g_oci->errhp,
        constraintstr,
        (ub4)strlen(constraintstr),
        OCI_NTV_SYNTAX,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_schema",
        (ub4)strlen("p_schema"),
        schema,
        (ub4)strlen(schema) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIBindByName(stmtp,
        &bndhp,
        g_oci->errhp,
        "p_constraint",
        (ub4)strlen("p_constraint"),
        constraint,
        (ub4)strlen(constraint) + 1,
        SQLT_STR,
        NULL,
        NULL,
        NULL,
        0,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        return -1;
    }
    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        1,
        constraint_type,
        sizeof(constraint_type),
        SQLT_STR,
        &constraint_type_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        2,
        table_name,
        sizeof(table_name),
        SQLT_STR,
        &table_name_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        3,
        search_condition,
        sizeof(search_condition),
        SQLT_STR,
        &search_condition_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        4,
        &r_owner,
        sizeof(r_owner),
        SQLT_STR,
        &r_owner_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        5,
        &r_constraint_name,
        sizeof(r_constraint_name),
        SQLT_STR,
        &r_constraint_name_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        6,
        &delete_rule,
        sizeof(delete_rule),
        SQLT_STR,
        &delete_rule_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        7,
        &status,
        sizeof(status),
        SQLT_STR,
        &status_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        8,
        &deferrable,
        sizeof(deferrable),
        SQLT_STR,
        &deferrable_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        9,
        &deferred,
        sizeof(deferred),
        SQLT_STR,
        &deferred_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        10,
        &validated,
        sizeof(validated),
        SQLT_STR,
        &validated_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIDefineByPos(stmtp,
        &defhp,
        g_oci->errhp,
        11,
        &generated,
        sizeof(generated),
        SQLT_STR,
        &generated_i,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }

    if (ChkOCIErrS(OCIStmtExecute(g_oci->svchp,
        stmtp,
        g_oci->errhp,
        0,
        0,
        NULL,
        NULL,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    if (ChkOCIErrS(OCIStmtFetch(stmtp,
        g_oci->errhp,
        1,
        OCI_FETCH_NEXT,
        OCI_DEFAULT)))
    {
        OCIHandleFree(stmtp, OCI_HTYPE_STMT);
        stmtp = NULL;
        return -1;
    }
    OCIHandleFree(stmtp, OCI_HTYPE_STMT);
    stmtp = NULL;
    if (strcmp(constraint_type, "C") || !match("\"*\" IS NOT NULL", search_condition))
    {
        SchemaPrintf(hWnd, "ALTER TABLE \"%s\" ADD", table_name);
        if (generated_i != (ub2)-1 && !strcmp(generated, "USER NAME"))
            SchemaPrintf(hWnd, " CONSTRAINT \"%s\"", constraint);
        if (!strcmp(constraint_type, "C"))
        {
            SchemaPrintf(hWnd, "\nCHECK (%s)\n", search_condition);
        }
        if (!strcmp(constraint_type, "P"))
        {
            SchemaPrintf(hWnd, "\nPRIMARY KEY (\n");
            ProcessConstraintColumns(hWnd, schema, constraint);
            SchemaPrintf(hWnd, ")\n");
        }
        if (!strcmp(constraint_type, "U"))
        {
            SchemaPrintf(hWnd, "\nUNIQUE (\n");
            ProcessConstraintColumns(hWnd, schema, constraint);
            SchemaPrintf(hWnd, ")\n");
        }
        if (!strcmp(constraint_type, "R"))
        {
            SchemaPrintf(hWnd, "\nFOREIGN KEY (\n");
            ProcessConstraintColumns(hWnd, schema, constraint);
            SchemaPrintf(hWnd, ")\nREFERENCES");
            ProcessForeignConstraint(hWnd, schema, r_owner, r_constraint_name);
            SchemaPrintf(hWnd, " (\n");
            ProcessConstraintColumns(hWnd, r_owner, r_constraint_name);
            SchemaPrintf(hWnd, ")\n");
        }
        if (deferrable_i != (ub2)-1)
            SchemaPrintf(hWnd, "%s\n", deferrable);
        if (deferred_i != (ub2)-1)
            SchemaPrintf(hWnd, "INITIALLY %s\n", deferred);
        if (status_i != (ub2)-1)
        {
            if (!strcmp(status, "ENABLED"))
            {
                SchemaPrintf(hWnd, "ENABLE");
                if (validated_i != (ub2)-1)
                {
                    if (!strcmp(validated, "VALIDATE"))
                        SchemaPrintf(hWnd, " VALIDATE");
                    else
                        SchemaPrintf(hWnd, " NOVALIDATE");
                }
                SchemaPrintf(hWnd, "\n");
            }
            else
                SchemaPrintf(hWnd, "DISABLE\n");
        }
        SchemaPrintf(hWnd, "/\n\n");
    }
    return 0;
}

