#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jbodbc.h"

#define MAX_ERROR_SIZE 200

ODBCEnvironment *g_odbc;
int odbc_connected = 0;

int CheckODBCError(SQLRETURN ret, SQLSMALLINT handle_type, SQLHANDLE handle, char *file, int line, int show, HWND hWnd)
{
	SQLCHAR state[6];
	SQLINTEGER native_error;
	SQLCHAR error_msg[MAX_ERROR_SIZE + 1];
	SQLSMALLINT msg_len;
	SQLRETURN status;

	status = SQLGetDiagRec(handle_type, handle, 1, state, &native_error, error_msg, MAX_ERROR_SIZE, &msg_len);
	if (status == SQL_SUCCESS || ret == SQL_SUCCESS)
	{

		switch (ret)
		{
			case SQL_SUCCESS:
				return 0;
				break;
			case SQL_SUCCESS_WITH_INFO:
				if (show == 1)
				{
					//WPrintf("\rODBC warning: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					WPrintf("\rWarning: %s\n", error_msg);
				}
				else if (show == 2)
				{
					//          AnalyzeError("ODBC warning: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					AnalyzeError("Warning: msg=%s\n", error_msg);
				}
				else if (show == 3)
				{
					//          SchemaPrintf(hWnd,"ODBC warning: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					SchemaPrintf(hWnd, "Warning: msg=%s\n", error_msg);
				}
				return 0;
				break;
			case SQL_ERROR:
				if (show == 1)
					//          WPrintf("\rODBC error: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					WPrintf("\r%s\n", error_msg);
				else if (show == 2)
					//          AnalyzeError("\rODBC error: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					AnalyzeError("\r%s\n", error_msg);
				else if (show == 3)
					//          SchemaPrintf(hWnd,"\rODBC error: SQLSTATE=%s, code=%d, msg=%s, file=%s, line=%d\n",state,native_error,error_msg,file,line);
					SchemaPrintf(hWnd, "\r%s\n", error_msg);
				break;
			case SQL_INVALID_HANDLE:
				if (show == 1)
					WPrintf("Error - ODBC_INVALID_HANDLE (%s, line %d)\n", file, line);
				else if (show == 2)
					AnalyzeError("Error - ODBC_INVALID_HANDLE (%s, line %d)", file, line);
				else if (show == 3)
					SchemaPrintf(hWnd, "Error - ODBC_INVALID_HANDLE (%s, line %d)", file, line);
				break;
			default:
				break;
		}
	}
	else
	{
		WPrintf("\rError looking up error code\n");
	}
	return 1;
}


int InitODBC(void)
{
	SQLRETURN ret;

	g_odbc = malloc(sizeof(ODBCEnvironment));
	if (g_odbc == NULL)
		return -1;
	memset(g_odbc, 0, sizeof(ODBCEnvironment));

	ret = SQLAllocHandle(SQL_HANDLE_ENV, NULL, &g_odbc->env_handle_source);
	if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
	{
		ChkODBCErr(ret, SQL_HANDLE_ENV, g_odbc->env_handle_source);
		return 1;
	}
	ret = SQLSetEnvAttr(g_odbc->env_handle_source, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, 0);
	if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
	{
		ChkODBCErr(ret, SQL_HANDLE_ENV, g_odbc->env_handle_source);
		return 1;
	}
	ret = SQLAllocHandle(SQL_HANDLE_DBC, g_odbc->env_handle_source, &g_odbc->dbc_handle_source);
	if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
	{
		ChkODBCErr(ret, SQL_HANDLE_ENV, g_odbc->env_handle_source);
		return 1;
	}
	g_odbc->stmt_handle_source = NULL;

	return 0;
}

int ConnectToODBC(char *username, char *password, char *database, int add)
{
	SQLRETURN ret;
	char str[512] = "", str2[512] = "";
	SQLSMALLINT strsize;

	if (odbc_connected)
		DisconnectFromODBC();
    if (ora_connected)
    {
        ClearList(hwndListView,1);
        DisconnectFromOracle();
    }


	SaveUserCredentials(username, database, DS_ODBC);

	ret = SQLConnect(g_odbc->dbc_handle_source, (SQLCHAR*)database, (SQLSMALLINT)strlen(database),
		(SQLCHAR*)username, (SQLSMALLINT)strlen(username),
		(SQLCHAR*)password, (SQLSMALLINT)strlen(password));
	if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
	{
		ChkODBCErr(ret, SQL_HANDLE_DBC, g_odbc->dbc_handle_source);
		return 1;
	}

	SQLGetInfo(g_odbc->dbc_handle_source, SQL_DBMS_NAME, str, 512, &strsize);
	SQLGetInfo(g_odbc->dbc_handle_source, SQL_DBMS_VER, str2, 512, &strsize);

	odbc_connected = 1;
	ora_connected = 0;
	WPrintf("\rConnected to %s %s\n", str, str2);
	strcpy(g_database, database);
	SetWindowTitle(username, database, 0, 0);
	ComposeConnectString(username, password, database, "", "odbc");
	strncpy(g_text->username, username, 100);
	_strupr(g_text->username);
	strncpy(g_text->database, database, 200);
	_strupr(g_text->database);
	strcpy(g_text->privilege, "");
	_strupr(g_text->privilege);
	ClearCommand(g_text, add, 0);
	GetDefaultOptions(g_text);
	SetPWD(g_text);
	return 0;
}

int DisconnectFromODBC(void)
{
	SQLDisconnect(g_odbc->dbc_handle_source);
	odbc_connected = 0;
	SetWindowTitle("", "", 0, 0);

	if (g_odbc->stmt_handle_source)
		SQLFreeHandle(SQL_HANDLE_DBC, g_odbc->stmt_handle_source);
	g_odbc->stmt_handle_source = NULL;
	if (g_odbc->dbc_handle_source)
		SQLFreeHandle(SQL_HANDLE_DBC, g_odbc->dbc_handle_source);
	g_odbc->dbc_handle_source = NULL;
	if (g_odbc->env_handle_source)
		SQLFreeHandle(SQL_HANDLE_ENV, g_odbc->env_handle_source);
	g_odbc->env_handle_source = NULL;

	g_text->username[0] = '\0';
	g_text->database[0] = '\0';
	g_text->privilege[0] = '\0';

	free(g_odbc);
	InitODBC();
	return 0;
}

