#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jbodbc.h"

SQLULEN rowsfetched;

extern char* write_buffer;
extern int write_buffer_offset;


void ODBC_PrintHeaders(TextWidget* text, ODBCColumnDesc* cols, int a_ncols)
{
    int x;
    char* lines;

    if (INT_OPTION(HEADING))
    {
        if (INT_OPTION(SCREEN))
        {
            WPrintf("\r");
            for (x = 0; x < (int)a_ncols; x++)
            {
                if (x)
                    WPrintf(" %-*s", MAX(cols[x].col_size - 1, strlen((char*)cols[x].col_name)), cols[x].col_name);
                else
                    WPrintf("%-*s", MAX(cols[x].col_size - 1, strlen((char*)cols[x].col_name)), cols[x].col_name);
            }
            WPrintfNL();
            for (x = 0; x < (int)a_ncols; x++)
            {
                if (x)
                    WPrintf(" ");
                lines = (char*)malloc(MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name)) + 1);
                memset(lines, '-', MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name)));
                lines[MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name))] = '\0';
                WPrintf("%s", lines);
                free(lines);
            }
            //text->noredraw=0;
            WPrintfNL();
        }
        else if (text->spoolfile != NULL)
        {
            size_t len;

            for (x = 0; x < (int)a_ncols; x++)
            {
                len = MAX(cols[x].col_size - 1, strlen((char*)cols[x].col_name));
                if (INT_OPTION(TRIMSPOOL) && x == a_ncols - 1)
                    len = strlen((char*)cols[x].col_name);
                if (x)
                    FPrintf(" %-*s", len, cols[x].col_name);
                else
                    FPrintf("%-*s", len, cols[x].col_name);
            }
            FPrintf("\n");
            for (x = 0; x < (int)a_ncols; x++)
            {
                if (x)
                    FPrintf(" ");
                lines = (char*)malloc(MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name)) + 1);
                memset(lines, '-', MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name)));
                lines[MAX(cols[x].col_size - 1, (int)strlen((char*)cols[x].col_name))] = '\0';
                FPrintf("%s", lines);
                free(lines);
            }
            FPrintf("\n");
        }
    }
}

int ODBC_PrepareSource(TextWidget * text, char* p_sqlstr, ODBCColumnDesc * *p_cols, int* p_ncols, char* action)
{
    int        x;
    SQLSMALLINT dtype;
    SQLCHAR     col_name[250];
    SQLSMALLINT col_name_len;
    SQLSMALLINT ncols;
    SQLSMALLINT dec_digits;
    SQLSMALLINT nullable;
    ODBCColumnDesc* pcols;
    SQLULEN    len;
    char* char_ptr;
    int        record_size = 0;
    int        total_memory, parse = 0;
    SQLRETURN  ret;

    *p_ncols = 0;
    if (!odbc_connected)
    {
        WPrintf("Not connected to a database\n");
        return -1;
    }
    if (g_odbc->stmt_handle_source == NULL)
    {
        parse = 1;
        ret = SQLAllocHandle(SQL_HANDLE_STMT, g_odbc->dbc_handle_source, &g_odbc->stmt_handle_source);
        if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
        {
            ChkODBCErr(ret, SQL_HANDLE_DBC, g_odbc->dbc_handle_source);
            return 1;
        }
        if (action[0] != '\0' && INT_OPTION(FEEDBACK))
        {
            HANDLE tmp = text->spoolfile;

            text->spoolfile = NULL;
            WPrintfOnly("\rParsing statement...");
            text->spoolfile = tmp;
        }
    }
    ret = SQLSetStmtAttr(g_odbc->stmt_handle_source, SQL_ATTR_ROW_ARRAY_SIZE, (SQLPOINTER)(unsigned long long)INT_OPTION(ARRAYSIZE), SQL_IS_UINTEGER);
    if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
    {
        ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
        return 1;
    }
    ret = SQLSetStmtAttr(g_odbc->stmt_handle_source, SQL_ATTR_ROWS_FETCHED_PTR, (SQLPOINTER)& rowsfetched, SQL_IS_POINTER);
    if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
    {
        ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
        return 1;
    }
    if (parse)
    {
        if (action[0] != '\0' && INT_OPTION(FEEDBACK))
        {
            HANDLE tmp = text->spoolfile;

            text->spoolfile = NULL;
            WPrintfOnly("\rExecuting query...");
            text->spoolfile = tmp;
        }
        ret = SQLExecDirect(g_odbc->stmt_handle_source, (SQLCHAR*)p_sqlstr, (SQLINTEGER)strlen(p_sqlstr));
        if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
        {
            ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
            return 1;
        }
    }
    ret = SQLNumResultCols(g_odbc->stmt_handle_source, &ncols);
    if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
    {
        ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
        return 1;
    }
    //  Log("Number of result columns: %d\n",(int)ncols);
    if ((pcols = (ODBCColumnDesc*)malloc(sizeof(ODBCColumnDesc) * ncols)) == NULL)
    {
        WPrintf("Can't perform malloc!\n");
        return 1;
    }
    for (x = 0; x < (int)ncols; x++)
    {
        ret = (SQLRETURN)SQLDescribeCol(g_odbc->stmt_handle_source, (SQLUSMALLINT)x + 1, col_name, 250, &col_name_len, &dtype, &len, &dec_digits, &nullable);
        if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
        {
            ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
            return 1;
        }
        pcols[x].col_name = (char*)malloc(col_name_len + 1);
        memcpy(pcols[x].col_name, col_name, col_name_len);
        pcols[x].col_name[col_name_len] = '\0';
        pcols[x].datatype = dtype;
        switch (dtype)
        {
            case SQL_TYPE_DATE:
            case SQL_TYPE_TIME:
            case SQL_TYPE_TIMESTAMP:
            case SQL_INTERVAL_MONTH:
            case SQL_INTERVAL_YEAR:
            case SQL_INTERVAL_YEAR_TO_MONTH:
            case SQL_INTERVAL_DAY:
            case SQL_INTERVAL_HOUR:
            case SQL_INTERVAL_MINUTE:
            case SQL_INTERVAL_SECOND:
            case SQL_INTERVAL_DAY_TO_HOUR:
            case SQL_INTERVAL_DAY_TO_MINUTE:
            case SQL_INTERVAL_DAY_TO_SECOND:
            case SQL_INTERVAL_HOUR_TO_MINUTE:
            case SQL_INTERVAL_HOUR_TO_SECOND:
            case SQL_INTERVAL_MINUTE_TO_SECOND:
                pcols[x].col_size = 40;
                break;
            case SQL_DECIMAL:
            case SQL_NUMERIC:
            case SQL_SMALLINT:
            case SQL_INTEGER:
            case SQL_REAL:
            case SQL_FLOAT:
            case SQL_DOUBLE:
            case SQL_BIT:
            case SQL_TINYINT:
            case SQL_BIGINT:
            case SQL_GUID:
                pcols[x].col_size = 40;
                break;
            case SQL_CHAR:
            case SQL_VARCHAR:
            case SQL_WCHAR:
            case SQL_WVARCHAR:
            case SQL_BINARY:
            case SQL_VARBINARY:
                pcols[x].col_size = len + 1;
                break;
            case SQL_LONGVARCHAR:
            case SQL_WLONGVARCHAR:
            case SQL_LONGVARBINARY:
                pcols[x].col_size = INT_OPTION(LONG_SIZE) + 1;
                break;
            default:
                WPrintf("\rUnknown datatype %d for column %s\n", dtype, pcols[x].col_name);
                return -5;
        }
        if (pcols[x].col_size < (SQLUINTEGER)strlen(STR_OPTION(NULLSTR)) + 1)
            pcols[x].col_size = (SQLUINTEGER)strlen(STR_OPTION(NULLSTR)) + 1;
        record_size += (int)pcols[x].col_size;
    }
    total_memory = INT_OPTION(ARRAYSIZE) * record_size;
    if (total_memory<0 || total_memory>(int)INT_OPTION(MAXMEM))
    {
        max_fetch_array_size = INT_OPTION(MAXMEM) / record_size;
        if (max_fetch_array_size < 1)
            max_fetch_array_size = 1;
    }
    else
        max_fetch_array_size = INT_OPTION(ARRAYSIZE);
    write_buffer = malloc(record_size);
    if (write_buffer == NULL)
    {
        WPrintf("Can't perform malloc!\n");
        return 1;
    }
    for (x = 0; x < (int)ncols; x++)
    {
        if ((pcols[x].indicator = (SQLLEN*)malloc(max_fetch_array_size * sizeof(SQLLEN))) == NULL)
        {
            WPrintf("Can't perform malloc!\n");
            return 1;
        }
        if ((pcols[x].value.str_value = (char*)malloc(max_fetch_array_size * (pcols[x].col_size))) == NULL)
        {
            WPrintf("Can't perform malloc!\n");
            return 1;
        }
        char_ptr = pcols[x].value.str_value;
        len = pcols[x].col_size;
        ret = SQLBindCol(g_odbc->stmt_handle_source, (SQLUSMALLINT)x + 1, SQL_CHAR, pcols[x].value.str_value, pcols[x].col_size, (SQLLEN*)pcols[x].indicator);
        if (ret != SQL_SUCCESS)
        {
            ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
            return 1;
        }
    }
    *p_ncols = ncols;
    *p_cols = pcols;

    return 0;
}

int ODBC_OutputColumn(TextWidget * text, int y, ODBCColumnDesc * cols, char* buffer)
{
    if (INT_OPTION(SCREEN))
    {
        if (y)
            WPrintf(" %-*s", MAX(cols[y].col_size - 1, strlen((char*)cols[y].col_name)), buffer);
        else
            WPrintf("\r%-*s", MAX(cols[y].col_size - 1, strlen((char*)cols[y].col_name)), buffer);
    }
    else if (text->spoolfile != NULL)
    {
        size_t len = MAX(cols[y].col_size - 1, strlen((char*)cols[y].col_name));
        if (INT_OPTION(TRIMSPOOL) && y == ncols - 1)
            len = strlen(buffer);
        if (y)
            FPrintf(" %-*s", len, buffer);
        else
            FPrintf("\r%-*s", len, buffer);
    }
    if (INT_OPTION(LISTVIEW))
    {
        WriteToBuffer(buffer, (int)cols[y].col_size);
    }
    return 0;
}


int ODBC_FetchData(TextWidget * text, ODBCColumnDesc * cols, int ncols, char* action)
{
    int nrows, x, y;
    SQLULEN fetch_array_size = max_fetch_array_size;
    sword ret;

    query_finished = 0;
    total_qrows = 0;
    while (!query_finished)
    {
        if (INT_OPTION(FAST_RESULTS) && total_qrows < (int)INT_OPTION(FAST_RESULTS))
            fetch_array_size = 1;
        else
            fetch_array_size = max_fetch_array_size;
        ret = SQLSetStmtAttr(g_odbc->stmt_handle_source, SQL_ATTR_ROW_ARRAY_SIZE, (SQLPOINTER)fetch_array_size, SQL_IS_UINTEGER);
        if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
        {
            ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
            return 1;
        }
        ret = SQLFetch(g_odbc->stmt_handle_source);
        if (ret == SQL_INVALID_HANDLE || ret == SQL_ERROR)
        {
            ChkODBCErr(ret, SQL_HANDLE_STMT, g_odbc->stmt_handle_source);
            return -1;
        }
        if (ret == SQL_NO_DATA)
        {
            return total_qrows;
        }

        nrows = (int)rowsfetched;
        if (total_qrows == 0 && rowsfetched)
            ODBC_PrintHeaders(text, cols, ncols);
        if (nrows != fetch_array_size)
            query_finished = 1;


        for (x = 0; x < nrows; x++)
        {
            write_buffer_offset = 0;

            for (y = 0; y < ncols; y++)
            {
                if (cols[y].indicator[x] == SQL_NULL_DATA)
                {
                    // memset(&cols[y].value.str_value[x*cols[y].col_size], 0, cols[y].col_size);
                    strcpy(&cols[y].value.str_value[x * cols[y].col_size], STR_OPTION(NULLSTR));
                }
                if (ODBC_OutputColumn(text, y, cols, &cols[y].value.str_value[x * cols[y].col_size]))
                    return total_qrows;
            }
            if (INT_OPTION(LISTVIEW))
            {
                if (fwrite(write_buffer, write_buffer_offset, 1, tmpfile_write) == 0)
                {
                    WPrintf("Cannot write to temp file: %s", strerror(errno));
                    return 1;
                }
                fflush(tmpfile_write);
            }
            //text->noredraw=0;
            if (INT_OPTION(SCREEN))
                WPrintfNL();
            else if (text->spoolfile != NULL)
                FPrintf("\n");
        }
        total_qrows += (int)rowsfetched;
        if (action[0] != '\0' && INT_OPTION(FEEDBACK))
        {
            HANDLE tmp = text->spoolfile;

            text->spoolfile = NULL;
            WPrintfOnly(action, total_qrows, total_qrows == 1 ? "" : "s");
            text->spoolfile = tmp;
        }
        if (INT_OPTION(LISTVIEW))
        {
            ListView_SetItemCountEx(hwndListView, total_qrows, LVSICF_NOINVALIDATEALL);
        }
    }
    return total_qrows;
}

int ODBC_FinishSource(ODBCColumnDesc * cols, int ncols)
{
    int x;

    for (x = 0; x < ncols; x++)
    {
        if (cols[x].value.str_value)
        {
            free(cols[x].value.str_value);
            cols[x].value.str_value = NULL;
        }
        free(cols[x].indicator);
    }
    SQLCloseCursor(g_odbc->stmt_handle_source);
    SQLFreeHandle(SQL_HANDLE_STMT, &g_odbc->stmt_handle_source);
    g_odbc->stmt_handle_source = NULL;
    if (write_buffer)
    {
        free(write_buffer);
        write_buffer = NULL;
    }
    return -1;
}


void ODBC_FreeCols(ODBCColumnDesc * *cols, int* ncols)
{
    if (*cols)
    {
        free(*cols);
        *cols = NULL;
        *ncols = 0;
    }
}