#include "jbsql.h"
#include <Wininet.h>

LRESULT CALLBACK VersionProc(HWND hWnd,
    UINT uMessage,
    WPARAM wParam,
    LPARAM lParam)
{
    switch (uMessage)
    {
        case WM_INITDIALOG:
            SetDlgItemText(hWnd, IDC_VERSIONTEXT, (LPCSTR)lParam);
            break;
        case WM_COMMAND:
            switch (GET_WM_COMMAND_ID(wParam, lParam))
            {
                case IDYES:
                    EndDialog(hWnd, TRUE);
                    return TRUE;
                case IDNO:
                    EndDialog(hWnd, FALSE);
                    return TRUE;
            }
            break;

        case WM_CLOSE:
            EndDialog(hWnd, FALSE);
            return TRUE;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case NM_CLICK:
                case NM_RETURN:
                    {
                        switch (GET_WM_COMMAND_ID(wParam, lParam))
                        {
                            case IDC_WHATSNEW:
                                {
                                    PNMLINK pNMLink = (PNMLINK)lParam;
                                    LITEM   item = pNMLink->item;

                                    ShellExecuteW(NULL, L"open", item.szUrl, NULL, NULL, SW_SHOW);
                                    break;
                                }
                        }
                        break;
                    }
            }
        default:
            break;
    }
    return FALSE;
}



int ShowNewVersionDialog(char* text)
{
    FARPROC cpproc;
    LRESULT res;

    cpproc = MakeProcInstance((FARPROC)VersionProc, hInst);
    res = DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_NEWVERSION), g_hWnd, (DLGPROC)cpproc, (LPARAM)text);
    FreeProcInstance(cpproc);

    return (int)res;
}

void CheckNewVersion(HWND hWnd, int feedback)
{
    char* server = "users.telenet.be";
    HINTERNET hInternet1, hInternet2;
    DWORD bytesavailable = 0, bytesread;

    hInternet1 = InternetOpen("JBSQL", INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0);
    if (hInternet1 == NULL)
    {
        DisplayLastError(hWnd);
        return;
    }
    hInternet2 = InternetOpenUrl(hInternet1, "http://users.telenet.be/jbosman/jbsql_version.txt", NULL, 0, 0, (DWORD_PTR)NULL);
    if (hInternet2 == NULL)
    {
        DisplayLastError(hWnd);
        InternetCloseHandle(hInternet1);
        return;
    }

    if (InternetQueryDataAvailable(hInternet2, &bytesavailable, 0, (DWORD_PTR)NULL))
    {
        DWORD httpstatus = 0, len = sizeof(DWORD);
        if (!HttpQueryInfo(hInternet2, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &httpstatus, &len, 0))
        {
            DisplayLastError(hWnd);
            InternetCloseHandle(hInternet1);
            InternetCloseHandle(hInternet2);
            return;
        }
        if (httpstatus != 200)
        {
            InternetCloseHandle(hInternet1);
            InternetCloseHandle(hInternet2);
            Error(hWnd, "HTTP request returned status %d", httpstatus);
            return;
        }
        char* str = malloc((size_t)bytesavailable + 1);
        if (!str)
        {
            InternetCloseHandle(hInternet1);
            InternetCloseHandle(hInternet2);
            Error(hWnd, "Out of memory");
            return;
        }
        InternetReadFile(hInternet2, str, bytesavailable, &bytesread);
        InternetCloseHandle(hInternet2);
        str[bytesread] = '\0';
        Trim(str);
        if (bytesread > 20)
        {
            Error(hWnd, "Unable to check for a new version, check your internet connection");
            InternetCloseHandle(hInternet1);
            return;
        }
        int p1, p2, p3, p4;
        if (sscanf(str, "%d.%d.%d.%d", &p1, &p2, &p3, &p4) != 4)
        {
            Error(hWnd, "Unable to check for a new version, check your internet connection");
            InternetCloseHandle(hInternet1);
            return;
        }
        int c1, c2, c3, c4;
        if (sscanf(JBSQL_VERSION, "%d.%d.%d.%d", &c1, &c2, &c3, &c4) != 4)
        {
            Error(hWnd, "Unable to check for a new version, check your internet connection");
            InternetCloseHandle(hInternet1);
            return;
        }
        if (p1 > c1 ||
            (p1 == c1 && p2 > c2) ||
            (p1 == c1 && p2 == c2 && p3 > c3) ||
            (p1 == c1 && p2 == c2 && p3 == c3 && p4 > c4)
           )
        {
            char msg[250];

            sprintf(msg, "A new version is available!\nCurrent version: %s\nNew version: %s\nDo you want to upgrade?", JBSQL_VERSION, str);
            if (ShowNewVersionDialog(msg))
            {
                char old_executable[MAX_PATH], new_executable[MAX_PATH];

                if ((sizeof(size_t) == 8))
                    hInternet2 = InternetOpenUrl(hInternet1, "http://users.telenet.be/jbosman/jbsql_x64.exe", NULL, 0, 0, (DWORD_PTR)NULL);
                else
                    hInternet2 = InternetOpenUrl(hInternet1, "http://users.telenet.be/jbosman/jbsql_x32.exe", NULL, 0, 0, (DWORD_PTR)NULL);
                if (hInternet2 == NULL)
                {
                    DisplayLastError(hWnd);
                    InternetCloseHandle(hInternet1);
                    return;
                }

                if (InternetQueryDataAvailable(hInternet2, &bytesavailable, 0, (DWORD_PTR)NULL))
                {
                    if (!HttpQueryInfo(hInternet2, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &httpstatus, &len, 0))
                    {
                        DisplayLastError(hWnd);
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        return;
                    }
                    if (httpstatus != 200)
                    {
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        Error(hWnd, "HTTP request returned status %d", httpstatus);
                        return;
                    }
                    char* buf = malloc(bytesavailable);
                    if (!buf)
                    {
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        Error(hWnd, "Out of memory");
                        return;
                    }
                    sprintf(new_executable, "%s.%s", executable, str);
                    FILE* f = fopen(new_executable, "wb");
                    if (f == NULL)
                    {
                        Error(hWnd, "Cannot create file %s: %s", new_executable, strerror(errno));
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        return;
                    }
                    for (;;)
                    {
                        InternetReadFile(hInternet2, buf, bytesavailable, &bytesread);
                        if (bytesread == 0)
                            break;
                        fwrite(buf, bytesread, 1, f);
                    }
                    fclose(f);
                    InternetCloseHandle(hInternet2);
                    sprintf(old_executable, "%s.%s", executable, JBSQL_VERSION);
                    _unlink(old_executable);
                    if (rename(executable, old_executable) != 0)
                    {
                        Error(hWnd, "Could not rename old exe %s to new exe %s: %s", executable, old_executable, strerror(errno));
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        return;
                    }
                    if (rename(new_executable, executable) != 0)
                    {
                        Error(hWnd, "Could not rename new exe %s to real exe %s", new_executable, executable, strerror(errno));
                        InternetCloseHandle(hInternet1);
                        InternetCloseHandle(hInternet2);
                        return;
                    }
                    NewWindow(g_text);
                    free(buf);
                    exit(0);
                }
            }
            free(str);
        }
        else
        {
            if (feedback)
                MessageBox(hWnd, "No new version available", "No new version available", MB_ICONASTERISK);
        }
    }
    else
    {
        DisplayLastError(hWnd);
        InternetCloseHandle(hInternet1);
        InternetCloseHandle(hInternet2);
        return;
    }
    InternetCloseHandle(hInternet1);
    InternetCloseHandle(hInternet2);
}


void CheckNewVersionSilent(HWND hWnd)
{
    CheckNewVersion(hWnd, 0);
}